/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip.h
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Main header file for the "Automatic Tour Into the Picture" library (ATIP)


#ifndef _ATIP_H_
#define _ATIP_H_

#include <atip_image.h>


//! Initialize the ATIP library
//! \return true if there was no error, else call atipGetError()
bool atipInit();

//! Add an image taken from a file to the ATIP set of images
//! \param fileName Name of the image file to load
//! \return true if there was no error, else call atipGetError()
bool atipAddImageFromFile(const char * fileName);

//! Add an image taken from memory to the ATIP set of images
//! \param image Image to send to the ATIP set of images
void atipAddImageFromMemory(const ATIPimage & image);

//----------------------------------------------------------------------------------------

//! Remove an image from the ATIP set of images
//! \note If the index does not correspond to an image, nothing appends
//! \param index Zero-based index of the image to remove
void atipRemoveImage(unsigned int index);

//! Get the desired image from the set of ATIP images
//! \param index Zero-based index of the image to get
//! \return Pointer to the desired image, 0 if the given index is invalid
const ATIPimage * atipGetImage(unsigned int index);

//! Get the number images inside the ATIP set of images
//! \return The number of loaded images
unsigned int atipGetNumLoadedImages();

//----------------------------------------------------------------------------------------

//! Tour Into the Picture data, created by the maker, and used by the navigators
typedef struct {
		float vertices[24];			//!< Array of 8 triplets (x, y, z), one per vertex,
									//!< unrotated version (screen aligned)
		float rotVertices[24];		//!< Array of 8 triplets (x, y, z), one per vertex,
									//!< rotated version (multiplied by rot[9])
		unsigned int indices[20];	//!< Array of 5 quadruplets (v0, v1, v2, v3)
									//!< representing the indices of the vertices
									//!< to use per quadrilateral
        ATIPimage * textures[5];	//!< Array of 5 textures (background, left
									//!< wall, right wall, floor, ceiling).
									//!< The pointers are nul if there is no texture
		float focal;				//!< Focal length (in screen space)
		float fov;					//!< Field-of-view (in degrees)
		float rot[9];				//!< 3x3 rotation matrix

		float distance;				//!< Distance from the viewer to the background plane
		float xLeft;				//!< Distance of the left wall from the point in
									//!< front of the viewer on the background plane
		float xRight;				//!< Distance of the right wall from the point in
									//!< front of the viewer on the background plane
		float yFloor;				//!< Distance of the floor from the point in
									//!< front of the viewer on the background plane
		float yCeiling;				//!< Distance of the ceiling from the point in
									//!< front of the viewer on the background plane

		unsigned int width;			//!< Width of the original image in pixels
		unsigned int height;		//!< Height of the original image in pixels
				}	ATIPdata;

//----------------------------------------------------------------------------------------

//@{
//! Processing flags constants

//! Detect the vanishing points of the image
#define ATIP_PROCESSFLAG_DETECT_VANISHING_POINTS		1

//! Edit the polygons of the image (with the already computed vanishing points)
#define ATIP_PROCESSFLAG_EDIT_POLYGONS					2

//! Compute the textures (with the vanishing points already computed, and
//! the polygons already edited)
#define ATIP_PROCESSFLAG_COMPUTE_TEXTURES				4

//! All processing steps
#define ATIP_PROCESSFLAG_ALL		(  ATIP_PROCESSFLAG_DETECT_VANISHING_POINTS		\
									 | ATIP_PROCESSFLAG_EDIT_POLYGONS				\
									 | ATIP_PROCESSFLAG_COMPUTE_TEXTURES)
//@}


//! Do the images processing to create ATIP compatible data
//! \param atipData Resulting ATIPdata structure
//! \param flags Combination of ATIP_PROCESSFLAG_xxx constants
//! \return true if there was no error, else call atipGetError()
bool atipProcess(ATIPdata & atipData, unsigned char flags = ATIP_PROCESSFLAG_ALL);

//! Definition of the polygons editor function pointer
//! \param atipData Reference to the data of the TIP
//! \param origImage Reference to the original image
//! \param nFiniteVanPoints Number of finite vanishing points
//! \param nInfiniteVanPoints Number of infinite vanishing points
//! \return true if there was no error, else call atipGetError()
typedef bool (* ATIPpolygonsEditorProc) (ATIPdata & atipData, ATIPimage & origImage,
										 unsigned int nFiniteVanPoints,
										 unsigned int nInfiniteVanPoints);

//! Set the polygons editor callback.
//! This callback is called when the polygons editor is needed
//! \param proc Pointer to a function which opens a polygons editor
void atipSetPolygonsEditorCallback(ATIPpolygonsEditorProc proc);


#endif	// _ATIP_H_