/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_image.h
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Image class (for any type of image)


#ifndef _ATIP_IMAGE_H_
#define _ATIP_IMAGE_H_

#include <atip_tools.h>
#include <memory.h>


//! \class ATIPimage
//! Image class.
//! This class contains information about the image, and the data.
//! \note The pixels are stored from the top to the bottom of the image.
class ATIPimage
{
public:

	//! Default constructor.
	//! Creates a default image for debugging (128x128 checkerboard in 24 bits)
	ATIPimage();

	//! Constructor which creates an image of the given dimensions.
	//! \warning The image data are not initialized
	//! \param width Width of the image in pixels
	//! \param height Height of the image in pixels
	//! \param bpp Number of bits per pixel.
	//!            If bpp is less or equal than 8, a palette is used (use SetPalette()
	//!            to set it) and is a greyscale by default. If bpp = 24, 8 bits
	//!            are used for the red, the green and the blue components.
	//!            If bpp = 32, a 8 bits alpha component is added.
	ATIPimage(unsigned int width, unsigned int height, unsigned char bpp);

	//! Constructor which creates an image from data in memory.
	//! Returns ATIP_INVALID_ARGUMENT if an argument is bad
	//! \param width Width of the image in pixels
	//! \param height Height of the image in pixels
	//! \param bpp Number of bits per pixel.
	//!            If bpp is less or equal than 8, a palette is used (use SetPalette()
	//!            to set it) and is a greyscale by default. If bpp = 24 or 32, 8 bits
	//!            are used for the red, the green and the blue components
	//! \param data The data of the image
	//! \param copy True if data must be copied (in this case, the original data can
	//!             be deleted)
	ATIPimage(unsigned int width, unsigned int height, unsigned char bpp,
			  void * data, bool copy = true);

	//! Copy constructor
	//! \param image Reference to the original image
	ATIPimage(const ATIPimage & image);

	//! Destructor
	~ATIPimage();

	//------------------------------------------------------------------------------------

	//! Assignment operator
	//! \param image Reference to the original image
	//! \return Reference to the new image
	ATIPimage & operator = (const ATIPimage & image);

	//! Load an image from a file.
	//! Formats supported : JPE, JPG, JPEG, BMP, ICO, PBM, PGM, PNM, PPM, PNG, BW, RGB,
	//!                     RGBA, SGI, TGA, TIF, TIFF, PCX, XPM, PSP, PSD, PIX, PXR,
	//!                     CUT, DCX, LIF
	//! Returns ATIP_FILE_DOESNT_EXIST if the file does not exist,
	//! ATIP_FILE_IO_ERROR if an error occurs when reading the file
	//! \param fileName Name of the file to load
	//! \return true if there was no error, else call atipGetError()
	bool LoadFromFile(const char * fileName);

	//! Save the image to a file
	//! Formats supported : JPE, JPG, JPEG, BMP, DDS, H, PBM, PGM, PPM, PNG, BW, RGB,
	//!                     RGBA, SGI, TGA, TIF, PCX, PSD
	//! ATIP_FILE_IO_ERROR if an error occurs when writing the file
	//! \param fileName Name of the file to write
	//! \return true if there was no error, else call atipGetError()
	bool SaveToFile(const char * fileName);

	//------------------------------------------------------------------------------------

	//! Get a pointer to the whole image
	//! \return A pointer to the data
	inline void * GetImage() const
		{	ATIP_ASSERT(_data);	return _data;	}

    //! Access to a row of the image
	//! \warning Use GetNumBytesPerPixel() to access the desired pixels
	//! \sa ATIPimage::GetNumBytesPerPixel()
	//! \param row The index of the row to access (0 <= row < height)
	//! \return A pointer to the row to access
	inline void * operator [] (unsigned int row) const
		{	ATIP_ASSERT(row < _height);	ATIP_ASSERT(_rows);	ATIP_ASSERT(_rows[row]);
			return _rows[row];
		}

	//! Get a pixel of the image
	//! \param row The row of the pixel (0 is the top of the image)
	//! \param col The column of the pixel 
	//! \return A pointer to the data of the pixel
	inline const void * GetPixel(unsigned int row, unsigned int col)
		{	ATIP_ASSERT((row < _height) && (col < _width));
			ATIP_ASSERT(_rows);	ATIP_ASSERT(_rows[row]);
			return _rows[row] + col * GetNumBytesPerPixel();
		}

	//! Set a pixel of the image
	//! \param row The row of the pixel (0 is the top of the image)
	//! \param col The column of the pixel 
	//! \param data Data of the pixel (size depending of the bitdepth)
	inline void SetPixel(unsigned int row, unsigned int col, const void * data)
		{	ATIP_ASSERT((row < _height) && (col < _width));
			ATIP_ASSERT(_rows);	ATIP_ASSERT(_rows[row]);	ATIP_ASSERT(data);
			memcpy(_rows[row] + col * GetNumBytesPerPixel(), data, GetNumBytesPerPixel());
		}

	//! Get the width of the image
	//! \return Width of the image in pixels
	inline unsigned int GetWidth() const
		{	ATIP_ASSERT(_width > 0);	return _width;	}

	//! Get the height of the image
	//! \return Height of the image in pixels
	inline unsigned int GetHeight() const
		{	ATIP_ASSERT(_height > 0);	return _height;	}

	//! Get the bitdepth of the image
	//! \return The bitdepth (1 to 8, 24 or 32)
	inline unsigned char GetBpp() const
		{	ATIP_ASSERT(((_bpp >= 1) && (_bpp <= 8)) || (_bpp == 24) || (_bpp == 32));
			return _bpp;
		}

	//! Get the number of bytes needed to store a pixel of the image.
	//! This method is used to access a pixel inside the image or inside a row.
	//! For example, to access the 6th pixel of the 3rd row :
	//!         image[2][5 * image.GetNumBytesPerPixel()].
	//! It is better to use GetPixel(2, 5)
	//! \return The number of bytes for a pixel from 1 to 4
	inline unsigned char GetNumBytesPerPixel() const
		{	ATIP_ASSERT((_nBytesPerPixel != 0) && (_nBytesPerPixel <= 4));
			return _nBytesPerPixel;
		}

	//! Does the palette have an alpha component ?
	//! \return true If the palette is coded in 32 bits instead of 24
	inline bool HasPaletteAlpha() const
		{	return _paletteAlpha;	}

	//------------------------------------------------------------------------------------

	//! Set the palette of the image (if bpp <= 8)
	//! Returns ATIP_INVALID_ARGUMENT if an argument is bad
	//! \param alpha True if the palette contains RGBA colors, else RGB colors are used
	//! \param palette Array of colors coded in 24 bits (RGB) or 32 bits (RGBA)
	//! \return true if there was no error, else call atipGetError()
	bool SetPalette(bool alpha, const void * palette);

	//! Get a pixel of a palettized image coded in RGB or RGBA
	//! \param row Row of the pixel (0 is the top of the image)
	//! \param col Column of the pixel 
	//! \return A pointer to the color of the pixel
	inline const void * GetPalPixel(unsigned int row, unsigned int col) const
		{	ATIP_ASSERT((row < _height) && (col < _width));
			ATIP_ASSERT(_rows);	ATIP_ASSERT(_rows[row]);
			return _palette + _rows[row][col * GetNumBytesPerPixel()]
									* (_paletteAlpha ? 4 : 3);
		}

	//------------------------------------------------------------------------------------

	//! Convert a 24 or 32 bits image to a grayscale one.
	//! If the image is a palettized one, no conversion occurs
	void ConvertToGreyscale();

	//! Convert a 32 bits, or a palettized image to a 24 bits one.
	//! If the image is already coded in 24 bits, nothing occurs
	void ConvertToRGB();

	//------------------------------------------------------------------------------------

	//! Convert an external image coded with floats to a greyscale image
	//! \warning The image must have the same dimensions than the external one
	//! \note If mim = max = 0.0f, the method computes these values
	//! \param image The image to convert
	//! \param max Maximum value of the floats of the image
	//! \param min Minimum value of the floats of the image
	//! \return true if there was no error, else call atipGetError()
	bool ConvertFromFloats(float * image, float max = 0.0f, float min = 0.0f);

	//------------------------------------------------------------------------------------

private:

	unsigned int _width;			//!< Width of the image in pixels
	unsigned int _height;			//!< Height of the image in pixels
	unsigned char _bpp;				//!< Bitdepth of the image (1 to 8, 24 or 32)
	unsigned char _nBytesPerPixel;	//!< Number of needed bytes to store a pixel

	unsigned char * _data;			//!< Data of the image (from the top to the bottom)
	unsigned char ** _rows;			//!< Array of pointers to the rows of the image

	bool _paletteAlpha;				//!< True if the palette is coded with RGBA colors
									//!< instead of RGB
	unsigned char * _palette;		//!< Data of the palette

	static unsigned int nImLoaded;	//!< Number of loaded images
};

#endif	// _ATIP_IMAGE_H_