/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_options.h
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Options for the "Automatic Tour Into the Picture" library (ATIP)


#ifndef _ATIP_OPTIONS_H_
#define _ATIP_OPTIONS_H_


//@{
//! Filtering methods constants for the edges detection

//! Deriche laplacian (Deriche laplacian filter, zero-crossing pixels detection,
//! detection ofthe maxima of the gradients on these pixels) (default method)
#define ATIP_FILTERING_DERICHE_LAPLACIAN			0

//! Simple laplacian (3x3 convolution mask, zero-crossing pixels detection,
//! detection ofthe maxima of the gradients on these pixels)
#define ATIP_FILTERING_SIMPLE_LAPLACIAN				1

//! Maxima of the gradients (Frei-Chen convolution mask to compute the gradients,
//! and maxima detection)
#define ATIP_FILTERING_FREICHEN_MAXIMA				2

//@}

//! Set the current filtering method for the edges detection
//! \param method One of the ATIP_FILTERING_xxx constant
void atipSetFilteringMethod(unsigned char method);

//! Get the current filtering method for the edges detection
//! \return One of the ATIP_FILTERING_xxx constant
unsigned char atipGetFilteringMethod();

//----------------------------------------------------------------------------------------

//@{
//! Thresholding methods constants for the gradients

//! Hysteresis thresholding of the gradients (uses the neighbors of the gradients)
//! (default method)
#define ATIP_THRESHOLDING_HYSTERESIS				0

//! Simple thresholding of the gradients (with a single level)
#define ATIP_THRESHOLDING_SIMPLE					1

//@}

//! Set the current thresholding method for the gradients
//! \param method One of the ATIP_THRESHOLDING_xxx constant
void atipSetThresholdingMethod(unsigned char method);

//! Get the current thresholding method for the gradients
//! \return One of the ATIP_THRESHOLDING_xxx constant
unsigned char atipGetThresholdingMethod();

//----------------------------------------------------------------------------------------

//! Set the current state for the neighbors check.
//! When hysteresis thresholding is done with a laplacian method, the neighbors of
//! the zero-crossing laplacians can be used to have a more precise edge detection
//! (but the computing time is longer)
//! \param state True to activate the neighbors check
void atipSetNeighborsCheckState(bool state);

//! Get the curent state of the neighbors check
//! \sa atipSetNeighborsCheck
//! \return True if the neighbors check is activated
bool atipGetNeighborsCheckState();

//----------------------------------------------------------------------------------------

//! Set the alpha parameter for the Deriche laplacian method
//! \param alpha Influence of the low-pass filter (positive, better if <= 1.0f)
//!              If alpha is small, the result is finer but more biased
void atipSetDericheAlpha(float alpha);

//! Get the alpha parameter for the Deriche laplacin method
//! \sa atipSetDericheAlpha
//! \return Deriche alpha parameter (>= 0.0f)
float atipGetDericheAlpha();

//----------------------------------------------------------------------------------------

//! Set the low level for the thresholding of the gradients (hysteresis method)
//! \param low Low level of the hysteresis thresholding (0.0f -> high level)
void atipSetThresholdingLowLevel(float low);

//! Get the low level for the thresholding of the gradients (hysteresis method)
//! \return Low level of the hysteresis thresholding (0.0f -> high level)
float atipGetThresholdingLowLevel();

//----------------------------------------------------------------------------------------

//! Set the high level for the thresholding of the gradients (hysteresis method),
//! or the single level for the simple thresholding method
//! \param high High level of the hysteresis thresholding
//!             (low level -> ATIP_MAX_THRESHOLD), or the single level
//!             (0.0f -> ATIP_MAX_THRESHOLD)
void atipSetThresholdingHighLevel(float high);

//! Get the high level for the thresholding of the gradients (hysteresis method),
//! or the single level for the simple thresholding method
//! \return High level of the hysteresis thresholding, or single level of the
//!         simple thresholding (0.0f -> ATIP_MAX_THRESHOLD)
float atipGetThresholdingHighLevel();

//----------------------------------------------------------------------------------------

//! Set the angle resolution (theta) of the first Hough transform.
//! The higher, the longer the computation is, with more precision.
//! If the parameter is too high, false lines can be detected
//! \param resolution Resolution of the theta parameter, 0 to fix it to the width
//!                   of the image
void atipSetHoughAngleResolution(unsigned int resolution);

//! Get the angle resolution (theta) of the first Hough transform
//! \return The angle resolution, 0 if it is fixed to the width of the image
unsigned int atipGetHoughAngleResolution();

//----------------------------------------------------------------------------------------

//! Set the level for the thresholding of the high-pass filtered polar Hough transform
//! \param level Level of thresholding (from 0.0f excluded to 1.0f excluded)
void atipSetPolarHoughThresholdingLevel(float level);

//! Get the level for the thresholding of the high-pass filtered polar Hough transform
//! \return The level of thresholding (from 0.0f excluded to 1.0f excluded)
float atipGetPolarHoughThresholdingLevel();

//----------------------------------------------------------------------------------------

//! Set the number of cells around the gaussian hemisphere (phi parameter)
//! for the Hough transform, to detect the vanishing points
//! \param resolution Number of cells
void atipSetGaussAzimuthResolution(unsigned int resolution);

//! Get the number of cells around the gaussian hemisphere (phi parameter)
//! for the Hough transform, to detect the vanishing points
//! \return Number of cells
unsigned int atipGetGaussAzimuthResolution();

//----------------------------------------------------------------------------------------

//! Set the number of cells along the gaussian hemisphere meridians (theta parameter)
//! for the Hough transform, to detect the vanishing points
//! \param resolution Number of cells
void atipSetGaussElevationResolution(unsigned int resolution);

//! Get the number of cells along the gaussian hemisphere meridians (theta parameter)
//! for the Hough transform, to detect the vanishing points
//! \return Number of cells
unsigned int atipGetGaussElevationResolution();

//----------------------------------------------------------------------------------------

//! Set the "Use main lines contributions" state
//! \param state True in order to use the number of edges points on the main lines
//!              as a contribution for the gaussian hemisphere Hough transform
void atipSetLinesContributionState(bool state);

//! Get the "Use main lines contributions" state
//! \return True if the main lines contributions are used
bool atipGetLinesContributionState();

//----------------------------------------------------------------------------------------

//! Set the gaussian hemisphere low-pass filter window size
//! \param size Size of the window (3, 5, 7, 9, ...)
void atipSetGaussWindowSize(unsigned int size);

//! Get the gaussian hemisphere low-pass filter window size
//! \return Size of the window (3, 5, 7, 9, ...)
unsigned int atipGetGaussWindowSize();

//----------------------------------------------------------------------------------------

//! Set the maximum distance of a line from a finite vanishing point to consider it
//! as a contributor (portion of min(width, height))
//! \param maxDist The maximum distance (0.0 to 1.0 excluded)
void atipSetMaxLineDistance(float maxDist);

//! Get the maximum distance of a line from a finite vanishing point to consider it
//! as a contributor (portion of min(width, height))
//! \return The maximum distance (0.0 to 1.0 excluded)
float atipGetMaxLineDistance();

//----------------------------------------------------------------------------------------

//! Set the theta spherical parameter for which a vanishing point is considered
//! as infinite if its angle is greater than the threshold
//! \param angle Angle in radians (less or equal to 1.57f, and positive)
void atipSetInfiniteAngle(float angle);

//! Get the theta spherical parameter for which a vanishing point is considered
//! as infinite if its angle is greater than the threshold
//! \return Angle in radians (less or equal to 1.57f, and positive)
float atipGetInfiniteAngle();

//----------------------------------------------------------------------------------------

//! Set the maximum angle of a line from an infinite vanishing point direction
//! to consider it as a contributor
//! \param angle Maximum angle in degrees (between 0 and 90 excluded)
void atipSetMaxLineAngle(unsigned int angle);

//! Get the maximum angle of a line from an infinite vanishing point direction
//! to consider it as a contributor
//! \return Maximum angle in degrees (between 0 and 90 excluded)
unsigned int atipGetMaxLineAngle();

//----------------------------------------------------------------------------------------

//@{
//! Computing speeds for the vanishing point detection

//! Normal speed, full precision
#define ATIP_COMPSPEED_NORMAL						0

//! Fast speed, less precision
#define ATIP_COMPSPEED_FAST							1

//! Very fast speed, the worst precision
#define ATIP_COMPSPEED_VERYFAST						2

//@}

//! Set the computing speed for the vanishing point detection
//! (the faster, the less precise is the detection)
//! \param speed One of the ATIP_COMPSPEED_xxx constant
void atipSetComputingSpeed(unsigned int speed);

//! Get the computing speed for the vanishing point detection
//! (the faster, the less precise is the detection)
//! \return One of the ATIP_COMPSPEED_xxx constant
unsigned int atipGetComputingSpeed();

//----------------------------------------------------------------------------------------

//! Set the field-of-view (FOV) for a perspective with one finite vanishing point
//! \param angle Vertical field-of-view (in degrees)
//!        (from the bottom to the top of the image)
void atipSetFOV(float angle);

//! Get the field-of-view (FOV) for a perspective with one finite vanishing point
//! \return Vertical field-of-view (in degrees) (from the bottom to the top of the image)
float atipGetFOV();

//----------------------------------------------------------------------------------------

//! Set the maximum texture size
//! \param maxTexSize Maximum texture size, power of 2 from 64 to 4096
void atipSetMaximumTextureSize(unsigned int maxTexSize);

//! Get the maximum texture size
//! \return Maximum texture size, power of 2 from 64 to 4096
unsigned int atipGetMaximumTextureSize();


#endif	// _ATIP_OPTIONS_H_