/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_tools.h
//! \author	BOULANGER Kevin
//! \date	26th June 2004
//! \brief	Main tools for the "Automatic Tour Into the Picture" library (ATIP)


#ifndef _ATIP_TOOLS_H_
#define _ATIP_TOOLS_H_


//! Definition of the error type
typedef unsigned int ATIPerror;

//@{
//! Error constants
#define ATIP_NO_ERROR							0
#define ATIP_INVALID_ARGUMENT					5
#define ATIP_INVALID_OPERATION					6
#define ATIP_FILE_READ_ERROR					10
#define ATIP_FILE_WRITE_ERROR					11
#define ATIP_BAD_FILE_FORMAT					15
#define ATIP_NOT_ENOUGH_EDGE_LINES				20
#define ATIP_NO_FINITE_VAN_POINT				22
#define ATIP_NO_THREE_DIFFERENT_VAN_POINTS		23
//@}

//! Return the last error code
//! \return An error code (ATIP_NO_ERROR if no error occured)
ATIPerror atipGetError();

//! Convert an error code to a string
//! \param error The code of the error (returned by atipGetError())
//! \param errorMessage An allocated array of chars which will contain the message,
//!                     "UNKNOWN ERROR" if the error code is invalid
void atipErrorMessage(ATIPerror error, char * errorMessage);


//----------------------------------------------------------------------------------------

//! Set the intermediate results flags
//! (to obtain intermediate images with the atipIntermediateResult callback)
//! \param intermResults Logical OR with the ATIP_INTRES_xxx flags.
void atipSetIntermediateResultsFlags(unsigned int intermResults);

//! Get the intermediate results flags
//! \return Logical OR with the ATIP_INTRES_xxx flags
unsigned int atipGetIntermediateResultsFlags();


//! Codes for the intermediate results flags

//! Original color image
#define ATIP_INTRES_ORIGINAL				0x00000001

//! Grayscale conversion after image loading
#define ATIP_INTRES_GRAYSCALE				0x00000002

//! After Deriche or simple filtering and detection of the zero-crossing laplacians
#define ATIP_INTRES_LAPLACIANS				0x00000004

//! Gradients obtained after Frei-Chen filtering
#define ATIP_INTRES_GRADIENTS				0x00000008

//! After thresholding
#define ATIP_INTRES_THRESHOLD				0x00000010

//! Polar Hough transform (to detect the main lines)
#define ATIP_INTRES_POLARHOUGH				0x00000020

//! After high-pass filtering of the polar Hough transform
#define ATIP_INTRES_HIGHPASSHOUGH			0x00000040

//! Lines detected by the first Hough transform
#define ATIP_INTRES_MAINLINES				0x00000080

//! Gaussian hemisphere Hough transform (to find the intersections of the main lines)
#define ATIP_INTRES_GAUSSHOUGH				0x00000100

//! Detected vanishing points (finite and infinite)
#define ATIP_INTRES_VANPOINTS				0x00000200

//! TIP textures
#define ATIP_INTRES_TEXTURES				0x00000400

//----------------------------------------------------------------------------------------

// Callbacks for the analysis of the process

class ATIPimage;

//! Definition of the intermediate result function pointer
//! \param image The intermediate image
//! \param title Title of the result
//! \return True if there was no error
typedef bool (* ATIPintermediateResultProc) (const ATIPimage & image,
											 const char * title);
//! Set the intermediate result callback.
//! The callback is called when an intermediate result is obtained
//! \param proc Pointer to a function which manages the intermediate results
void atipSetIntermediateResultCallback(ATIPintermediateResultProc proc);

//! Definition of the percentage function pointer
//! \param percentage Current percentage of the processing
typedef void (* ATIPpercentageProc) (float percentage);

//! Set the percentage callback.
//! This callback is called each time a percentage of the processing is known.
//! It can be used for a progression bar.
//! \param proc Pointer to a function which manages the percentage values
void atipSetPercentageCallback(ATIPpercentageProc proc);


//----------------------------------------------------------------------------------------

#ifdef _DEBUG

//! Do an assertion error if the given expression is not true (in debug mode).
//! If an assertion error occurs, the user callback set by atipSetAssertionCallback()
//! is called
//! \param expr Expression which is evaluated (it should be equals to true)
#define ATIP_ASSERT(expr)	\
		{	if (!(expr)) AssertionError(#expr, __FILE__, __LINE__);	}

//! Do an assertion error.
//! The user callback set by atipSetAssertionCallback() is called if available
//! \param expr Expression which has been evaluated to false
//! \param sourceFile Name of the source file in which the error occured
//! \param sourceLine Line of the expression inside the source file in which the error
//!                   occured
void AssertionError(const char * expr, const char * sourceFile, int sourceLine);

#else

//! Void definition of the assertion macro for the release mode
#define ATIP_ASSERT(expr)

#endif	// _DEBUG


//----------------------------------------------------------------------------------------

//! Out of memory error.
//! The user callback set by atipSetOutOfMemoryCallback() is called if available
void OutOfMemoryError();

//! Allocation of an object
//! \warning The pointers MUST always be initialised with 0 at the beginning of
//!          the program execution
//! \note If the macro is called with an already allocated object,
//!       it is firstly deallocated
//! \param t Type of the object to allocate
//! \param ptr The pointer which will point to the object (t* type)
#define ATIP_NEW(t, ptr)			{												\
										if ((ptr)) delete (ptr);					\
										ptr = new t;								\
										if (!(ptr)) OutOfMemoryError();				\
									}

//! Deallocate an object
//! \param ptr A pointer to an object (can be 0)
#define ATIP_DELETE(ptr)			{												\
										if ((ptr)) delete (ptr);					\
										ptr = 0;									\
									}

//! Allocation of an array of objects
//! \warning The pointers MUST always be initialised with 0 at the beginning of
//!          the program execution
//! \note If the macro is called with an already allocated array,
//!       it is firstly deallocated
//! \param t Type of the object to allocate
//! \param ptr The pointer which will point to the object (t* type)
//! \param n Number of items inside the array
#define ATIP_NEW_ARRAY(t, ptr, n)	{												\
										if ((ptr)) delete[] (ptr);					\
										ptr = new t [n];							\
										if (!(ptr)) OutOfMemoryError();				\
									}

//! Deallocate an array of objects
//! \param ptr A pointer to an array of objects (can be 0)
#define ATIP_DELETE_ARRAY(ptr)		{												\
										if ((ptr)) delete[] (ptr);					\
										ptr = 0;									\
									}


#endif	// _ATIP_TOOLS_H_