/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_private_tools.h
//! \author	BOULANGER Kevin
//! \date	27th June 2004
//! \brief	Private tools reserved for the ATIP implementation


#ifndef _ATIP_PRIVATE_TOOLS_H_
#define _ATIP_PRIVATE_TOOLS_H_

#include <atip.h>


//! Set the last error which occured (got by a atipGetError() call)
//! \param error Code of the error
void atipSetError(ATIPerror error);

//! Get the last error
//! \return Code of the last error
ATIPerror atipGetLastError();

//----------------------------------------------------------------------------------------

//! Set the intermediate results flags
//! \param intermResults Logical OR with the ATIP_INTRES_xxx flags.
void atipSetIntermResultsFlags(unsigned int intermResults);

//! Test the presence of a given intermediate result flag
//! \param intermResult ATIP_INTRES_xxx flag
//! \return true if the desired intermediate result must be given
//!         (through a callback)
bool atipIsIntermResultActive(unsigned int intermResult);

//! Get the intermediate results flags
//! \return Logical OR with the ATIP_INTRES_xxx flags
unsigned int atipGetIntermResultsFlags();

//! Set the intermediate result callback.
//! The callback is called when an intermediate result is obtained
//! \param proc Pointer to a function whichi manages the intermediate results
void atipSetIntermResultCallback(ATIPintermediateResultProc proc);

//! Call the intermediate result callback
//! \param image Intermediate image
//! \param title Title of the result
bool atipCallIntermResultCallback(const ATIPimage & image, const char * title);

//----------------------------------------------------------------------------------------

//! Set the percentage callback.
//! This callback is called each time a percentage of the processing is known.
//! It can be used for a progression bar.
//! \param proc Pointer to a function which manages the percentage values
void atipSetPercCallback(ATIPpercentageProc proc);

//! Call the percentage callback
//! \param percentage Current percentage
void atipCallPercentageCallback(float percentage);

//----------------------------------------------------------------------------------------

//! Set the polygons editor callback.
//! This callback is called when the polygons editor is needed
//! \param proc Pointer to a function which opens a polygons editor
void atipSetPolyEdCallback(ATIPpolygonsEditorProc proc);

//! Call the polygons editor callback
//! \param atipData Reference to the data of the TIP
//! \param origImage Original image
//! \param nFiniteVanPoints Number of finite vanishing points
//! \param nInfiniteVanPoints Number of infinite vanishing points
//! \return true if there was no error, else call atipGetError()
bool atipCallPolygonsEditorCallback(ATIPdata & atipData, ATIPimage & origImage,
									unsigned int nFiniteVanPoints,
									unsigned int nInfiniteVanPoints);

//----------------------------------------------------------------------------------------

//@{
//! Useful constants
#define M_2PI		6.28318530718			// 2 * PI
#define M_0_5PI		1.5707963268			// PI / 2
//@}

//@{
//! Useful macros
#define MIN(a, b)		((a) > (b) ? (b) : (a))
#define MAX(a, b)		((a) > (b) ? (a) : (b))
//@}

//----------------------------------------------------------------------------------------

//! Test if a number is a power of 2
//! \param n The argument to test
inline bool atipIsPowerOf2(unsigned int n)
{
	if (n == 0) return false;
	while ((n & 1) == 0) n >>= 1;
	n >>= 1;
	return n == 0;
}


//! Compute the next power of 2
//! \param n The argument for which the next power of 2 must be found
//! \return The next power of 2 (equality if the argument is already a power of 2)
inline unsigned int atipNextPowerOf2(unsigned int n)
{
	ATIP_ASSERT(n >= 1);
	unsigned int pow = 0;
	while ((unsigned int)(1 << pow) < n) pow++;
	return 1 << pow;
}


//! Logarithm in base 2 of a power of 2
//! \param n Original number (> 0)
//! \return Logarithm in base 2 of n
inline unsigned int atipLog2(unsigned int n)
{
	ATIP_ASSERT(n > 0);
	unsigned int i = 0;
	while (n > 1)
	{
		i++;
		n >>= 1;
	}
	return i;
}

//----------------------------------------------------------------------------------------

//! Swap two float numbers
//! \param x1 The first variable to swap
//! \param x2 The second variable to swap
inline void swap(float & x1, float & x2)
{
	float t = x1;
	x1 = x2;
	x2 = t;
}


#endif	// _ATIP_H_