/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_filter.h
//! \author	BOULANGER Kevin
//! \date	29th June 2004
//! \brief	Image filtering functions


#ifndef _ATIP_FILTER_H_
#define _ATIP_FILTER_H_


//! Convert a 24 bits (RGB) or 32 bits (RGBA) image into a grayscale one
//! \param grayImage Grayscale image, coded in 8 bits
//! \param colorImage Source color image
//! \param width Width in pixel of both images
//! \param height Height in pixel of both images
//! \param bpp Bits per pixel of the source image (24 or 32)
//! \return True if there was no error, else call atipGetError()
bool atipConvertToGrayscale(unsigned char * grayImage, const unsigned char * colorImage,
							unsigned int width, unsigned int height, unsigned char bpp);

//! Convert an image coded in 8 bits to an image of floats
//! \param dstImage Resulting image after conversion
//! \param srcImage Source image before conversion
//! \param width Width in pixel of both images
//! \param height Height in pixel of both images
//! \return True if there was no error, else call atipGetError()
bool atipConvertImageToFloats(float * dstImage, const unsigned char * srcImage,
							  unsigned int width, unsigned int height);

//----------------------------------------------------------------------------------------

//! Reduce the size of an image coded in 8 bits by a factor of 2
//! \param smallImage Resulting image after the reduction (coded in 8 bits)
//!                   (allocated by the function)
//! \param smallWidth Resulting width of the small image (in pixels)
//! \param smallHeight Resulting height of the small image (in pixels)
//! \param bigImage Source image (coded in 8 bits)
//! \param bigWidth Resulting width of the small image (in pixels)
//! \param bigHeight Resulting height of the small image (in pixels)
//! \return True if there was no error, else call atipGetError()
bool atipReduceImage2_8(unsigned char * & smallImage,
						unsigned int & smallWidth, unsigned int & smallHeight,
						const unsigned char * bigImage,
						unsigned int bigWidth, unsigned int bigHeight);

//! Reduce the size of an image coded in 24 bits by a factor of 2
//! \param smallImage Resulting image after the reduction (coded in 24 bits)
//!                   (allocated by the function)
//! \param smallWidth Resulting width of the small image (in pixels)
//! \param smallHeight Resulting height of the small image (in pixels)
//! \param bigImage Source image (coded in 24 bits)
//! \param bigWidth Resulting width of the small image (in pixels)
//! \param bigHeight Resulting height of the small image (in pixels)
//! \return True if there was no error, else call atipGetError()
bool atipReduceImage2_24(unsigned char * & smallImage,
						 unsigned int & smallWidth, unsigned int & smallHeight,
						 const unsigned char * bigImage,
						 unsigned int bigWidth, unsigned int bigHeight);

//----------------------------------------------------------------------------------------

//! Apply a Deriche laplacian filter to a luminance image coded with floats
//! and look for the zero-crossing laplacians to find the edges
//! \param dstImage Destination image coded with 1 bit (stored inside a byte),
//!                 pixels with value 1 are edges
//! \param srcImage Source image, before filtering
//! \param width Width in pixel of both images
//! \param height Height in pixel of both images
//! \param alpha Influence of the low-pass filter (positive, better if <= 1.0f)
//!              If alpha is small, the result is finer but more biased
//! \return True if there was no error, else call atipGetError()
bool atipDericheLaplacians(unsigned char * dstImage, const float * srcImage,
						   unsigned int width, unsigned int height, float alpha);

//! Apply a simple laplacian filter to a luminance image coded with 8 bits
//! and look for the zero-crossing laplacians to find the edges
//! \param dstImage Destination image coded with 1 bit (stored inside a byte),
//!                 pixels with value 1 are edges
//! \param srcImage Source image, before filtering
//! \param width Width in pixel of both images
//! \param height Height in pixel of both images
//! \return True if there was no error, else call atipGetError()
bool atipSimpleLaplacians(unsigned char * dstImage, const unsigned char * srcImage,
						  unsigned int width, unsigned int height);

//! Compute the images of gradient of an original image using the 3x3 Frei-Chen filter
//! \note The gradients are nul for the first and last rows and columns of gradX and gradY
//! \param gradX Resulting image containing the gradients along the X axis
//! \param gradY Resulting image containing the gradients along the Y axis
//! \param srcImage Source image, before filtering
//! \param width Width in pixel of both images
//! \param height Height in pixel of both images
//! \return True if there was no error, else call atipGetError()
bool atipFreiChenFilter(float * gradX, float * gradY, const unsigned char * srcImage,
						unsigned int width, unsigned int height);

//----------------------------------------------------------------------------------------

//! Maximum threshold
#define ATIP_MAX_THRESHOLD		360.624f		// = sqrt(255^2 + 255^2)

//! Extract the pixels of an image where gradients are greater than a single threshold
//! and where the mask pixels are 1's
//! \param edges Resulting binary image, same format as mask, allocated by the function
//! \param edgePoints Array of size 2 * width * height integers which will contain
//!                   nEdgePoints groups (x,y) which are the edge points coordinates
//! \param nEdgePoints Number of detected edge points (modified by the function)
//! \param mask 1 bit image where 1's are where pixels must be tested (coded with 8 bits)
//! \param srcImage Original image (coded with floats)
//! \param width Width in pixel of mask and srcImage
//! \param height Height in pixel of mask and srcImage
//! \param threshold Threshold of the test (0.0f -> ATIP_MAX_THRESHOLD)
//! \return True if there was no error, else call atipGetError()
bool atipSimpleThreshold(unsigned char * & edges,
						 int * edgePoints, unsigned int & nEdgePoints,
						 const unsigned char * mask, const float * srcImage,
						 unsigned int width, unsigned int height, float threshold);

//! Extract the pixels of an image where gradients norm are greater than the
//! thresholding level
//! \param edges Resulting binary image, same format as mask, allocated by the function
//! \param edgePoints Array of size 2 * width * height integers which will contain
//!                   nEdgePoints groups (x,y) which are the edge points coordinates
//! \param nEdgePoints Number of detected edge points (modified by the function)
//! \param gradX Gradients along the X axis of the image
//! \param gradY Gradients along the Y axis of the image
//! \param width Width in pixel of mask and srcImage
//! \param height Height in pixel of mask and srcImage
//! \param threshold Threshold level (0.0f -> ATIP_MAX_THRESHOLD)
//! \return True if there was no error, else call atipGetError()
bool atipSimpleThreshold(unsigned char * & edges,
						 int * edgePoints, unsigned int & nEdgePoints,
						 const float * gradX, const float * gradY,
						 unsigned int width, unsigned int height, float threshold);

//! Extract the pixels of an image where the mask pixels are 1's, and where gradients norm
//! are greater than the high thresholding level, or between the low and the high ones
//! if ALL neighbors have a gradient greater than the high level
//! \param edges Resulting binary image, same format as mask, allocated by the function
//! \param edgePoints Array of size 2 * width * height integer which will contain
//!                   nEdgePoints groups (x,y) which are the edge points coordinates
//! \param nEdgePoints Number of detected edge points (modified by the function)
//! \param mask 1 bit image where 1's are where pixels must be tested (coded with 8 bits)
//! \param srcImage Original image (coded with floats)
//! \param width Width in pixel of mask and srcImage
//! \param height Height in pixel of mask and srcImage
//! \param lowThreshold Low threshold level (0.0f -> highThreshold)
//! \param highThreshold High threshold level (lowThreshold -> ATIP_MAX_THRESHOLD)
//! \return True if there was no error, else call atipGetError()
bool atipHysteresisThresholdNeighbor(unsigned char * & edges,
									 int * edgePoints, unsigned int & nEdgePoints,
									 const unsigned char * mask, const float * srcImage,
									 unsigned int width, unsigned int height,
			                         float lowThreshold, float highThreshold);

//! Extract the pixels of an image where the mask pixels are 1's, and where gradients norm
//! are greater than the high thresholding level, or between the low and the high ones
//! if neighbors ON THE MASK have a gradient greater than the high level
//! \param edges Resulting binary image, same format as mask, allocated by the function
//! \param edgePoints Array of size 2 * width * height integer which will contain
//!                   nEdgePoints groups (x,y) which are the edge points coordinates
//! \param nEdgePoints Number of detected edge points (modified by the function)
//! \param mask 1 bit image where 1's are where pixels must be tested (coded with 8 bits)
//! \param srcImage Original image (coded with floats)
//! \param width Width in pixel of mask and srcImage
//! \param height Height in pixel of mask and srcImage
//! \param lowThreshold Low threshold level (0.0f -> highThreshold)
//! \param highThreshold High threshold level (lowThreshold -> ATIP_MAX_THRESHOLD)
//! \return True if there was no error, else call atipGetError()
bool atipHysteresisThreshold(unsigned char * & edges,
							 int * edgePoints, unsigned int & nEdgePoints,
							 const unsigned char * mask, const float * srcImage,
							 unsigned int width, unsigned int height,
							 float lowThreshold, float highThreshold);

//! Extract the pixels of an image where gradients norm are greater than the high
//! thresholding level, or between the low and the high ones if neighbors have a
//! gradient greater than the high level
//! \param edges Resulting binary image, same format as mask, allocated by the function
//! \param edgePoints Array of size 2 * width * height integer which will contain
//!                   nEdgePoints groups (x,y) which are the edge points coordinates
//! \param nEdgePoints Number of detected edge points (modified by the function)
//! \param gradX Gradients along the X axis of the image
//! \param gradY Gradients along the Y axis of the image
//! \param width Width in pixel of mask and srcImage
//! \param height Height in pixel of mask and srcImage
//! \param lowThreshold Low threshold level (0.0f -> highThreshold)
//! \param highThreshold High threshold level (lowThreshold -> ATIP_MAX_THRESHOLD)
//! \return True if there was no error, else call atipGetError()
bool atipHysteresisThreshold(unsigned char * & edges,
							 int * edgePoints, unsigned int & nEdgePoints,
							 const float * gradX, const float * gradY,
							 unsigned int width, unsigned int height,
							 float lowThreshold, float highThreshold);

//----------------------------------------------------------------------------------------

//! Apply a mean filter to a luminance image coded with 8 bits
//! \note The luminance is nul near the sides, the width of these rows and columns
//!       is int(windowSize / 2)
//! \param dstImage Destination image, after filtering
//! \param srcImage Source image, before filtering
//! \param width Width in pixels of both images
//! \param height Height in pixels of both images
//! \param windowSize An odd number, 7 for example for a 7x7 window
//! \return True if there was no error, else call atipGetError()
bool atipMeanFilter8(unsigned char * dstImage, const unsigned char * srcImage,
					 unsigned int width, unsigned int height,
					 unsigned int windowSize);

//! Apply a mean filter to a luminance image coded with 32 bits
//! \note The luminance is nul near the sides, the width of these rows and columns
//!       is int(windowSize / 2)
//! \param dstImage Destination image, after filtering
//! \param srcImage Source image, before filtering
//! \param width Width in pixels of both images
//! \param height Height in pixels of both images
//! \param windowSize An odd number, 7 for example for a 7x7 window
//! \return True if there was no error, else call atipGetError()
bool atipMeanFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					  unsigned int width, unsigned int height,
					  unsigned int windowSize);

//! Apply a mean filter to a luminance image coded with 32 bits (with maximum detection)
//! \note The luminance is nul near the sides, the width of these rows and columns
//!       is int(windowSize / 2)
//! \param dstImage Destination image, after filtering
//! \param srcImage Source image, before filtering
//! \param width Width in pixels of both images
//! \param height Height in pixels of both images
//! \param windowSize An odd number, 7 for example for a 7x7 window
//! \param max Resulting maximum value of the filtered image
//! \return True if there was no error, else call atipGetError()
bool atipMeanFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					  unsigned int width, unsigned int height,
					  unsigned int windowSize, unsigned int & max);

//! Apply a filter that computes the sum of the neighbor pixels in a luminance image
//! coded with 32 bits (with maximum detection)
//! \note The luminance is nul near the sides, the width of these rows and columns
//!       is int(windowSize / 2)
//! \param dstImage Destination image, after filtering
//! \param srcImage Source image, before filtering
//! \param width Width in pixels of both images
//! \param height Height in pixels of both images
//! \param windowSize An odd number, 7 for example for a 7x7 window
//! \param max Resulting maximum value of the filtered image
//! \param xMax X coordinate of the maximum
//! \param yMax Y coordinate of the maximum
//! \return True if there was no error, else call atipGetError()
bool atipSumFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					 unsigned int width, unsigned int height,
					 unsigned int windowSize, unsigned int & max,
					 unsigned int & xMax, unsigned int & yMax);

//! Apply a high-pass filter to a luminance image coded with 32 bits
//! \note The luminance is nul in the first and last lines and columns
//! \param dstImage Destination image, after filtering
//! \param srcImage Source image, before filtering
//! \param width Width in pixels of both images
//! \param height Height in pixels of both images
//! \param max Resulting maximum value of the filtered image
//! \return True if there was no error, else call atipGetError()
bool atipHighPassFilter32(unsigned int * dstImage, const unsigned int * srcImage,
						  unsigned int width, unsigned int height, unsigned int & max);

#endif	// _ATIP_H_