/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_gaussian_hemisphere.h
//! \author	BOULANGER Kevin
//! \date	20th July 2004
//! \brief	Hough transform on a gaussian hemisphere


#ifndef _ATIP_GAUSSIAN_HEMISPHERE_H_
#define _ATIP_GAUSSIAN_HEMISPHERE_H_


//! Given a list of lines, compute the Hough transform on the gaussian hemisphere
//! \param hough Resulting Hough transform, array of
//!              (nTheta + windowSize - 1) * (nPhi + windowSize - 1) accumulation values,
//!              allocated by the function (phi for the lines)
//! \param maxHough Resulting maximum accumulation value of the Hough transform
//! \param lines nLines quadruplets (x0, y0, dx, dy), one for each line
//! \param lineCoeffs Array of coefficients, one per line, to use their contributions,
//!                   0 if all lines must have a coefficient of 1
//! \param nLines Number of input lines
//! \param nTheta Number of samples for the theta parameter (elevation)
//! \param nPhi Number of samples for the phi parameter (azimuth)
//! \param windowSize Size of the window used for the maxima detection (3, 5, 7, ...)
//! \param width Width of the original image in pixels
//! \param height Height of the original image in pixels
//! \return True if there was no error, else call atipGetError()
bool atipGaussHough(unsigned int * & hough, unsigned int & maxHough,
					const float * lines, const unsigned int * lineCoeffs,
					unsigned int nLines, unsigned int nTheta, unsigned int nPhi, 
					unsigned int windowSize, unsigned int width, unsigned int height);

//! Look for three orthogonal vanishing points
//! inside the gaussian hemisphere Hough transform
//! \param vanPoints Resulting array of 3 quadruplets (x, y, phi, theta), one per
//!                  vanishing point, allocated by the user.
//!                  If the point is finite, (x, y) are the coordinates of the point
//!                  inside the image ; if the point is infinite, it is the unit direction
//!                  vector. (phi, theta) are the gaussian hemisphere coordinates
//! \param infiniteVanPoints Array of 3 booleans which are true if the corresponding
//!                          vanishing points are infinite
//! \param lineClusters Array of 4 user allocated pointers to arrays containing
//!                     quadruplets (x0, y0, dx, dy) representing the lines which
//!                     contributed to detect the vanishing points. The number of lines
//!                     per cluster is nLinesPercluster[0..3]. The 4 contained arrays are
//!                     allocated by the function. The first 3 arrays correspond to the
//!                     three vanishing points. The last one is for the remaining lines
//! \param nLinesPerCluster Number of lines which contributed to detect the vanishing points,
//!                         user allocated array of 4 unsigned integers
//! \param hough Gaussian hemisphere Hough transform, array of
//!              (nTheta + windowSize - 1) * (nPhi + windowSize - 1) accumulation values,
//!              modified by the function
//! \param lines nLines quadruplets (x0, y0, dx, dy), one for each source line
//! \param lineCoeffs Array of coefficients, one per line, to use their contributions,
//!                   0 if all lines must have a coefficient of 1
//! \param nLines Number of input lines
//! \param nTheta Number of samples for the theta parameter (elevation)
//! \param nPhi Number of samples for the phi parameter (azimuth)
//! \param windowSize Size of the window used for the maxima detection (3, 5, 7, ...)
//! \param width Width of the original image in pixels
//! \param height Height of the original image in pixels
//! \param maxLineDistance Maximum distance from a point to a line to consider the line
//!                        as a member of the cluster associated with the point
//!                        (portion of min(width, height))
//! \param thetaInfinite Minimum theta value on the gaussian hemisphere for which a point
//!                      is considered as an infinite one
//! \param maxLineAngle Maximum angle from a direction to a line to consider
//!                     them as colinear (in degrees)
//! \return True if there was no error, else call atipGetError()
bool atipSearchVanPoints(float vanPoints[12], bool infiniteVanPoints[3],
						 float * lineClusters[4], unsigned int nLinesPerCluster[4],
						 unsigned int * hough, const float * lines,
						 const unsigned int * lineCoeffs, unsigned int nLines,
						 unsigned int nTheta, unsigned int nPhi, unsigned int windowSize,
						 unsigned int width, unsigned int height,
						 float maxLineDistance, float thetaInfinite,
						 unsigned int maxLineAngle);

//! Given 3 vanishing points, copy the source image to the destination image,
//! draw the vanishing points, and the line clusters (red, green ,blue, white)
//! \param dstImage Destination image of size width * height coded with 24 bits
//! \param srcImage Source image of size width * height coded with 24 bits
//! \param width Width of the images in pixels
//! \param height Height of the images in pixels
//! \param vanPoints Array of 3 quadruplets (x, y, phi, theta), one per vanishing point.
//!                  If the point is finite, (x, y) are the coordinates of the point
//!                  inside the image ; if the point is infinite, it is the unit direction
//!                  vector. (phi, theta) are the gaussian hemisphere coordinates
//! \param infiniteVanPoints Array of 3 booleans which are true if the corresponding
//!                          vanishing points are infinite
//! \param lineClusters Array of 4 arrays containing quadruplets (x0, y0, dx, dy)
//!                     representing the lines which contributed to detect the vanishing
//!                     points. The number of lines per cluster is nLinesPercluster[0..3].
//!                     The first 3 arrays correspond to the three vanishing points.
//!                     The last one is for the remaining lines
//! \param nLinesPerCluster Number of lines which contributed to detect the vanishing points,
//!                     array of integers
//! \return True if there was no error, else call atipGetError()
bool atipDrawVanPoints24(unsigned char * dstImage, const unsigned char * srcImage,
						 unsigned int width, unsigned int height,
						 const float vanPoints[12], const bool infiniteVanPoints[3],
						 float * const lineClusters[4],
						 const unsigned int nLinesPerCluster[4]);

#endif	// _ATIP_GAUSSIAN_HEMISPHERE_H_