/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_hough.h
//! \author	BOULANGER Kevin
//! \date	13th July 2004
//! \brief	Hough transforms


#ifndef _ATIP_HOUGH_H_
#define _ATIP_HOUGH_H_


//! Given a list of points, compute the polar Hough transform
//! \param hough Resulting Hough transform, array of rhoMax * nTheta accumulation values,
//!              allocated by the function
//! \param rhoMax Resulting width of the Hough transform
//! \param maxHough Resulting maximum accumulation value of the Hough transform
//! \param points nPoints groups (x, y) of point coordinates
//! \param nPoints Number of input points
//! \param width Width in pixels of the original image
//! \param height Height in pixels of the original image
//! \param nTheta Number of samples for the theta polar parameter
//! \return True if there was no error, else call atipGetError()
bool atipPolarHough(unsigned int * & hough, int & rhoMax, unsigned int & maxHough,
					const int * points, unsigned int nPoints,
					unsigned int width, unsigned int height, unsigned int nTheta);

//! Given a polar Hough transform, detect the maxima which will give the main lines
//! of the image
//! \param lines Resulting array of quadruplets (x0, y0, dx, dy), one for each line,
//!              allocated by the function
//! \param lineCoeffs Resulting array of coefficients, one per detected line, in order
//!                  to use their contribution, allocated by the function
//! \param nLines Resulting number of detected lines
//! \param houghHigh Polar Hough transform after a high-pass filter
//! \param hough Polar Hough after before the high-pass filter
//! \param nTheta Number of samples for the theta polar parameter
//! \param rhoMax Width of the Hough transform
//! \param maxHoughHigh Maximum accumulation value of the Polar Hough transform
//!                     after the high-pass filter
//! \param threshold Value of the threshold (between 0.0f and 1.0f)
//! \return True if there was no error, else call atipGetError()
bool atipPolarHoughThreshold(float * & lines, unsigned int * & lineCoeffs,
							 unsigned int & nLines,
							 const unsigned int * houghHigh, const unsigned int * hough,
							 unsigned int nTheta, int rhoMax,
							 unsigned int maxHoughHigh, float threshold);

//! Given a list of lines, copy the source image to the destination image and
//! draw the lines with clipping
//! \param dstImage Destination image of size width * height coded with 24 bits
//! \param srcImage Source image of size width * height coded with 24 bits
//! \param width Width of the images in pixels
//! \param height Height of the images in pixels
//! \param lines Array of quadruplets (x0, y0, dx, dy) for each line
//! \param lineCoeffs Array of coefficients, one per line, to use their contribution
//! \param maxLineCoeff Maximum line coefficient
//! \param nLines Number of lines to draw
//! \return True if there was no error, else call atipGetError()
bool atipDrawHoughLines24(unsigned char * dstImage, const unsigned char * srcImage,
						  unsigned int width, unsigned int height,
						  const float * lines, const unsigned int * lineCoeffs,
						  unsigned int maxLineCoeff, unsigned int nLines);


#endif	// _ATIP_HOUGH_H_