/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip.cpp
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Main source file for the "Automatic Tour Into the Picture" library (ATIP)


#include <atip.h>
#include <atip_options.h>
#include <van_point/atip_filter.h>
#include <van_point/atip_hough.h>
#include <van_point/atip_gaussian_hemisphere.h>
#include <van_point/atip_camera_calib.h>
#include <tools/atip_private_tools.h>
#include <atip_tip_maker.h>
#include <vector>
#include <math.h>


//! Array of images to process
std::vector<ATIPimage> srcImages;

//----------------------------------------------------------------------------------------

bool atipInit()
{
	srcImages.clear();
	return true;
}

//----------------------------------------------------------------------------------------

bool atipAddImageFromFile(const char * fileName)
{
	// Create a small image
	ATIPimage image(1, 1, 24);

	// Copy it to the set of images
	srcImages.push_back(image);

	// Load data from the file
	if (srcImages[srcImages.size() - 1].LoadFromFile(fileName)) return true;

	// In case of error, remove the image from the list
	srcImages.erase(srcImages.end() - 1);
	return false;
}

//----------------------------------------------------------------------------------------

void atipAddImageFromMemory(const ATIPimage & image)
{
	srcImages.push_back(image);
}

//----------------------------------------------------------------------------------------

void atipRemoveImage(unsigned int index)
{
	if (index < srcImages.size())
	{
		std::vector<ATIPimage>::iterator pos = srcImages.begin() + index;
		srcImages.erase(pos);
	}
}

//----------------------------------------------------------------------------------------

const ATIPimage * atipGetImage(unsigned int index)
{
	if (index < srcImages.size())
	{
		std::vector<ATIPimage>::iterator pos = srcImages.begin() + index;
		return &(*pos);
	}
	return 0;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetNumLoadedImages()
{
	return (unsigned int)srcImages.size();
}

//----------------------------------------------------------------------------------------

//! Number of process steps for each image (used for the percentage of processing)
#define ATIP_NSTEPS_PER_IMAGE		12

//! Macro to delete the allocated arrays
#define ATIP_PROCESS_DELETE_ARRAYS()	\
{										\
	ATIP_DELETE_ARRAY(image);			\
	ATIP_DELETE(smallOriginal);			\
	ATIP_DELETE_ARRAY(fImage);			\
	ATIP_DELETE_ARRAY(edgesMap);		\
	ATIP_DELETE_ARRAY(threshEdges);		\
	ATIP_DELETE_ARRAY(gradX);			\
	ATIP_DELETE_ARRAY(gradY);			\
	ATIP_DELETE_ARRAY(edgePoints);		\
	ATIP_DELETE_ARRAY(pHough);			\
	ATIP_DELETE_ARRAY(pHoughHigh);		\
	ATIP_DELETE_ARRAY(lines);			\
	ATIP_DELETE_ARRAY(lineCoeffs);		\
	ATIP_DELETE_ARRAY(gaussHough);		\
	ATIP_DELETE_ARRAY(lineClusters[0]);	\
	ATIP_DELETE_ARRAY(lineClusters[1]);	\
	ATIP_DELETE_ARRAY(lineClusters[2]);	\
	ATIP_DELETE_ARRAY(lineClusters[3]);	\
}

bool atipProcess(ATIPdata & atipData, unsigned char flags)
{
	// If an error occurred, quit directly
	if (atipGetError() != ATIP_NO_ERROR) return false;

	// Compute the number of total steps
	unsigned int nTotalSteps = 0;
	if (flags & ATIP_PROCESSFLAG_DETECT_VANISHING_POINTS)
		nTotalSteps += (ATIP_NSTEPS_PER_IMAGE - 2 + atipGetComputingSpeed())
									* (unsigned int)srcImages.size();
	if (flags & ATIP_PROCESSFLAG_EDIT_POLYGONS)
		nTotalSteps += (unsigned int)srcImages.size();
	if (flags & ATIP_PROCESSFLAG_COMPUTE_TEXTURES)
		nTotalSteps += (unsigned int)srcImages.size();

	std::vector<ATIPimage>::iterator im;
	unsigned int imIndex;
	char title[256] = "";
	unsigned int width, height;
	unsigned int currentStep = 0;
	unsigned char * image = 0;
	ATIPimage * smallOriginal = 0;
	float * fImage = 0;
	unsigned char * edgesMap = 0;
	float * gradX = 0;
	float * gradY = 0;
	unsigned char * threshEdges = 0;
	int * edgePoints = 0;
	unsigned int nEdgePoints = 0;
	unsigned int houghAngleResolution = 0;
	unsigned int * pHough = 0;
	unsigned int * pHoughHigh = 0;
	int rhoMax = 0;
	unsigned int maxHough = 0;
	unsigned int maxHoughHigh = 0;
	float * lines = 0;
	unsigned int * lineCoeffs = 0;
	unsigned int nLines = 0;
	unsigned int * gaussHough = 0;
	unsigned int maxGaussHough;
	float vanPoints[12];
	bool infiniteVanPoints[3] = { true, true, true };
	float * lineClusters[4] = { 0 };
	unsigned int nLinesPerCluster[4] = { 0 };
	unsigned int nFiniteVanPoints = 0;
	unsigned int nInfiniteVanPoints = 0;

	// Initialize the ATIPdata structure
	if (flags & ATIP_PROCESSFLAG_DETECT_VANISHING_POINTS)
	{
		(atipData.vertices, 0, 24 * sizeof(float));
		memset(atipData.rotVertices, 0, 24 * sizeof(float));
		memset(atipData.indices, 0, 20 * sizeof(unsigned int));
		ATIP_DELETE(atipData.textures[0]);
		ATIP_DELETE(atipData.textures[1]);
		ATIP_DELETE(atipData.textures[2]);
		ATIP_DELETE(atipData.textures[3]);
		ATIP_DELETE(atipData.textures[4]);
		atipData.focal = 0.0f;
		atipData.fov = 0.0f;
		memset(atipData.rot, 0, 9 * sizeof(float));
		atipData.rot[0] = atipData.rot[4] = atipData.rot[8] = 1.0f;
		atipData.distance = 100.0f;
		atipData.xLeft =  20.0f;
		atipData.xRight = 20.0f;
		atipData.yFloor = 20.0f;
		atipData.yCeiling = 20.0f;
		atipData.width = 0;
		atipData.height = 0;
	}

	// For each image to process
	for (im = srcImages.begin(), imIndex = 0; im != srcImages.end(); im++, imIndex++)
	{
		// Get some info about the image
		atipData.width = width = im->GetWidth();
		atipData.height = height = im->GetHeight();

		if (flags & ATIP_PROCESSFLAG_DETECT_VANISHING_POINTS)
		{
			// If the original image, or the image with the main lines,
			// or the image with the detected vanishing points must be displayed,
			// create a small version of the image
			if (   (   atipIsIntermResultActive(ATIP_INTRES_ORIGINAL)
					|| atipIsIntermResultActive(ATIP_INTRES_MAINLINES)
					|| atipIsIntermResultActive(ATIP_INTRES_VANPOINTS))
				&& atipGetComputingSpeed() >= ATIP_COMPSPEED_FAST)
			{
				// Convert the image to 24 bits
				ATIPimage tempIm(*im);
				tempIm.ConvertToRGB();
				unsigned char * smallIm = 0;
				unsigned int smallOrigWidth = 0;
				unsigned int smallOrigHeight = 0;
				
				// Reduce a first time
				if (!atipReduceImage2_24(smallIm, smallOrigWidth, smallOrigHeight,
										(unsigned char *)tempIm.GetImage(),
										width, height))
				{
					ATIP_PROCESS_DELETE_ARRAYS();
					return false;
				}

				// Reduce a second time for the very fast mode
				if (atipGetComputingSpeed() >= ATIP_COMPSPEED_VERYFAST)
				{
					unsigned char * verySmallIm = 0;
					if (!atipReduceImage2_24(verySmallIm, smallOrigWidth, smallOrigHeight,
											smallIm, smallOrigWidth, smallOrigHeight))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_DELETE_ARRAY(smallIm);
					smallIm = verySmallIm;
				}

				// Create the small image object
				ATIP_NEW(ATIPimage(smallOrigWidth, smallOrigHeight, 24, smallIm, false),
						smallOriginal);
			}

			// Display the original image if desired
			if (atipIsIntermResultActive(ATIP_INTRES_ORIGINAL))
			{
				sprintf(title, "Image %d : Original", imIndex + 1);
				unsigned int compSpeed = atipGetComputingSpeed();
				if (compSpeed == ATIP_COMPSPEED_FAST)
					strcat(title, " - Zoom 50%");
				else if (compSpeed == ATIP_COMPSPEED_VERYFAST)
					strcat(title, " - Zoom 25%");
				if (compSpeed == ATIP_COMPSPEED_NORMAL)
					atipCallIntermResultCallback(*im, title);
				else
				{
					ATIP_ASSERT(smallOriginal);
					atipCallIntermResultCallback(*smallOriginal, title);
				}
			}

			// Convert in greyscale if necessary
			ATIP_NEW_ARRAY(unsigned char, image, width * height);
			if (im->GetBpp() > 8)
			{
				if (!atipConvertToGrayscale(image, (unsigned char *)im->GetImage(),
											width, height, im->GetBpp()))
				{
					ATIP_PROCESS_DELETE_ARRAYS();
					return false;
				}
			}
			else
				// else copy the image
				memcpy(image, im->GetImage(), width * height);
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);

			// Reduce the size of the image if the fast or very fast mode is enabled
			if (atipGetComputingSpeed() >= ATIP_COMPSPEED_FAST)
			{
				unsigned char * smallImage = 0;
				if (!atipReduceImage2_8(smallImage, width, height, image, width, height))
				{
					ATIP_DELETE_ARRAY(smallImage);
					ATIP_PROCESS_DELETE_ARRAYS();
					return false;
				}
				ATIP_DELETE_ARRAY(image);
				image = smallImage;
				atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);

				if (atipGetComputingSpeed() >= ATIP_COMPSPEED_VERYFAST)
				{
					smallImage = 0;
					if (!atipReduceImage2_8(smallImage, width, height,
											image, width, height))
					{
						ATIP_DELETE_ARRAY(smallImage);
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_DELETE_ARRAY(image);
					image = smallImage;
					atipCallPercentageCallback((++currentStep)
												/ (float)nTotalSteps * 100.0f);
				}
			}

			// Display the grayscale version if desired
			if (atipIsIntermResultActive(ATIP_INTRES_GRAYSCALE))
			{
				ATIPimage grayIm(width, height, 8, image);
				sprintf(title, "Image %d : Grayscale version", imIndex + 1);
				if (atipGetComputingSpeed() == ATIP_COMPSPEED_FAST)
					strcat(title, " - Zoom 50%");
				else if (atipGetComputingSpeed() == ATIP_COMPSPEED_VERYFAST)
					strcat(title, " - Zoom 25%");
				atipCallIntermResultCallback(grayIm, title);
			}

			// Convert the image to an image of floats
			ATIP_NEW_ARRAY(float, fImage, width * height);
			if (!atipConvertImageToFloats(fImage, image, width, height))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);

			ATIP_NEW_ARRAY(unsigned char, edgesMap, width * height);
			ATIP_NEW_ARRAY(int, edgePoints, 2 * width * height);
			nEdgePoints = 0;

			// Filtering method
			switch (atipGetFilteringMethod())
			{
				// Deriche laplacian filter
				case ATIP_FILTERING_DERICHE_LAPLACIAN:

					// Compute the zero-crossing laplacians of the image
					if (!atipDericheLaplacians(edgesMap, fImage, width, height,
											atipGetDericheAlpha()))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					atipCallPercentageCallback((++currentStep)
												/ (float)nTotalSteps * 100.0f);
					if (atipIsIntermResultActive(ATIP_INTRES_LAPLACIANS))
					{
						ATIPimage edgesMapIm(width, height, 1, edgesMap);
						sprintf(title,
								"Image %d : Edges map after a Deriche laplacian filter",
								imIndex + 1);
						atipCallIntermResultCallback(edgesMapIm, title);
					}

					// Thresholding
					switch (atipGetThresholdingMethod())
					{
					case ATIP_THRESHOLDING_HYSTERESIS:

							if (atipGetNeighborsCheckState())
							{
								if (!atipHysteresisThresholdNeighbor(threshEdges,
										edgePoints, nEdgePoints, edgesMap, fImage,
										width, height, atipGetThresholdingLowLevel(),
                                        atipGetThresholdingHighLevel()))
								{
									ATIP_PROCESS_DELETE_ARRAYS();
									return false;
								}
							}
							else
							{
								if (!atipHysteresisThreshold(threshEdges,
										edgePoints, nEdgePoints, edgesMap, fImage, 
										width, height, atipGetThresholdingLowLevel(),
										atipGetThresholdingHighLevel()))
								{
									ATIP_PROCESS_DELETE_ARRAYS();
									return false;
								}
							}
							break;

						case ATIP_THRESHOLDING_SIMPLE:

							if (!atipSimpleThreshold(threshEdges,
													edgePoints, nEdgePoints, edgesMap,
													fImage, width, height,
													atipGetThresholdingHighLevel()))
							{
								ATIP_PROCESS_DELETE_ARRAYS();
								return false;
							}
							break;
					}
					break;

				// Simple laplacian filter
				case ATIP_FILTERING_SIMPLE_LAPLACIAN:

					// Compute the zero-crossing laplacians of the image
					if (!atipSimpleLaplacians(edgesMap, image, width, height))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					atipCallPercentageCallback((++currentStep)
												/ (float)nTotalSteps * 100.0f);
					if (atipIsIntermResultActive(ATIP_INTRES_LAPLACIANS))
					{
						ATIPimage edgesMapIm(width, height, 1, edgesMap);
						sprintf(title,
								"Image %d : Edges map after a simple laplacian filter",
								imIndex + 1);
						atipCallIntermResultCallback(edgesMapIm, title);
					}

					// Thresholding
					switch (atipGetThresholdingMethod())
					{
					case ATIP_THRESHOLDING_HYSTERESIS:

							if (atipGetNeighborsCheckState())
							{
								if (!atipHysteresisThresholdNeighbor(threshEdges,
										edgePoints, nEdgePoints, edgesMap, fImage,
										width, height, atipGetThresholdingLowLevel(),
										atipGetThresholdingHighLevel()))
								{
									ATIP_PROCESS_DELETE_ARRAYS();
									return false;
								}
							}
							else
							{
								if (!atipHysteresisThreshold(threshEdges,
										edgePoints, nEdgePoints, edgesMap, fImage,
										width, height, atipGetThresholdingLowLevel(),
										atipGetThresholdingHighLevel()))
								{
									ATIP_PROCESS_DELETE_ARRAYS();
									return false;
								}
							}
							break;

						case ATIP_THRESHOLDING_SIMPLE:

							if (!atipSimpleThreshold(threshEdges, edgePoints,
										nEdgePoints, edgesMap, fImage, width, height,
										atipGetThresholdingHighLevel()))
							{
								ATIP_PROCESS_DELETE_ARRAYS();
								return false;
							}
							break;
					}
					break;


				// Frei-Chen gradients filter
				case ATIP_FILTERING_FREICHEN_MAXIMA:

					// Compute the gradients of the image
					ATIP_NEW_ARRAY(float, gradX, width * height);
					ATIP_NEW_ARRAY(float, gradY, width * height);
					if (!atipFreiChenFilter(gradX, gradY, image, width, height))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					atipCallPercentageCallback((++currentStep)
												/ (float)nTotalSteps * 100.0f);
					if (atipIsIntermResultActive(ATIP_INTRES_GRADIENTS))
					{
						ATIPimage gradIm(width, height, 8);
						gradIm.ConvertFromFloats(gradX);
						sprintf(title,
								"Image %d : X Gradients after a Frei-Chen filter",
								imIndex + 1);
						atipCallIntermResultCallback(gradIm, title);
						gradIm.ConvertFromFloats(gradY);
						sprintf(title,
								"Image %d : Y Gradients after a Frei-Chen filter",
								imIndex + 1);
						atipCallIntermResultCallback(gradIm, title);
					}

					// Thresholding
					switch (atipGetThresholdingMethod())
					{
						case ATIP_THRESHOLDING_HYSTERESIS:

							if (!atipHysteresisThreshold(threshEdges,
														edgePoints, nEdgePoints,
														gradX, gradY, width, height,
														atipGetThresholdingLowLevel(),
														atipGetThresholdingHighLevel()))
							{
								ATIP_PROCESS_DELETE_ARRAYS();
								return false;
							}
							break;

						case ATIP_THRESHOLDING_SIMPLE:

							if (!atipSimpleThreshold(threshEdges,
													edgePoints, nEdgePoints,
													gradX, gradY, width, height,
													atipGetThresholdingHighLevel()))
							{
								ATIP_PROCESS_DELETE_ARRAYS();
								return false;
							}
							break;
					}
					break;
			}

			// Some cleaning to reduce memory usage
			ATIP_DELETE_ARRAY(image);
			ATIP_DELETE_ARRAY(fImage);
			ATIP_DELETE_ARRAY(edgesMap);
			ATIP_DELETE_ARRAY(gradX);
			ATIP_DELETE_ARRAY(gradY);

			// Display the result of the thresholding
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_THRESHOLD))
			{
				ATIPimage edgesMapThIm(width, height, 1, threshEdges);
				switch (atipGetThresholdingMethod())
				{
					case ATIP_THRESHOLDING_HYSTERESIS:	
						if (atipGetNeighborsCheckState())
							sprintf(title,
									"Image %d : Edges map after hysteresis thresholding"
									" and checking of all neighbors", imIndex + 1);
							sprintf(title,
									"Image %d : Edges map after hysteresis thresholding",
									imIndex + 1);
						break;

					case ATIP_THRESHOLDING_SIMPLE:
						sprintf(title, "Image %d : Edges map after simple thresholding",
								imIndex + 1);
						break;
				}
				atipCallIntermResultCallback(edgesMapThIm, title);
			}

			// Polar Hough transform to find the main lines of the image
			houghAngleResolution = atipGetHoughAngleResolution();
			if (houghAngleResolution == 0) houghAngleResolution = width;
			if (!atipPolarHough(pHough, rhoMax, maxHough, edgePoints, nEdgePoints,
								width, height, houghAngleResolution))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_POLARHOUGH))
			{
				unsigned int nAccum = rhoMax * houghAngleResolution;
				unsigned char * pHoughUChar = 0;
				ATIP_NEW_ARRAY(unsigned char, pHoughUChar, nAccum);
				for (unsigned int p = 0; p < nAccum; p++)
					pHoughUChar[p] = (unsigned char)(pHough[p] * 255 / maxHough);
				ATIPimage pHoughIm((unsigned int)rhoMax, houghAngleResolution,
								8, pHoughUChar, false);
				sprintf(title, "Image %d : Polar Hough transform, %d points, "
						"max : %d", imIndex + 1, nEdgePoints, maxHough);
				atipCallIntermResultCallback(pHoughIm, title);
			}

			// Some cleaning to reduce memory usage
			ATIP_DELETE_ARRAY(threshEdges);
			ATIP_DELETE_ARRAY(edgePoints);

			// High-pass filter on the polar Hough transform
			// to have a better maxima detection
			ATIP_NEW_ARRAY(unsigned int, pHoughHigh, rhoMax * houghAngleResolution);
			if (!atipHighPassFilter32(pHoughHigh, pHough, rhoMax, houghAngleResolution,
									maxHoughHigh))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_HIGHPASSHOUGH))
			{
				unsigned int nAccum = rhoMax * houghAngleResolution;
				unsigned char * pHoughUChar = 0;
				ATIP_NEW_ARRAY(unsigned char, pHoughUChar, nAccum);
				for (unsigned int p = 0; p < nAccum; p++)
					pHoughUChar[p] = (unsigned char)(pHoughHigh[p] * 255 / maxHoughHigh);
				ATIPimage pHoughIm((unsigned int)rhoMax, houghAngleResolution,
								8, pHoughUChar, false);
				sprintf(title, "Image %d : High-pass filter of the polar "
						"Hough transform, max : %d", imIndex + 1, maxHoughHigh);
				atipCallIntermResultCallback(pHoughIm, title);
			}

			// Look for the main lines of the image
			if (!atipPolarHoughThreshold(lines, lineCoeffs, nLines, pHoughHigh, pHough,
										houghAngleResolution, rhoMax, maxHoughHigh,
										atipGetPolarHoughThresholdingLevel()))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			// If the number of detected main (or edge) lines is not sufficient
			if (nLines < 4)
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				atipSetError(ATIP_NOT_ENOUGH_EDGE_LINES);
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_MAINLINES))
			{
				unsigned char * mainLines = 0;
				ATIPimage * mainLinesIm = 0;
				if (atipGetComputingSpeed() == ATIP_COMPSPEED_NORMAL)
				{
					ATIP_NEW_ARRAY(unsigned char, mainLines, width * height * 3);
					if (!atipDrawHoughLines24(mainLines, (unsigned char *)im->GetImage(),
								width, height, lines, lineCoeffs, maxHough, nLines))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_NEW(ATIPimage(width, height, 24, mainLines, false), mainLinesIm);
				}
				else				
				{	
					ATIP_NEW_ARRAY(unsigned char, mainLines,
							smallOriginal->GetWidth() * smallOriginal->GetHeight() * 3);
					if (!atipDrawHoughLines24(mainLines,
							(unsigned char *)smallOriginal->GetImage(),
							smallOriginal->GetWidth(), smallOriginal->GetHeight(),
							lines, lineCoeffs, maxHough, nLines))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_NEW(ATIPimage(smallOriginal->GetWidth(),
									   smallOriginal->GetHeight(),
									   24, mainLines, false), mainLinesIm);
				}
				sprintf(title, "Image %d : Main lines after the polar Hough transform, "
						"%d lines, threshold : %1.2f", imIndex + 1, nLines,
						atipGetPolarHoughThresholdingLevel());
				if (atipGetComputingSpeed() == ATIP_COMPSPEED_FAST)
					strcat(title, " - Zoom 50%");
				else if (atipGetComputingSpeed() == ATIP_COMPSPEED_VERYFAST)
					strcat(title, " - Zoom 25%");
				atipCallIntermResultCallback(*mainLinesIm, title);
				ATIP_DELETE(mainLinesIm);
			}

			// Some cleaning to reduce memory usage
			ATIP_DELETE_ARRAY(pHough);
			ATIP_DELETE_ARRAY(pHoughHigh);

			// Hough transform on the gaussian hemisphere to find the vanishing points
			unsigned int gaussWindowSize = atipGetGaussWindowSize();
			unsigned int gaussNPhi = atipGetGaussAzimuthResolution();
			unsigned int gaussNTheta = atipGetGaussElevationResolution();
			unsigned int gaussNPhi2 = gaussNPhi + gaussWindowSize - 1;
			unsigned int gaussNTheta2 = gaussNTheta + gaussWindowSize - 1;
			if (!atipGaussHough(gaussHough, maxGaussHough, lines,
								atipGetLinesContributionState() ? lineCoeffs : 0, nLines,
								gaussNTheta, gaussNPhi, gaussWindowSize, width, height))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_GAUSSHOUGH))
			{
				unsigned int nAccum = gaussNTheta2 * gaussNPhi2;
				unsigned char * gaussHoughUChar = 0;
				ATIP_NEW_ARRAY(unsigned char, gaussHoughUChar, nAccum);
				for (unsigned int p = 0; p < nAccum; p++)
					gaussHoughUChar[p] = (unsigned char)
											(gaussHough[p] * 255 / maxGaussHough);
				ATIPimage gaussHoughIm(gaussNTheta2, gaussNPhi2, 8,
									   gaussHoughUChar, false);
				sprintf(title, "Image %d : Hough transform on the gaussian hemisphere, "
						"%d lines, max : %d", imIndex + 1, nLines, maxGaussHough);
				atipCallIntermResultCallback(gaussHoughIm, title);
			}

			// Search the vanishing points
			if (!atipSearchVanPoints(vanPoints, infiniteVanPoints, lineClusters,
						nLinesPerCluster, gaussHough, lines,
						(atipGetLinesContributionState() ? lineCoeffs : 0), nLines,
						gaussNTheta, gaussNPhi, gaussWindowSize, width, height,
						atipGetMaxLineDistance(), atipGetInfiniteAngle(),
						atipGetMaxLineAngle()))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_VANPOINTS))
			{
				unsigned char * vanPoints24 = 0;
				ATIPimage * vanPointsIm = 0;
				if (atipGetComputingSpeed() == ATIP_COMPSPEED_NORMAL)
				{
					ATIP_NEW_ARRAY(unsigned char, vanPoints24, width * height * 3);
					if (!atipDrawVanPoints24(vanPoints24, (unsigned char *)im->GetImage(),
										width, height, vanPoints, infiniteVanPoints,
										lineClusters, nLinesPerCluster))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_NEW(ATIPimage(width, height, 24, vanPoints24, false),
							 vanPointsIm);
				}
				else				
				{	
					ATIP_NEW_ARRAY(unsigned char, vanPoints24,
							smallOriginal->GetWidth() * smallOriginal->GetHeight() * 3);
					if (!atipDrawVanPoints24(vanPoints24,
							(unsigned char *)smallOriginal->GetImage(),
							smallOriginal->GetWidth(), smallOriginal->GetHeight(),
							vanPoints, infiniteVanPoints, lineClusters, nLinesPerCluster))
					{
						ATIP_PROCESS_DELETE_ARRAYS();
						return false;
					}
					ATIP_NEW(ATIPimage(smallOriginal->GetWidth(),
									   smallOriginal->GetHeight(),
									   24, vanPoints24, false), vanPointsIm);
				}
				sprintf(title, "Image %d : Detected vanishing points "
						"%s(%.2f, %.2f, %.2f,%.2f) %s(%.2f, %.2f, %.2f,%.2f) "
						"%s(%.2f, %.2f, %.2f,%.2f)",
						imIndex + 1,
						infiniteVanPoints[0] ? "I" : "F", vanPoints[0], vanPoints[1],
														vanPoints[2], vanPoints[3],
						infiniteVanPoints[1] ? "I" : "F", vanPoints[4], vanPoints[5],
														vanPoints[6], vanPoints[7],
						infiniteVanPoints[2] ? "I" : "F", vanPoints[8], vanPoints[9],
														vanPoints[10], vanPoints[11]);
				if (atipGetComputingSpeed() == ATIP_COMPSPEED_FAST)
					strcat(title, " - Zoom 50%");
				else if (atipGetComputingSpeed() == ATIP_COMPSPEED_VERYFAST)
					strcat(title, " - Zoom 25%");
				atipCallIntermResultCallback(*vanPointsIm, title);
				ATIP_DELETE(vanPointsIm);
			}

			// Some cleaning to reduce memory usage
			ATIP_DELETE(smallOriginal);
			ATIP_DELETE_ARRAY(lines);
			ATIP_DELETE_ARRAY(lineCoeffs);
			ATIP_DELETE_ARRAY(gaussHough);
			ATIP_DELETE_ARRAY(lineClusters[0]);
			ATIP_DELETE_ARRAY(lineClusters[1]);
			ATIP_DELETE_ARRAY(lineClusters[2]);
			ATIP_DELETE_ARRAY(lineClusters[3]);
		
			// Calibrate the camera
			if (!atipCalibrateCamera(atipData.focal, atipData.fov, atipData.rot,
									 vanPoints, infiniteVanPoints, width, height))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
		}

		bool polygonsEditorCancelled = false;
		if (flags & ATIP_PROCESSFLAG_EDIT_POLYGONS)
		{
			// Number of finite and infinite vanishing points
			nInfiniteVanPoints = 0;
			if (infiniteVanPoints[0]) nInfiniteVanPoints++;
			if (infiniteVanPoints[1]) nInfiniteVanPoints++;
			if (infiniteVanPoints[2]) nInfiniteVanPoints++;
			nFiniteVanPoints = 3 - nInfiniteVanPoints;

			// Compute the coordinates of the vertices of the TIP polygons
			if (!atipCallPolygonsEditorCallback(atipData, *im,
												nFiniteVanPoints, nInfiniteVanPoints))
				polygonsEditorCancelled = true;
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
		}

		if ((flags & ATIP_PROCESSFLAG_COMPUTE_TEXTURES) && !polygonsEditorCancelled) 
		{
			//! Compute the textures used for the TIP polygons
			if (!atipComputeTextures(atipData.textures, *im, atipData.rotVertices,
									atipData.indices, atipData.focal))
			{
				ATIP_PROCESS_DELETE_ARRAYS();
				return false;
			}
			atipCallPercentageCallback((++currentStep) / (float)nTotalSteps * 100.0f);
			if (atipIsIntermResultActive(ATIP_INTRES_TEXTURES))
			{
				if (atipData.textures[0])
				{
					sprintf(title, "Image %d : Background texture", imIndex + 1);
					atipCallIntermResultCallback(*(atipData.textures[0]), title);
				}
				if (atipData.textures[1])
				{
					sprintf(title, "Image %d : Left wall texture", imIndex + 1);
					atipCallIntermResultCallback(*(atipData.textures[1]), title);
				}
				if (atipData.textures[2])
				{
					sprintf(title, "Image %d : Right wall texture", imIndex + 1);
					atipCallIntermResultCallback(*(atipData.textures[2]), title);
				}
				if (atipData.textures[3])
				{
					sprintf(title, "Image %d : Floor texture", imIndex + 1);
					atipCallIntermResultCallback(*(atipData.textures[3]), title);
				}
				if (atipData.textures[4])
				{
					sprintf(title, "Image %d : Ceiling texture", imIndex + 1);
					atipCallIntermResultCallback(*(atipData.textures[4]), title);
				}
			}
		}
	}

	// Do some cleaning
	ATIP_PROCESS_DELETE_ARRAYS();

	return (atipGetError() == ATIP_NO_ERROR);
}

//----------------------------------------------------------------------------------------

void atipSetPolygonsEditorCallback(ATIPpolygonsEditorProc proc)
{
	// In atip_private_tools.h
	atipSetPolyEdCallback(proc);
}