/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_file.cpp
//! \author	BOULANGER Kevin
//! \date	13th August 2004
//! \brief	Management of the ATIP file format


#include <atip_file.h>
#include <tools/atip_private_tools.h>
#include <ext/IL/il.h>
#include <stdio.h>
#include <string.h>


//@{
//! Chunks identifiers
#define ATIP_CHUNK_ORIGINAL_IMAGE_SIZE			0x0100
#define ATIP_CHUNK_VERTICES						0x0200
#define ATIP_CHUNK_ROTVERTICES					0x0201
#define ATIP_CHUNK_INDICES						0x0210
#define ATIP_CHUNK_POLYGONS_PARAMETERS			0x0220
#define ATIP_CHUNK_TEXTURE						0x0300
#define ATIP_CHUNK_FOCAL_FOV					0x0400
#define ATIP_CHUNK_CAMERA_ROTATION				0x0410
//@}

//----------------------------------------------------------------------------------------

//! Read an original image size chunk
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkOriginalImageSize(ATIPdata & atipData, FILE * atipFile)
{
	// Read the width of the image
	size_t nItems = fread(&(atipData.width), 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (atipData.width < 4) return 2;

	// Read the height of the image
	nItems = fread(&(atipData.height), 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (atipData.height < 4) return 2;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write an original image size chunk
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkOriginalImageSize(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_ORIGINAL_IMAGE_SIZE;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the width of the image
	if (atipData.width < 4) return 2;
	nItems = fwrite(&(atipData.width), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the height of the image
	if (atipData.height < 4) return 2;
	nItems = fwrite(&(atipData.height), 4, 1, atipFile);
	if (nItems != 1) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a vertices chunk (unrotated vertices describing the polygons)
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkVertices(ATIPdata & atipData, FILE * atipFile)
{
	// Read the number of vertices
	unsigned int nVertices;
	size_t nItems = fread(&nVertices, 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (nVertices != 8) return 2;		// For the moment, only 8 vertices are supported

	// Read the vertices coordinates
	nItems = fread(atipData.vertices, 3 * 4, nVertices, atipFile);
	if (nItems != nVertices) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a vertices chunk (unrotated vertices describing the polygons)
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkVertices(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_VERTICES;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the number of vertices
	unsigned int nVertices = 8;			// For the moment, only 8 vertices are supported
	nItems = fwrite(&nVertices, 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the vertices coordinates
	nItems = fwrite(atipData.vertices, 3 * 4, nVertices, atipFile);
	if (nItems != nVertices) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a rotated vertices chunk (describing the polygons)
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkRotVertices(ATIPdata & atipData, FILE * atipFile)
{
	// Read the number of vertices
	unsigned int nVertices;
	size_t nItems = fread(&nVertices, 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (nVertices != 8) return 2;		// For the moment, only 8 vertices are supported

	// Read the rotated vertices coordinates
	nItems = fread(atipData.rotVertices, 3 * 4, nVertices, atipFile);
	if (nItems != nVertices) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a rotated vertices chunk (describing the polygons)
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkRotVertices(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_ROTVERTICES;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the number of vertices
	unsigned int nVertices = 8;			// For the moment, only 8 vertices are supported
	nItems = fwrite(&nVertices, 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the vertices coordinates
	nItems = fwrite(atipData.rotVertices, 3 * 4, nVertices, atipFile);
	if (nItems != nVertices) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read an indices chunk (for the vertex array describing the polygons)
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkIndices(ATIPdata & atipData, FILE * atipFile)
{
	// Read the number of quadrilaterals
	unsigned int nQuads;
	size_t nItems = fread(&nQuads, 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (nQuads != 5) return 2;		// For the moment, only 5 quadrilaterals are supported

	// Read the indices
	nItems = fread(atipData.indices, 4 * 4, nQuads, atipFile);
	if (nItems != nQuads) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write an indices chunk (for the vertex array describing the polygons)
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkIndices(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_INDICES;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the number of quadrilaterals
	unsigned int nQuads = 5;		// For the moment, only 8 vertices are supported
	nItems = fwrite(&nQuads, 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the indices
	nItems = fwrite(atipData.indices, 4 * 4, nQuads, atipFile);
	if (nItems != nQuads) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a polygons parameters chunk
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkPolygonsParameters(ATIPdata & atipData, FILE * atipFile)
{
	// Read the distance of the background plane
	size_t nItems = fread(&(atipData.distance), 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (atipData.distance <= 0.0f) return 2;

	// Read the distance from the central point of the background plane to the left wall
	nItems = fread(&(atipData.xLeft), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Read the distance from the central point of the background plane to the right wall
	nItems = fread(&(atipData.xRight), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Read the distance from the central point of the background plane to the floor
	nItems = fread(&(atipData.yFloor), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Read the distance from the central point of the background plane to the ceiling
	nItems = fread(&(atipData.yCeiling), 4, 1, atipFile);
	if (nItems != 1) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a polygons parameters chunk
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkPolygonsParameters(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_POLYGONS_PARAMETERS;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the distance of the background plane
	if (atipData.distance <= 0.0f) return 2;
	nItems = fwrite(&(atipData.distance), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the distance from the central point of the background plane to the left wall
	nItems = fwrite(&(atipData.xLeft), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the distance from the central point of the background plane to the right wall
	nItems = fwrite(&(atipData.xRight), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the distance from the central point of the background plane to the floor
	nItems = fwrite(&(atipData.yFloor), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the distance from the central point of the background plane to the ceiling
	nItems = fwrite(&(atipData.yCeiling), 4, 1, atipFile);
	if (nItems != 1) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a texture chunk
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkTexture(ATIPdata & atipData, FILE * atipFile)
{
	// Read the index of the texture (for the atipData.textures array)
	unsigned int texIndex;
	size_t nItems = fread(&texIndex, 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (texIndex > 4) return 2;			// For the moment, only 5 textures are supported

	// Read the size of the texture file in bytes
	unsigned int texFileSize;
	nItems = fread(&texFileSize, 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (texFileSize == 0) return 2;

	// Read the texture data
	unsigned char * texData = 0;
	ATIP_NEW_ARRAY(unsigned char, texData, texFileSize);
	nItems = fread(texData, texFileSize, 1, atipFile);
	if (nItems != 1) return 1;

	// Create the DevIL image object
	ILuint imIndex;
	ilGenImages(1, &imIndex);
	ilBindImage(imIndex);

	// We want that the saved image is in the correct orientation
	ilOriginFunc(IL_ORIGIN_LOWER_LEFT);
	ilEnable(IL_ORIGIN_SET);

	if (!ilLoadL(IL_JPG, texData, texFileSize)) return 1;
	ATIP_DELETE_ARRAY(texData);

	// Create the ATIPimage texture object
	unsigned int width = ilGetInteger(IL_IMAGE_WIDTH);
	unsigned int height = ilGetInteger(IL_IMAGE_HEIGHT);
	if ((width == 0) || (height == 0)) return 2;
	ATIP_NEW(ATIPimage(width, height, 24, ilGetData()), atipData.textures[texIndex]);

	ilDeleteImages(1, &imIndex);
	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a texture chunk
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \param texIndex Index of the texture inside the atipData.textures array
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkTexture(FILE * atipFile, ATIPdata & atipData,
								   unsigned int texIndex)
{
	// Reference to the image to save
	ATIPimage & image = *(atipData.textures[texIndex]);

	// Create a DevIL image object
	ILuint imIndex;
	ilGenImages(1, &imIndex);
	ilBindImage(imIndex);

	// We want that the saved image is in the correct orientation
	ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
	ilEnable(IL_ORIGIN_SET);

	// Create an image object
	if (!ilTexImage(image.GetWidth(), image.GetHeight(), 1, 3,
					IL_RGB, IL_UNSIGNED_BYTE, image.GetImage())) return 1;

	// Save this to a temporary file
#define ATIP_TEMP_JPG_FILE			"temp5f9ea012.jpg"

	if (!ilSave(IL_JPG, ATIP_TEMP_JPG_FILE)) return 1;

	// Get the size of the file
	FILE * tmpFile = fopen(ATIP_TEMP_JPG_FILE, "rb");
	fseek(tmpFile, 0, SEEK_END);
	unsigned int texFileSize = ftell(tmpFile);
	if (texFileSize == 0)
	{
		fclose(tmpFile);
		remove(ATIP_TEMP_JPG_FILE);
		return 1;
	}

	// Load the JPG temporary file and delete it from the disk
	fseek(tmpFile, 0, SEEK_SET);
	unsigned char * texData = 0;
	ATIP_NEW_ARRAY(unsigned char, texData, texFileSize);
	size_t nItems = fread(texData, texFileSize, 1, tmpFile);
	fclose(tmpFile);
	remove(ATIP_TEMP_JPG_FILE);
	if (nItems != 1) return 1;

	// Write the chunk

	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_TEXTURE;
	nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the index of the texture
	if (texIndex > 4) return 2;			// For the moment, only 5 textures are supported
	if (!atipData.textures[texIndex]) return 2;		// The image must exist
	nItems = fwrite(&texIndex, 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the size of the texture file in bytes
	nItems = fwrite(&texFileSize, 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the texture data
	nItems = fwrite(texData, texFileSize, 1, atipFile);
	if (nItems != 1) return 1;

	// Delete the temporary data
	ilDeleteImages(1, &imIndex);
	ATIP_DELETE_ARRAY(texData);

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a focal and FOV chunk
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkFocalFov(ATIPdata & atipData, FILE * atipFile)
{
	// Read the focal length of the camera
	size_t nItems = fread(&(atipData.focal), 4, 1, atipFile);
	if (nItems != 1) return 1;
	if (atipData.focal <= 0.0f) return 2;

	// Read the field-of-view of the camera
	nItems = fread(&(atipData.fov), 4, 1, atipFile);
	if (nItems != 1) return 1;
	if ((atipData.fov <= 0.0f) || (atipData.fov >= 180.0f)) return 2;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a focal and FOV chunk
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkFocalFov(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_FOCAL_FOV;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the focal length of the camera
	if (atipData.focal <= 0.0f) return 2;
	nItems = fwrite(&(atipData.focal), 4, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the field-of-view of the camera
	if ((atipData.fov <= 0.0f) || (atipData.fov >= 180.0f)) return 2;
	nItems = fwrite(&(atipData.fov), 4, 1, atipFile);
	if (nItems != 1) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Read a camera rotation chunk
//! \param atipData Modified ATIPdata structure after the chunk reading
//! \param atipFile File inside which the chunk is read
//! \return 0 if there was no error, 1 if there was a file read error, 2 if there
//!         was a bad chunk error
unsigned int atipReadChunkCameraRotation(ATIPdata & atipData, FILE * atipFile)
{
	// Read the camera rotation matrix
	size_t nItems = fread(atipData.rot, 4, 9, atipFile);
	if (nItems != 9) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

//! Write a camera rotation chunk
//! \param atipFile File where the chunk is written
//! \param atipData ATIPdata structure containing the data
//! \return 0 if there was no error, 1 if there was a file write error, 2 if there
//!         was a bad data error
unsigned int atipWriteChunkCameraRotation(FILE * atipFile, ATIPdata & atipData)
{
	// Write the chunk ID
	unsigned short chunkID = ATIP_CHUNK_CAMERA_ROTATION;
	size_t nItems = fwrite(&chunkID, 2, 1, atipFile);
	if (nItems != 1) return 1;

	// Write the camera rotation matrix
	nItems = fwrite(atipData.rot, 4, 9, atipFile);
	if (nItems != 9) return 1;

	return 0;
}

//----------------------------------------------------------------------------------------

// Useful macros

#define ATIP_RETURN_READ		{	fclose(atipFile);						\
									atipSetError(ATIP_FILE_READ_ERROR);		\
									ATIP_DELETE(newData.textures[0]);		\
									ATIP_DELETE(newData.textures[1]);		\
									ATIP_DELETE(newData.textures[2]);		\
									ATIP_DELETE(newData.textures[3]);		\
									ATIP_DELETE(newData.textures[4]);		\
									return false;	 }
#define ATIP_RETURN_BAD			{	fclose(atipFile);						\
									atipSetError(ATIP_BAD_FILE_FORMAT);		\
									ATIP_DELETE(newData.textures[0]);		\
									ATIP_DELETE(newData.textures[1]);		\
									ATIP_DELETE(newData.textures[2]);		\
									ATIP_DELETE(newData.textures[3]);		\
									ATIP_DELETE(newData.textures[4]);		\
									return false;	 }

#define ATIP_RETURN_WRITE		{	fclose(atipFile);						\
									remove(fileName);						\
									atipSetError(ATIP_FILE_WRITE_ERROR);	\
									return false;	 }
#define ATIP_RETURN_INTERNAL	{	fclose(atipFile);						\
									remove(fileName);						\
									atipSetError(ATIP_INVALID_ARGUMENT);	\
									return false;	 }

//----------------------------------------------------------------------------------------

bool atipReadFile(ATIPdata & atipData, const char * fileName)
{
	// Initialize DevIL
	ATIP_ASSERT(ilGetInteger(IL_VERSION_NUM) >= IL_VERSION);
	ilInit();

	ATIPdata newData;
	memset(&newData, 0, sizeof(ATIPdata));

	// Open the file for reading
	FILE * atipFile = fopen(fileName, "rb");
	if (!atipFile)
	{
		atipSetError(ATIP_FILE_READ_ERROR);
		return false;
	}

	// Read the file ID
	char buffer[128];
	buffer[8] = '\0';
	size_t nItems = fread(buffer, 8, 1, atipFile);
	if (nItems != 1) ATIP_RETURN_READ;
	if (strcmp(buffer, "ATIPFILE")) ATIP_RETURN_BAD;

	// Read every chunks
	unsigned short chunkID;
	while (true)
	{
		// Read the chunk ID
		nItems = fread(&chunkID, 2, 1, atipFile);
		if (nItems != 1)
		{
			if (feof(atipFile)) break;
			ATIP_RETURN_READ;
		}

		switch (chunkID)
		{
			case ATIP_CHUNK_ORIGINAL_IMAGE_SIZE:
				atipReadChunkOriginalImageSize(newData, atipFile);
				break;

			case ATIP_CHUNK_VERTICES:
				atipReadChunkVertices(newData, atipFile);
				break;

			case ATIP_CHUNK_ROTVERTICES:
				atipReadChunkRotVertices(newData, atipFile);
				break;

			case ATIP_CHUNK_INDICES:
				atipReadChunkIndices(newData, atipFile);
				break;

			case ATIP_CHUNK_POLYGONS_PARAMETERS:
				atipReadChunkPolygonsParameters(newData, atipFile);
				break;

			case ATIP_CHUNK_TEXTURE:
				atipReadChunkTexture(newData, atipFile);
				break;

			case ATIP_CHUNK_FOCAL_FOV:
				atipReadChunkFocalFov(newData, atipFile);
				break;

			case ATIP_CHUNK_CAMERA_ROTATION:
				atipReadChunkCameraRotation(newData, atipFile);
				break;

			default:
				ATIP_RETURN_BAD;
		}
	}
	fclose(atipFile);

	// Clean the atipData structure
	ATIP_DELETE(atipData.textures[0]);
	ATIP_DELETE(atipData.textures[1]);
	ATIP_DELETE(atipData.textures[2]);
	ATIP_DELETE(atipData.textures[3]);
	ATIP_DELETE(atipData.textures[4]);

	// Copy the new ATIPdata structure
	atipData = newData;
	return true;
}

//----------------------------------------------------------------------------------------

bool atipWriteFile(const char * fileName, ATIPdata & atipData)
{
	// Open the file for writing
	FILE * atipFile = fopen(fileName, "wb");
	if (!atipFile)
	{
		atipSetError(ATIP_FILE_WRITE_ERROR);
		return false;
	}

	// Write the file ID
	char buffer[128] = "ATIPFILE";
	size_t nItems = fwrite(buffer, 8, 1, atipFile);
	if (nItems != 1) ATIP_RETURN_WRITE;

	// Write the chunks
	unsigned int error;
	error = atipWriteChunkOriginalImageSize(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	error = atipWriteChunkVertices(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	error = atipWriteChunkRotVertices(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	error = atipWriteChunkIndices(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	error = atipWriteChunkPolygonsParameters(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	if (atipData.textures[0])
	{
		error = atipWriteChunkTexture(atipFile, atipData, 0);
		if (error == 1) ATIP_RETURN_WRITE;
		if (error == 2) ATIP_RETURN_INTERNAL;
	}
	if (atipData.textures[1])
	{
		error = atipWriteChunkTexture(atipFile, atipData, 1);
		if (error == 1) ATIP_RETURN_WRITE;
		if (error == 2) ATIP_RETURN_INTERNAL;
	}
	if (atipData.textures[2])
	{
		error = atipWriteChunkTexture(atipFile, atipData, 2);
		if (error == 1) ATIP_RETURN_WRITE;
		if (error == 2) ATIP_RETURN_INTERNAL;
	}
	if (atipData.textures[3])
	{
		error = atipWriteChunkTexture(atipFile, atipData, 3);
		if (error == 1) ATIP_RETURN_WRITE;
		if (error == 2) ATIP_RETURN_INTERNAL;
	}
	if (atipData.textures[4])
	{
		error = atipWriteChunkTexture(atipFile, atipData, 4);
		if (error == 1) ATIP_RETURN_WRITE;
		if (error == 2) ATIP_RETURN_INTERNAL;
	}
	error = atipWriteChunkFocalFov(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;
	error = atipWriteChunkCameraRotation(atipFile, atipData);
	if (error == 1) ATIP_RETURN_WRITE;
	if (error == 2) ATIP_RETURN_INTERNAL;

	fclose(atipFile);
	return true;
}