/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_image.cpp
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Image class (for any type of image)


#include <atip_image.h>
#include <tools/atip_private_tools.h>
#include <ext/IL/il.h>
#include <math.h>


// Static members
unsigned int ATIPimage::nImLoaded = 0;

//----------------------------------------------------------------------------------------

ATIPimage::ATIPimage()
:	_width(128),
	_height(128),
	_bpp(24),
	_nBytesPerPixel(3),
	_data(0),
	_rows(0),
	_paletteAlpha(false),
	_palette(0)
{
	// Data allocation for the image
	ATIP_NEW_ARRAY(unsigned char, _data, _width * _height * _nBytesPerPixel);

	// Fill the array of pointers to the rows of the image
	ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
	unsigned int i;
	_rows[0] = _data;
	for (i = 1; i < _height; i++)
		_rows[i] = _rows[i - 1] + _width * _nBytesPerPixel;

	// Create the checkerboard image
	unsigned int j;
	unsigned int offset = 0;
	for (i = 0; i < _height; i++)
		for (j = 0; j < _width; j++)
		{
			//! \todo Verify the checkerboard pattern
			_rows[i][j * _nBytesPerPixel    ] = (i >> 2) ^ (j >> 2);
			_rows[i][j * _nBytesPerPixel + 1] = (i >> 2) ^ (j >> 2);
			_rows[i][j * _nBytesPerPixel + 2] = (i >> 2) ^ (j >> 2);
		}

	if (nImLoaded == 0)
	{
		// Check if the shared library's version of DevIL matches the executable's version
		ATIP_ASSERT(ilGetInteger(IL_VERSION_NUM) >= IL_VERSION);
		ilInit();
	}
	nImLoaded++;
}

//----------------------------------------------------------------------------------------

ATIPimage::ATIPimage(unsigned int width, unsigned int height, unsigned char bpp)
:	_width(width),
	_height(height),
	_bpp(bpp),
	_data(0),
	_rows(0),
	_paletteAlpha(false),
	_palette(0)
{
	ATIP_ASSERT(((bpp >= 1) && (bpp <= 8)) || (bpp == 24) || (bpp == 32));

	// Compute the number of bytes per pixel
	_nBytesPerPixel = bpp / 8;
	if (_nBytesPerPixel == 0) _nBytesPerPixel = 1;

	ATIP_ASSERT(width > 0);
	ATIP_ASSERT(height > 0);
	if ((width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return;
	}

	// Create an empty image (uninitialized data)
	ATIP_NEW_ARRAY(unsigned char, _data, _width * _height * _nBytesPerPixel);

	// Fill the array of pointers to the rows of the image
	ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
	unsigned int i;
	_rows[0] = _data;
	for (i = 1; i < _height; i++)
		_rows[i] = _rows[i - 1] + _width * _nBytesPerPixel;

	// Fill the palette with grayscale if necessary
	if (bpp <= 8)
	{
		ATIP_NEW_ARRAY(unsigned char, _palette, (1 << bpp) * 3);
		for (i = 0; i < unsigned(1 << bpp); i++)
			_palette[i * 3] = _palette[i * 3 + 1] = _palette[i * 3 + 2]
							= (unsigned char)((int)i * 255 / int((1 << bpp) - 1));
	}

	if (nImLoaded == 0)
	{
		// Check if the shared library's version of DevIL matches the executable's version
		ATIP_ASSERT(ilGetInteger(IL_VERSION_NUM) >= IL_VERSION);
		ilInit();
	}
	nImLoaded++;
}

//----------------------------------------------------------------------------------------

ATIPimage::ATIPimage(unsigned int width, unsigned int height, unsigned char bpp,
					 void * data, bool copy)
:	_width(width),
	_height(height),
	_bpp(bpp),
	_data(0),
	_rows(0),
	_paletteAlpha(false),
	_palette(0)
{
	ATIP_ASSERT(((bpp >= 1) && (bpp <= 8)) || (bpp == 24) || (bpp == 32));
	ATIP_ASSERT(data);

	// Compute the number of bytes per pixel
	_nBytesPerPixel = bpp / 8;
	if (_nBytesPerPixel == 0) _nBytesPerPixel = 1;

	ATIP_ASSERT(width > 0);
	ATIP_ASSERT(height > 0);
	if ((width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return;
	}

	if (copy)
	{
		// Create the image from a copy of data
		ATIP_NEW_ARRAY(unsigned char, _data, _width * _height * _nBytesPerPixel);
		memcpy(_data, data, _width * _height * _nBytesPerPixel);
	}
	else
		// Copy only the pointer to the data
		_data = (unsigned char *)data;

	// Fill the array of pointers to the rows of the image
	ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
	unsigned int i;
	_rows[0] = _data;
	for (i = 1; i < _height; i++)
		_rows[i] = _rows[i - 1] + _width * _nBytesPerPixel;

	// Fill the palette with grayscale if necessary
	if (bpp <= 8)
	{
		ATIP_NEW_ARRAY(unsigned char, _palette, (1 << _bpp) * 3);
		for (i = 0; i < unsigned(1 << _bpp); i++)
			_palette[i * 3] = _palette[i * 3 + 1] = _palette[i * 3 + 2]
							= (unsigned char)((int)i * 255 / int((1 << _bpp) - 1));
	}

	if (nImLoaded == 0)
	{
		// Check if the shared library's version of DevIL matches the executable's version
		ATIP_ASSERT(ilGetInteger(IL_VERSION_NUM) >= IL_VERSION);
		ilInit();
	}
	nImLoaded++;
}

//----------------------------------------------------------------------------------------

ATIPimage::ATIPimage(const ATIPimage & image)
{
	*this = image;
	nImLoaded++;
}

//----------------------------------------------------------------------------------------

ATIPimage::~ATIPimage()
{
	ATIP_DELETE_ARRAY(_data);
	ATIP_DELETE_ARRAY(_rows);
	ATIP_DELETE_ARRAY(_palette);

	// Shutdown DevIL if the last image is deleted
	if (--nImLoaded == 0)
		ilShutDown();
}

//----------------------------------------------------------------------------------------

ATIPimage & ATIPimage::operator = (const ATIPimage & image)
{
	ATIP_ASSERT(image._width > 0);
	ATIP_ASSERT(image._height > 0);
	ATIP_ASSERT(((image._bpp >= 1) && (image._bpp <= 8)) || (image._bpp == 24)
				|| (image._bpp == 32));
	ATIP_ASSERT(image._data);

	_width = image._width;
	_height = image._height;
	_bpp = image._bpp;
	_nBytesPerPixel = image._nBytesPerPixel;
	_data = 0;
	_rows = 0;
	_paletteAlpha = image._paletteAlpha;
	_palette = 0;

	// Create the image from a copy of the original data
	ATIP_NEW_ARRAY(unsigned char, _data, _width * _height * _nBytesPerPixel);
	memcpy(_data, image._data, _width * _height * _nBytesPerPixel);

	// Fill the array of pointers to the rows of the image
	ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
	unsigned int i;
	_rows[0] = _data;
	for (i = 1; i < _height; i++)
		_rows[i] = _rows[i - 1] + _width * _nBytesPerPixel;

	// Copy the palette from the original image
	if (_nBytesPerPixel == 1)
	{
		ATIP_NEW_ARRAY(unsigned char, _palette, (1 << _bpp) * (_paletteAlpha ? 4 : 3));
		memcpy(_palette, image._palette, (1 << _bpp) * (_paletteAlpha ? 4 : 3));
	}

	return *this;
}

//----------------------------------------------------------------------------------------

bool ATIPimage::LoadFromFile(const char * fileName)
{
	// Generate an image object
	ILuint imgId;
	ilGenImages(1, &imgId);
	ilBindImage(imgId);

	// We want that the loaded image is in the correct orientation
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_UPPER_LEFT);

	// Loads the image specified by fileName into the image named by imgId
	if (!ilLoadImage(ILstring(fileName)))
	{
		ILenum error = ilGetError();
		if ((error == IL_COULD_NOT_OPEN_FILE) || (error == IL_FILE_READ_ERROR))
				atipSetError(ATIP_FILE_READ_ERROR);
		else	atipSetError(ATIP_BAD_FILE_FORMAT);
		return false;
	}

	// The image must be coded with 8 bits components
	if (ilGetInteger(IL_IMAGE_TYPE) != IL_UNSIGNED_BYTE)
	{
		atipSetError(ATIP_BAD_FILE_FORMAT);
		return false;
	}

	// Basic information about the image
	unsigned int width = ilGetInteger(IL_IMAGE_WIDTH);
	unsigned int height = ilGetInteger(IL_IMAGE_HEIGHT);
	unsigned char bpp = (unsigned char)ilGetInteger(IL_IMAGE_BITS_PER_PIXEL);
	unsigned char nBytesPerPixel = bpp / 8;
	if (nBytesPerPixel == 0) nBytesPerPixel = 1;

	// Allocation of the new array for the image data
	unsigned char * newData = 0;
	ATIP_NEW_ARRAY(unsigned char, newData, width * height * nBytesPerPixel);
	unsigned char * data = ilGetData();
	unsigned int nBytes = width * height * nBytesPerPixel;

	bool newPaletteAlpha = false;
	unsigned char * newPalette = 0;
		
	// For 24 and 32 bits file formats
	if ((nBytesPerPixel == 3) || (nBytesPerPixel == 4))
	{
		if (nBytesPerPixel == 3)
				ilConvertImage(IL_RGB, IL_UNSIGNED_BYTE);
		else	ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE);
		memcpy(newData, data, width * height * nBytesPerPixel);
	}

	// For palettized formats
	else if (nBytesPerPixel == 1)
	{
		// Alpha component of the palette
		unsigned int paletteType = ilGetInteger(IL_PALETTE_TYPE);
		newPaletteAlpha = (paletteType == IL_PAL_RGBA32) || (paletteType == IL_PAL_BGRA32);
		bool paletteAlpha =    (paletteType == IL_PAL_RGBA32) || (paletteType == IL_PAL_BGRA32)
							|| (paletteType == IL_PAL_RGB32) || (paletteType == IL_PAL_BGR32);

		// Determine the size of the palette
		unsigned int paletteNumCols;
		if (paletteType == IL_PAL_NONE) paletteNumCols = 256;
		else paletteNumCols = ilGetInteger(IL_PALETTE_NUM_COLS);

		unsigned int paletteSize = atipNextPowerOf2(paletteNumCols);
		bpp = atipLog2(paletteSize);

		// Allocate the array for the new palette
		unsigned int nBytesNewPal = paletteSize * (newPaletteAlpha ? 4 : 3);
		unsigned int nBytesPal = paletteNumCols * (paletteAlpha ? 4 : 3);
		ATIP_NEW_ARRAY(unsigned char, newPalette, nBytesNewPal);
		memset(newPalette, 0, nBytesNewPal);
		unsigned char * palette = ilGetPalette();

		unsigned int i, j;

		// If the palette does not exist, create a grayscale one
		if (paletteType == IL_PAL_NONE)
			for (i = 0; i < 256; i++)
				newPalette[i * 3] = newPalette[i * 3 + 1]
								  = newPalette[i * 3 + 2]
								  = (unsigned char)((int)i * 255 / int((1 << bpp) - 1));

		// If the red and blue components must be swapped in 24 bits
		else if (paletteType == IL_PAL_BGR24)
			for (i = 0; i < nBytesPal; i += 3)
			{
				newPalette[i    ] = palette[i + 2];
				newPalette[i + 1] = palette[i + 1];
				newPalette[i + 2] = palette[i    ];
			}
		
		// If the red and blue components must be swapped in 24/32 bits
		else if (paletteType == IL_PAL_BGR32)
			for (i = 0, j = 0; j < nBytesPal; i += 3, j += 4)
			{
				newPalette[i    ] = palette[j + 2];
				newPalette[i + 1] = palette[j + 1];
				newPalette[i + 2] = palette[j    ];
			}
		
		// If the red and blue components must be swapped in 32 bits
		else if (paletteType == IL_PAL_BGRA32)
			for (i = 0; i < nBytesPal; i += 4)
			{
				newPalette[i    ] = palette[i + 2];
				newPalette[i + 1] = palette[i + 1];
				newPalette[i + 2] = palette[i    ];
				newPalette[i + 3] = palette[i + 3];
			}

		// Direct copy of the palette
		else
			memcpy(newPalette, palette, nBytesPal);

		// Copy the data of the image
		memcpy(newData, data, width * height * nBytesPerPixel);
	}

	// Bad pixel format
	else
	{
		atipSetError(ATIP_BAD_FILE_FORMAT);
		return false;
	}

	// Now the image is loaded, members can be updated
	_width = width;
	_height = height;
	_bpp = bpp;
	_nBytesPerPixel = _bpp / 8;
	if (_nBytesPerPixel == 0) _nBytesPerPixel = 1;

	ATIP_DELETE_ARRAY(_data);
	_data = newData;

	ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
	unsigned int i;
	_rows[0] = _data;
	for (i = 1; i < _height; i++)
		_rows[i] = _rows[i - 1] + _width * _nBytesPerPixel;

	_paletteAlpha = newPaletteAlpha;

	ATIP_DELETE_ARRAY(_palette);
	_palette = newPalette;

	// Delete the DevIL image object
	ilDeleteImages(1, &imgId);

	return true;
}

//----------------------------------------------------------------------------------------

bool ATIPimage::SaveToFile(const char * fileName)
{
	ILuint imIndex;
	ilGenImages(1, &imIndex);
	ilBindImage(imIndex);
	
	//! \todo Verify for _bpp < 8
	ILenum format = IL_RGB;
	switch (_nBytesPerPixel)
	{
		case 1:	format = IL_COLOUR_INDEX;	break;
		case 3:	format = IL_RGB;			break;
		case 4:	format = IL_RGBA;			break;
		default:	ilDeleteImages(1, &imIndex);
					atipSetError(ATIP_BAD_FILE_FORMAT);
					return false;
	}

	// We want that the saved image is in the correct orientation
	ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
	ilEnable(IL_ORIGIN_SET);

	// Create an image object
	if (!ilTexImage(_width, _height, 1, _nBytesPerPixel, format, IL_UNSIGNED_BYTE, _data))
	{
		ilDeleteImages(1, &imIndex);
		atipSetError(ATIP_INVALID_OPERATION);
		return false;
	}

	// Set the palette if necessary
	if (_nBytesPerPixel == 1)
	{
		//! \todo Verify the functionality
		ilRegisterPal(_palette, (1 << _bpp) * (_paletteAlpha ? 4 : 3),
					  (_paletteAlpha ? IL_PAL_RGB24 : IL_PAL_RGBA32));
		ilEnable(IL_CONV_PAL);
	}

	ilEnable(IL_FILE_OVERWRITE);
	if (!ilSaveImage(ILstring(fileName)))
	{
		ilDeleteImages(1, &imIndex);
		atipSetError(ATIP_FILE_WRITE_ERROR);
		return false;
	}

	// Delete the DevIL image object
	ilDeleteImages(1, &imIndex);

	return true;
}

//----------------------------------------------------------------------------------------

bool ATIPimage::SetPalette(bool alpha, const void * palette)
{
	ATIP_ASSERT((_bpp >= 1) && (_bpp <= 8));
	if (!palette)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}
	
	// If a resizing of the palette is needed
	if (_paletteAlpha != alpha)
	{
		_paletteAlpha = alpha;
		ATIP_NEW_ARRAY(unsigned char, _palette, (1 << _bpp) * (_paletteAlpha ? 4 : 3));
	}

	// Copy the palette
	memcpy(_palette, palette, (1 << _bpp) * (_paletteAlpha ? 4 : 3));

	return true;
}

//----------------------------------------------------------------------------------------

void ATIPimage::ConvertToGreyscale()
{
	// Only if the image is coded in 24 or 32 bits
	if ((_bpp == 24) || (_bpp == 32))
	{
		unsigned int nPixels = _width * _height;

		// Allocate a temporary image
		unsigned char * newData = 0;
		ATIP_NEW_ARRAY(unsigned char, newData, nPixels);

		// Do the conversion
		unsigned int i, p;
		for (i = 0, p = 0; i < nPixels; i++, p += _nBytesPerPixel)
			newData[i] = (unsigned char)floor(  (float)_data[p] * 0.299f
											  + (float)_data[p + 1] * 0.587f
											  + (float)_data[p + 2] * 0.114f);

		// Delete the old image and replace by the new one
		ATIP_NEW_ARRAY(unsigned char, _data, nPixels);
		_data = newData;

		// Update the members
		_bpp = 8;
		_nBytesPerPixel = 1;

		// Fill the array of pointers to the rows of the image
		ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
		_rows[0] = _data;
		for (i = 1; i < _height; i++)
			_rows[i] = _rows[i - 1] + _width;

		// Fill the palette with grayscale
		_paletteAlpha = false;
		ATIP_NEW_ARRAY(unsigned char, _palette, 256 * 3);
		unsigned int j;
		for (i = 0, j = 0; i < 256; i++, j += 3)
			_palette[j] = _palette[j + 1] = _palette[j + 2] = i;
	}
}

//----------------------------------------------------------------------------------------

void ATIPimage::ConvertToRGB()
{
	if (_bpp != 24)
	{
		unsigned int nBytes = _width * _height * 3;

		// Allocate a temporary image
		unsigned char * newData = 0;
		ATIP_NEW_ARRAY(unsigned char, newData, nBytes);

		// Do the conversion from 32 bits
		if (_bpp == 32)
		{
			unsigned int p24, p32;
			for (p24 = 0, p32 = 0; p24 < nBytes; p24 += 3, p32 += 4)
			{
				newData[p24    ] = _data[p32    ];
				newData[p24 + 1] = _data[p32 + 1];
				newData[p24 + 2] = _data[p32 + 2];
			}
		}

		// Do the conversion from a palettized image
		if (_bpp <= 8)
		{
			unsigned int p8, p24;
			if (_paletteAlpha)
				for (p8 = 0, p24 = 0; p24 < nBytes; p8++, p24 += 3)
				{
					newData[p24    ] = _palette[ _data[p8] << 2     ];
					newData[p24 + 1] = _palette[(_data[p8] << 2) + 1];
					newData[p24 + 2] = _palette[(_data[p8] << 2) + 2];
				}
			else
				for (p8 = 0, p24 = 0; p24 < nBytes; p8++, p24 += 3)
				{
					newData[p24    ] = _palette[_data[p8] * 3    ];
					newData[p24 + 1] = _palette[_data[p8] * 3 + 1];
					newData[p24 + 2] = _palette[_data[p8] * 3 + 2];
				}
		}

		// Delete the old image and replace by the new one
		ATIP_NEW_ARRAY(unsigned char, _data, nBytes);
		_data = newData;

		// Update the members
		_bpp = 24;
		_nBytesPerPixel = 3;

		// Fill the array of pointers to the rows of the image
		ATIP_NEW_ARRAY(unsigned char *, _rows, _height);
		_rows[0] = _data;
		unsigned int i;
		for (i = 1; i < _height; i++)
			_rows[i] = _rows[i - 1] + _width * 3;

		// Delete the palette
		_paletteAlpha = false;
		ATIP_DELETE_ARRAY(_palette);
	}
}

//----------------------------------------------------------------------------------------

bool ATIPimage::ConvertFromFloats(float * image, float max, float min)
{
	ATIP_ASSERT(image);
	if (max < min)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int nPixels = _width * _height;
	unsigned int i;
	
	// If desired, look for the maximum and minimum values inside the image
	if ((max == 0.0f) && (min == 0.0f))
		for (i = 0; i < nPixels; i++)
			if (image[i] < min) min = image[i];
			else if (image[i] > max) max = image[i];

	// Convert the image
	float amplitude = max - min;
	for (i = 0; i < nPixels; i++)
		_data[i] = (unsigned char)floor((image[i] - min) / amplitude * 255.0f);

	// Fill the palette with grayscale
	_paletteAlpha = false;
	ATIP_NEW_ARRAY(unsigned char, _palette, 256 * 3);
	unsigned int j;
	for (i = 0, j = 0; i < 256; i++, j += 3)
		_palette[j] = _palette[j + 1] = _palette[j + 2] = i;

	return true;
}
