/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_options.cpp
//! \author	BOULANGER Kevin
//! \date	25th June 2004
//! \brief	Options for the "Automatic Tour Into the Picture" library (ATIP)


#include <atip_options.h>
#include <van_point/atip_filter.h>
#include <tools/atip_private_tools.h>


// Options of the ATIP processing (with their default values)
unsigned char filteringMethod = ATIP_FILTERING_SIMPLE_LAPLACIAN;
unsigned char thresholdingMethod = ATIP_THRESHOLDING_SIMPLE;
bool neighborsCheckState = false;
float dericheAlpha = 1.0f;
float thresholdingLowLevel = 5.0f;
float thresholdingHighLevel = 20.0f;
unsigned int houghAngleResolution = 0;
float polarHoughThreshLevel = 0.2f;
bool linesContribState = false;
unsigned int gaussAzimuthRes = 800;
unsigned int gaussElevationRes = 600;
unsigned int gaussWindowSize = 5;
float maxLineDistance = 0.04f;
float infiniteAngle = 1.525f;
unsigned int maxLineAngle = 6;
unsigned int computingSpeed = ATIP_COMPSPEED_FAST;
float fov = 48.0f;
unsigned int maximumTextureSize = 1024;

//----------------------------------------------------------------------------------------

void atipSetFilteringMethod(unsigned char method)
{
	ATIP_ASSERT(   method == ATIP_FILTERING_DERICHE_LAPLACIAN
				|| method == ATIP_FILTERING_SIMPLE_LAPLACIAN
				|| method == ATIP_FILTERING_FREICHEN_MAXIMA);
	filteringMethod = method;
}

//----------------------------------------------------------------------------------------

unsigned char atipGetFilteringMethod()
{
	return filteringMethod;
}

//----------------------------------------------------------------------------------------

void atipSetThresholdingMethod(unsigned char method)
{
	ATIP_ASSERT(   method == ATIP_THRESHOLDING_HYSTERESIS
				|| method == ATIP_THRESHOLDING_SIMPLE);
	thresholdingMethod = method;
}

//----------------------------------------------------------------------------------------

unsigned char atipGetThresholdingMethod()
{
	return thresholdingMethod;
}

//----------------------------------------------------------------------------------------

void atipSetNeighborsCheckState(bool state)
{
	neighborsCheckState = state;
}

//----------------------------------------------------------------------------------------

bool atipGetNeighborsCheckState()
{
	return neighborsCheckState;
}

//----------------------------------------------------------------------------------------

void atipSetDericheAlpha(float alpha)
{
	ATIP_ASSERT(alpha >= 0.0f);
	dericheAlpha = alpha;
}

//----------------------------------------------------------------------------------------

float atipGetDericheAlpha()
{
	return dericheAlpha;
}

//----------------------------------------------------------------------------------------

void atipSetThresholdingLowLevel(float low)
{
	ATIP_ASSERT((low >= 0.0f) && (low < thresholdingHighLevel));
	thresholdingLowLevel = low;
}

//----------------------------------------------------------------------------------------

float atipGetThresholdingLowLevel()
{
	return thresholdingLowLevel;
}

//----------------------------------------------------------------------------------------

void atipSetThresholdingHighLevel(float high)
{
	ATIP_ASSERT(   (thresholdingMethod == ATIP_THRESHOLDING_HYSTERESIS
					&& (high > thresholdingLowLevel) && (high <= ATIP_MAX_THRESHOLD))
				|| (thresholdingMethod == ATIP_THRESHOLDING_SIMPLE
					&& (high >= 0.0f) && (high <= ATIP_MAX_THRESHOLD)) );
	thresholdingHighLevel = high;
}

//----------------------------------------------------------------------------------------

float atipGetThresholdingHighLevel()
{
	return thresholdingHighLevel;
}

//----------------------------------------------------------------------------------------

void atipSetHoughAngleResolution(unsigned int resolution)
{
	houghAngleResolution = resolution;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetHoughAngleResolution()
{
	return houghAngleResolution;
}

//----------------------------------------------------------------------------------------

void atipSetPolarHoughThresholdingLevel(float level)
{
	ATIP_ASSERT((level > 0.0f) && (level < 1.0f));
	polarHoughThreshLevel = level;
}

//----------------------------------------------------------------------------------------

float atipGetPolarHoughThresholdingLevel()
{
	return polarHoughThreshLevel;
}

//----------------------------------------------------------------------------------------

void atipSetGaussAzimuthResolution(unsigned int resolution)
{
	ATIP_ASSERT(resolution >= 4);
	gaussAzimuthRes = resolution;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetGaussAzimuthResolution()
{
	return gaussAzimuthRes;
}

//----------------------------------------------------------------------------------------

void atipSetGaussElevationResolution(unsigned int resolution)
{
	ATIP_ASSERT(resolution >= 4);
	gaussElevationRes = resolution;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetGaussElevationResolution()
{
	return gaussElevationRes;
}

//----------------------------------------------------------------------------------------

void atipSetLinesContributionState(bool state)
{
	linesContribState = state;
}

//----------------------------------------------------------------------------------------

bool atipGetLinesContributionState()
{
	return linesContribState;
}

//----------------------------------------------------------------------------------------

void atipSetGaussWindowSize(unsigned int size)
{
	ATIP_ASSERT(((size & 1) == 1) && (size >= 3));
	gaussWindowSize = size;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetGaussWindowSize()
{
	return gaussWindowSize;
}

//----------------------------------------------------------------------------------------

void atipSetMaxLineDistance(float maxDist)
{
	ATIP_ASSERT((maxDist > 0.0f) && (maxDist < 1.0f));
	maxLineDistance = maxDist;
}

//----------------------------------------------------------------------------------------

float atipGetMaxLineDistance()
{
	return maxLineDistance;
}

//----------------------------------------------------------------------------------------

void atipSetInfiniteAngle(float angle)
{
	ATIP_ASSERT((angle > 0.0f) && (angle <= 1.57f));
	infiniteAngle = angle;
}

//----------------------------------------------------------------------------------------

float atipGetInfiniteAngle()
{
	return infiniteAngle;
}

//----------------------------------------------------------------------------------------

void atipSetMaxLineAngle(unsigned int angle)
{
	ATIP_ASSERT((angle > 0) && (angle < 90));
	maxLineAngle = angle;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetMaxLineAngle()
{
	return maxLineAngle;
}

//----------------------------------------------------------------------------------------

void atipSetComputingSpeed(unsigned int speed)
{
	ATIP_ASSERT(   (speed == ATIP_COMPSPEED_NORMAL)
				|| (speed == ATIP_COMPSPEED_FAST)
				|| (speed == ATIP_COMPSPEED_VERYFAST));
	computingSpeed = speed;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetComputingSpeed()
{
	return computingSpeed;
}

//----------------------------------------------------------------------------------------

void atipSetFOV(float angle)
{
	ATIP_ASSERT((angle >= 1.0f) && (angle <= 179.0f));
	fov = angle;
}

//----------------------------------------------------------------------------------------

float atipGetFOV()
{
	return fov;
}

//----------------------------------------------------------------------------------------

void atipSetMaximumTextureSize(unsigned int maxTexSize)
{
	ATIP_ASSERT(atipIsPowerOf2(maxTexSize) && (maxTexSize >= 64) && (maxTexSize <= 4096));
	maximumTextureSize = maxTexSize;
}

//----------------------------------------------------------------------------------------

unsigned int atipGetMaximumTextureSize()
{
	return maximumTextureSize;
}