/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_tip_maker.cpp
//! \author	BOULANGER Kevin
//! \date	02nd August 2004
//! \brief	Compute the TIP parameters (vertices, textures) from a calibrated camera


#include <atip_tip_maker.h>
#include <atip_options.h>
#include <tools/atip_private_tools.h>
#include <math.h>


void atipVerticesCoords(ATIPdata & atipData)
{
	// Unrotated vertices
	atipData.vertices[0 ] = - atipData.xLeft;
	atipData.vertices[1 ] = - atipData.yFloor;
	atipData.vertices[2 ] = - atipData.distance;
	atipData.vertices[3 ] =   atipData.xRight;
	atipData.vertices[4 ] = - atipData.yFloor;
	atipData.vertices[5 ] = - atipData.distance;
	atipData.vertices[6 ] =   atipData.xRight;
	atipData.vertices[7 ] =   atipData.yCeiling;
	atipData.vertices[8 ] = - atipData.distance;
	atipData.vertices[9 ] = - atipData.xLeft;
	atipData.vertices[10] =   atipData.yCeiling;
	atipData.vertices[11] = - atipData.distance;
	atipData.vertices[12] = - atipData.xLeft;
	atipData.vertices[13] = - atipData.yFloor;
	atipData.vertices[14] =   0.0f;
	atipData.vertices[15] =   atipData.xRight;
	atipData.vertices[16] = - atipData.yFloor;
	atipData.vertices[17] =   0.0f;
	atipData.vertices[18] =   atipData.xRight;
	atipData.vertices[19] =   atipData.yCeiling;
	atipData.vertices[20] =   0.0f;
	atipData.vertices[21] = - atipData.xLeft;
	atipData.vertices[22] =   atipData.yCeiling;
	atipData.vertices[23] =   0.0f;

	// Rotated vertices
	unsigned int v;
	for (v = 0; v < 8; v++)
	{
		atipData.rotVertices[v * 3    ] =  atipData.rot[0] * atipData.vertices[v * 3    ]
										 + atipData.rot[1] * atipData.vertices[v * 3 + 1]
										 + atipData.rot[2] * atipData.vertices[v * 3 + 2];
		atipData.rotVertices[v * 3 + 1] =  atipData.rot[3] * atipData.vertices[v * 3    ]
										 + atipData.rot[4] * atipData.vertices[v * 3 + 1]
										 + atipData.rot[5] * atipData.vertices[v * 3 + 2];
		atipData.rotVertices[v * 3 + 2] =  atipData.rot[6] * atipData.vertices[v * 3    ]
										 + atipData.rot[7] * atipData.vertices[v * 3 + 1]
										 + atipData.rot[8] * atipData.vertices[v * 3 + 2];
	}

	// Indices
	atipData.indices[0] = 0;	atipData.indices[1] = 1;
	atipData.indices[2] = 2;	atipData.indices[3] = 3;

	atipData.indices[4] = 4;	atipData.indices[5] = 0;
	atipData.indices[6] = 3;	atipData.indices[7] = 7;

	atipData.indices[8] = 1;	atipData.indices[9] = 5;
	atipData.indices[10] = 6;	atipData.indices[11] = 2;

	atipData.indices[12] = 1;	atipData.indices[13] = 0;
	atipData.indices[14] = 4;	atipData.indices[15] = 5;

	atipData.indices[16] = 3;	atipData.indices[17] = 2;
	atipData.indices[18] = 6;	atipData.indices[19] = 7;
}

//----------------------------------------------------------------------------------------

//! Compute the bilinear interpolation inside image with the coordinates (s, t)
//! \note If the coordinates are outside the image, a black pixel is written
//! \param color Destination array containing a RGB color
//! \param image Original image data (24 bits)
//! \param width Width of the image (unsigned int number)
//! \param fwidth Width of the image (float number)
//! \param fheight Height of the image (float number)
//! \param s S coordinate inside the image
//! \param t T coordinate inside the image
inline void BilinearInterp(unsigned char color[3], const unsigned char * image,
						   unsigned int width, float fwidth, float fheight,
						   float s, float t)
{
	ATIP_ASSERT(color && image);

	if ((s < 0.0f) || (s >= 1.0f) || (t < 0.0f) || (t >= 1.0f))
	{
		color[0] = color[1] = color[2] = 0;
		return;
	}

	t = 1.0f - t;

	float sf = floor(s * (fwidth - 2.0f));
	float tf = floor(t * (fheight - 2.0f));
	float sm = s * (fwidth - 2.0f) - sf;
	float tm = t * (fheight - 2.0f) - tf;
	float _sm = 1.0f - sm;
	float _tm = 1.0f - tm;
	unsigned int offset1 = ((unsigned int)tf * width + (unsigned int)sf) * 3;
	unsigned int offset2 = offset1 + 3;
	unsigned int offset3 = offset1 + width * 3;
	unsigned int offset4 = offset3 + 3;

	color[0] = (unsigned char)floor(
							  _sm * (  _tm * (float)image[offset1]
									 +  tm * (float)image[offset3])
							+  sm * (  _tm * (float)image[offset2]
									 +  tm * (float)image[offset4])
									);
	color[1] = (unsigned char)floor(
							  _sm * (  _tm * (float)image[offset1 + 1]
									 +  tm * (float)image[offset3 + 1])
							+  sm * (  _tm * (float)image[offset2 + 1]
									 +  tm * (float)image[offset4 + 1])
									);
	color[2] = (unsigned char)floor(
							  _sm * (  _tm * (float)image[offset1 + 2]
									 +  tm * (float)image[offset3 + 2])
							+  sm * (  _tm * (float)image[offset2 + 2]
									 +  tm * (float)image[offset4 + 2])
									);
}


//----------------------------------------------------------------------------------------

bool atipComputeTextures(ATIPimage * textures[5], ATIPimage & image,
						 const float rotVertices[24], const unsigned int indices[20],
						 float focal)
{
	ATIP_ASSERT(textures && rotVertices && indices && (focal > 0.0f));
	if (!textures || !rotVertices || !indices || (focal <= 0.0f))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Conversion of the original image if necessary (in 24 bits RGB)
	ATIPimage srcImage(image);
	srcImage.ConvertToRGB();

	// Do the projection of the 8 vertices to the screen plane
	// (if z >= -focal, invert the z component)
	float screenVertices[16];
	unsigned int v;
	for (v = 0; v < 8; v++)
	{
		if (rotVertices[v * 3 + 2] < focal)
		{
			screenVertices[ v << 1     ] = - focal * rotVertices[v * 3    ]
												/ rotVertices[v * 3 + 2];
			screenVertices[(v << 1) + 1] = - focal * rotVertices[v * 3 + 1]
												/ rotVertices[v * 3 + 2];
		}
		else
		{
			screenVertices[ v << 1     ] = focal * rotVertices[v * 3    ]
											/ (rotVertices[v * 3 + 2] + 2.0f * focal);
			screenVertices[(v << 1) + 1] = focal * rotVertices[v * 3 + 1]
											/ (rotVertices[v * 3 + 2] + 2.0f * focal);
		}
	}

	// List the vertices which are inside the screen
	bool inScreenVertices[8];
	for (v = 0; v < 8; v++)
		inScreenVertices[v] =    (screenVertices[ v << 1     ] >= -1.0f)
							  && (screenVertices[ v << 1     ] <=  1.0f)
							  && (screenVertices[(v << 1) + 1] >= -1.0f)
							  && (screenVertices[(v << 1) + 1] <=  1.0f);

	// List the polygons which are inside the screen (at least a part)
	bool used[5];
	unsigned int p, p2;
	for (p = 0; p < 5; p++)
	{
		p2 = p << 2;
		used[p] =    (inScreenVertices[indices[p2    ]])
				  || (inScreenVertices[indices[p2 + 1]])
				  || (inScreenVertices[indices[p2 + 2]])
				  || (inScreenVertices[indices[p2 + 3]]);
	}

	// Minimum and maximum X and Y of the polygons
	float minX[5];
	float maxX[5];
	float minY[5];
	float maxY[5];
	for (p = 0; p < 5; p++)
		if (used[p])
		{
			p2 = p << 2;
			minX[p] = MIN(screenVertices[ indices[p2    ] << 1],
					  MIN(screenVertices[ indices[p2 + 1] << 1],
					  MIN(screenVertices[ indices[p2 + 2] << 1],
						  screenVertices[ indices[p2 + 3] << 1])));
			maxX[p] = MAX(screenVertices[ indices[p2    ] << 1],
					  MAX(screenVertices[ indices[p2 + 1] << 1],
					  MAX(screenVertices[ indices[p2 + 2] << 1],
						  screenVertices[ indices[p2 + 3] << 1])));
			minY[p] = MIN(screenVertices[(indices[p2    ] << 1) + 1],
					  MIN(screenVertices[(indices[p2 + 1] << 1) + 1],
					  MIN(screenVertices[(indices[p2 + 2] << 1) + 1],
						  screenVertices[(indices[p2 + 3] << 1) + 1])));
			maxY[p] = MAX(screenVertices[(indices[p2    ] << 1) + 1],
					  MAX(screenVertices[(indices[p2 + 1] << 1) + 1],
					  MAX(screenVertices[(indices[p2 + 2] << 1) + 1],
						  screenVertices[(indices[p2 + 3] << 1) + 1])));
		}

	// Compute the size of the textures, and then the data
	unsigned int width[5], height[5];
	unsigned char * data[5] = { 0 };
	float x0, y0, z0, v0x, v0y, v0z, v1x, v1y, v1z;
	unsigned int widthIm;
	float fwidthIm, fheightIm, ratio;
	unsigned int u;
	float s, t, x, y, z;
	float screenX, screenY;
	for (p = 0; p < 5; p++)
		if (used[p])
		{
			// Compute the size of the texture
			width[p] = atipNextPowerOf2((unsigned int)ceil((maxX[p] - minX[p]) / 2.0f
														* (float)srcImage.GetWidth()));
			if (width[p] < 4) width[p] = 4;
			if (width[p] > atipGetMaximumTextureSize())
				width[p] = atipGetMaximumTextureSize();
			height[p] = atipNextPowerOf2((unsigned int)ceil((maxY[p] - minY[p]) / 2.0f
														* (float)srcImage.GetHeight()));
			if (height[p] < 4) height[p] = 4;
			if (height[p] > atipGetMaximumTextureSize())
				height[p] = atipGetMaximumTextureSize();

			// Compute the base vectors of the current face
			p2 = p << 2;
			x0 =  rotVertices[indices[p2    ] * 3    ];
			y0 =  rotVertices[indices[p2    ] * 3 + 1];
			z0 =  rotVertices[indices[p2    ] * 3 + 2];
			v0x = rotVertices[indices[p2 + 1] * 3    ] - x0;
			v0y = rotVertices[indices[p2 + 1] * 3 + 1] - y0;
			v0z = rotVertices[indices[p2 + 1] * 3 + 2] - z0;
			v1x = rotVertices[indices[p2 + 3] * 3    ] - x0;
			v1y = rotVertices[indices[p2 + 3] * 3 + 1] - y0;
			v1z = rotVertices[indices[p2 + 3] * 3 + 2] - z0;

			ATIP_NEW_ARRAY(unsigned char, data[p], width[p] * height[p] * 3);

			widthIm = srcImage.GetWidth();
			fwidthIm = (float)widthIm;
			fheightIm = (float)srcImage.GetHeight();
			ratio = fheightIm / fwidthIm;

			// For each texel
			for (v = 0; v < height[p]; v++)
				for (u = 0; u < width[p]; u++)
				{
					// Texel coordinates
					s = (float)u / (float)width[p];
					t = (float)v / (float)height[p];
					
					// Real coordinates
					x = x0 + s * v0x + t * v1x;
					y = y0 + s * v0y + t * v1y;
					z = z0 + s * v0z + t * v1z;
					
					// Projection to the screen plane
					if (z < focal)
					{
						screenX = - focal * x * ratio / z;
						screenY = - focal * y / z;
					}
					else
					{
						screenX = focal * x * ratio / (z + 2.0f * focal);
						screenY = focal * y / (z + 2.0f * focal);
					}

					// Bilinear interpolation of the original image
					BilinearInterp(data[p] + (v * width[p] + u) * 3,
								   (unsigned char *)srcImage.GetImage(),
								   widthIm, fwidthIm, fheightIm,
								   (screenX + 1.0f) / 2.0f, (screenY + 1.0f) / 2.0f);
				}
		}

	// Save the textures to ATIPimage objects
	for (p = 0; p < 5; p++)
		if (used[p])
			ATIP_NEW(ATIPimage(width[p], height[p], 24, data[p], false), textures[p]);

	return true;
}