/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_primitives.cpp
//! \author	BOULANGER Kevin
//! \date	01st July 2004
//! \brief	Drawing primitives for images (lines, ...)


#include <tools/atip_primitives.h>
#include <tools/atip_private_tools.h>
#include <memory.h>
#include <stdlib.h>


void atipDrawSegment8(unsigned char * image, unsigned int width, unsigned int height,
					  unsigned int x0, unsigned int y0,
					  unsigned int x1, unsigned int y1,
					  unsigned char color)
{
	ATIP_ASSERT(image && (width > 0) && (height > 0));
	if (!image || (width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return;
	}

	// For vertical lines
	if (x0 == x1)
		if (y0 <= y1)
		{
			// From top to bottom

			// Verify the bounds of the segment
			if (y1 >= height) y1 = height - 1;

			// Draw the segment
			for (unsigned int i = y0, p = y0 * width + x0; i <= y1; i++, p += width)
				image[p] = color;
		}
		else
		{
			// From bottom to top

			// Verifiy the bounds of the segment
			if (y0 >= height) y0 = height - 1;

			// Draw the segment
			for (unsigned int i = y1, p = y1 * width + x0; i <= y0; i++, p += width)
				image[p] = color;
		}

	// For horizontal lines
	else if (y0 == y1)
		if (x0 <= x1)
		{
			// From left to right

			// Verify the bounds of the segment
			if (x1 >= width) x1 = width - 1;

			// Draw the segment
			memset(image + y0 * width + x0, color, x1 - x0 + 1);
		}
		else
		{
			// From right to left

			// Verify the bounds of the segment
			if (x0 >= width) x0 = width - 1;

			// Draw the segment
			memset(image + y0 * width + x1, color, x0 - x1 + 1);
		}

	// For diagonal lines
	else
	{
		int x = (int)x0;
		int y = (int)y0;
		int dx = abs((int)x1 - (int)x0);
		int dy = abs((int)y1 - (int)y0);
		int sx = (x1 > x0) ? 1 : -1;
		int sy = (y1 > y0) ? 1 : -1;
		int eps = 0;

		// Swap x and y if necessary
		if (dy > dx)
		{
			int t;
			t = x;		x = y;		y = t;
			t = dx;		dx = dy;	dy = t;
			t = sx;		sx = sy;	sy = t;

			// Draw the line
			for (int i = 0; i < dx; i++)
			{
				image[x * width + y] = color;
				eps += dy;
				if ((eps << 1) >= dx)
				{
					y += sy;
					eps -= dx;
				}
				x += sx;
			}
		}
		else
			// Draw the line
			for (int i = 0; i < dx; i++)
			{
				image[y * width + x] = color;
				eps += dy;
				if ((eps << 1) >= dx)
				{
					y += sy;
					eps -= dx;
				}
				x += sx;
			}
	}
}

//----------------------------------------------------------------------------------------

void atipDrawSegment24(unsigned char * image, unsigned int width, unsigned int height,
					   unsigned int x0, unsigned int y0,
					   unsigned int x1, unsigned int y1,
					   unsigned char red, unsigned char green, unsigned char blue)
{
	ATIP_ASSERT(image && (width > 0) && (height > 0));
	if (!image || (width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return;
	}

	// For vertical lines
	if (x0 == x1)
		if (y0 <= y1)
		{
			// From top to bottom

			// Verify the bounds of the segment
			if (y1 >= height) y1 = height - 1;

			// Draw the segment
			for (unsigned int i = y0, p = (y0 * width + x0) * 3;
				 i <= y1; i++, p += width * 3)
			{
				image[p    ] = red;
				image[p + 1] = green;
				image[p + 2] = blue;
			}
		}
		else
		{
			// From bottom to top

			// Verifiy the bounds of the segment
			if (y0 >= height) y0 = height - 1;

			// Draw the segment
			for (unsigned int i = y1, p = (y1 * width + x0) * 3;
				 i <= y0; i++, p += width * 3)
			{
				image[p    ] = red;
				image[p + 1] = green;
				image[p + 2] = blue;
			}
		}

	// For horizontal lines
	else if (y0 == y1)
		if (x0 <= x1)
		{
			// From left to right

			// Verify the bounds of the segment
			if (x1 >= width) x1 = width - 1;

			// Draw the segment
			for (unsigned int i = x0, p = (y0 * width + x0) * 3;
				 i <= x1; i++, p += 3)
			{
				image[p    ] = red;
				image[p + 1] = green;
				image[p + 2] = blue;
			}
		}
		else
		{
			// From right to left

			// Verify the bounds of the segment
			if (x0 >= width) x0 = width - 1;

			// Draw the segment
			for (unsigned int i = x1, p = (y0 * width + x1) * 3;
				 i <= x0; i++, p += 3)
			{
				image[p    ] = red;
				image[p + 1] = green;
				image[p + 2] = blue;
			}
		}

	// For diagonal lines
	else
	{
		int x = (int)x0;
		int y = (int)y0;
		int dx = abs((int)x1 - (int)x0);
		int dy = abs((int)y1 - (int)y0);
		int sx = (x1 > x0) ? 1 : -1;
		int sy = (y1 > y0) ? 1 : -1;
		int eps = 0;
		int offset = 0;

		// Swap x and y if necessary
		if (dy > dx)
		{
			int t;
			t = x;		x = y;		y = t;
			t = dx;		dx = dy;	dy = t;
			t = sx;		sx = sy;	sy = t;

			// Draw the line
			for (int i = 0; i < dx; i++)
			{
				offset = (x * width + y) * 3;
				image[offset    ] = red;
				image[offset + 1] = green;
				image[offset + 2] = blue;
				eps += dy;
				if ((eps << 1) >= dx)
				{
					y += sy;
					eps -= dx;
				}
				x += sx;
			}
		}
		else
			// Draw the line
			for (int i = 0; i < dx; i++)
			{
				offset = (y * width + x) * 3;
				image[offset    ] = red;
				image[offset + 1] = green;
				image[offset + 2] = blue;
				eps += dy;
				if ((eps << 1) >= dx)
				{
					y += sy;
					eps -= dx;
				}
				x += sx;
			}
	}
}
