/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_camera_calib.cpp
//! \author	BOULANGER Kevin
//! \date	29th July 2004
//! \brief	Camera calibration from vanishing points


#include <van_point/atip_camera_calib.h>
#include <tools/atip_private_tools.h>
#include <atip_options.h>
#include <math.h>


//! Normalization of a vector
//! \param v Vector to normalize
inline void Normalize(float v[3])
{
	float recNorm = 1.0f / (float)sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
	v[0] *= recNorm;
	v[1] *= recNorm;
	v[2] *= recNorm;
}

//----------------------------------------------------------------------------------------

//! Normalization of a 3x3 matrix column
//! \param v Vector to normalize, column taken from a 3x3 matrix
inline void NormalizeColumn(float * v)
{
	float recNorm = 1.0f / (float)sqrt(v[0] * v[0] + v[3] * v[3] + v[6] * v[6]);
	v[0] *= recNorm;
	v[3] *= recNorm;
	v[6] *= recNorm;
}

//----------------------------------------------------------------------------------------

//! Dot product between two vectors in 3 dimensions
//! \param v1 First vector
//! \param v2 Second vector
//! \return Dot product
inline float Dot3(const float v1[3], const float v2[3])
{
	return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}

//----------------------------------------------------------------------------------------

//! Cross product between two vectors in 3 dimensions
//! \param v1 First vector
//! \param v2 Second vector
//! \param c Resulting cross product
inline void CrossProduct(float c[3], const float v1[3], const float v2[3])
{
	c[0] = v1[1] * v2[2] - v1[2] * v2[1];
	c[1] = v1[2] * v2[0] - v1[0] * v2[2];
	c[2] = v1[0] * v2[1] - v1[1] * v2[0];
}

//----------------------------------------------------------------------------------------

//! Cross product between two columns of a 3x3 matrix
//! \param v1 First vector, column taken from a 3x3 matrix
//! \param v2 Second vector, column taken from a 3x3 matrix
//! \param c Resulting cross product, stored inside a column of a 3x3 matrix
inline void CrossProductColumn(float * c, const float * v1, const float * v2)
{
	c[0] = v1[3] * v2[6] - v1[6] * v2[3];
	c[3] = v1[6] * v2[0] - v1[0] * v2[6];
	c[6] = v1[0] * v2[3] - v1[3] * v2[0];
}

//----------------------------------------------------------------------------------------

bool atipCalibrateCamera(float & focal, float & fov, float rot[9],
						 float vanPoints[12], bool infiniteVanPoints[3],
						 unsigned int width, unsigned int height)
{
	ATIP_ASSERT(vanPoints && infiniteVanPoints && (width > 0) && (height > 0));
	if (!vanPoints || !infiniteVanPoints || (width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Compute the number of finite points
	// (the finite point should be at the beginning of the array)
	unsigned int nFiniteVanPoints = 0;
	if (!infiniteVanPoints[0])
	{
		nFiniteVanPoints++;
		if (!infiniteVanPoints[1])
		{
			nFiniteVanPoints++;
			if (!infiniteVanPoints[2])
				nFiniteVanPoints++;
		}
	}
	if (nFiniteVanPoints == 0)
	{
		atipSetError(ATIP_NO_FINITE_VAN_POINT);
		return false;
	}

	// In function of the number of finite vanishing points
	switch (nFiniteVanPoints)
	{
		// 1 finite vanishing point, 2 infinite ones
		case 1:
			{
// IN ORDER TO USE THE NEW METHOD (AUTOMATIC FOCAL CALCULATION), DEFINE ATIP_NEW_METHOD,
// BUT NOT FOR ORTHOGONAL INFINITE VANISHING POINTS
//#define ATIP_NEW_METHOD

#ifndef ATIP_NEW_METHOD
				// Compute the focal length
				fov = atipGetFOV();
				focal = float(1.0 / tan(fov * M_2PI / 720.0));

				// Compute the rotation matrix
				// Coordinates of the finite vanishing points F
				float fx = (vanPoints[2] * 2.0f - (float)width) / (float)height;
				float fy = - (vanPoints[3] * 2.0f - (float)height) / (float)height;

				// Norm of the vector OF (from the camera to the finite vanishing point)
				float of = (float)sqrt(fx * fx + fy * fy + focal * focal);

				// Vectors of the objects base
				rot[2] = - fx / of;
				rot[5] = - fy / of;
				rot[8] = focal / of;

				float fvx, fvy;
				if (fabs(vanPoints[7]) < fabs(vanPoints[11]))
				{
					// If the second vanishing point (infinite)
					// is more horizontal than the third one
					
					// Copy the vertical direction
					if (vanPoints[11] < 0.0f)
					{
						fvx = - vanPoints[10];
						fvy = - vanPoints[11];
					}
					else
					{
						fvx = vanPoints[10];
						fvy = vanPoints[11];
					}
				}
				else
				{
					// If the third vanishing point (infinite)
					// is more horizontal than the second one
					
					// Copy the vertical direction
					if (vanPoints[7] < 0.0f)
					{
						fvx = - vanPoints[6];
						fvy = - vanPoints[7];
					}
					else
					{
						fvx = vanPoints[6];
						fvy = vanPoints[7];
					}
				}

				// Compute the vertical direction (the parallel projection of this
				// direction must give the vanishing point direction)
				rot[7] = - (fvx * rot[2] + fvy * rot[5]) / rot[8];
				float recNorm = 1.0f / (float)sqrt(
									fvx * fvx + fvy * fvy + rot[7] * rot[7]);
				rot[1] = fvx * recNorm;
				rot[4] = fvy * recNorm;
				rot[7] *= recNorm;

				// Compute the second direction (is becoming a finite vanishing point)
				CrossProductColumn(rot, rot + 1, rot + 2);
#else
				// Coordinates of the finite vanishing points F
				float fx = (vanPoints[2] * 2.0f - (float)width) / (float)height;
				float fy = - (vanPoints[3] * 2.0f - (float)height) / (float)height;

				// Directions of the infinite vanishing points
				float i1x, i1y, i2x, i2y;
				if (fabs(vanPoints[7]) < fabs(vanPoints[11]))
				{
					// If the second vanishing point (infinite)
					// is more horizontal than the third one
					
					// Copy the vertical direction
					if (vanPoints[11] < 0.0f)
					{
						i1x = - vanPoints[10];
						i1y = - vanPoints[11];
					}
					else
					{
						i1x = vanPoints[10];
						i1y = vanPoints[11];
					}

					// Copy the horizontal direction
					if (vanPoints[6] < 0.0f)
					{
						i2x = - vanPoints[6];
						i2y = - vanPoints[7];
					}
					else
					{
						i2x = vanPoints[6];
						i2y = vanPoints[7];
					}
				}
				else
				{
					// If the third vanishing point (infinite)
					// is more horizontal than the second one
					
					// Copy the vertical direction
					if (vanPoints[7] < 0.0f)
					{
						i1x = - vanPoints[6];
						i1y = - vanPoints[7];
					}
					else
					{
						i1x = vanPoints[6];
						i1y = vanPoints[7];
					}

					// Copy the horizontal direction
					if (vanPoints[10] < 0.0f)
					{
						i2x = - vanPoints[10];
						i2y = - vanPoints[11];
					}
					else
					{
						i2x = vanPoints[10];
						i2y = vanPoints[11];
					}
				}

				// Compute the focal length and the fov
				focal = - ( (i1x * fx + i1y * fy) * (i2x * fx + i2y * fy) )
							/ (i1x * i2x + i1y * i2y);
				//ATIP_ASSERT(focal >= 0.0f);
				focal = (float)sqrt((double)fabs(focal));
				fov = float(atan(1.0f / focal) * 360.0 / M_2PI) * 2.0f;

				// Compute the rotation matrix
				rot[0] = i2x;
				rot[3] = i2y;
				rot[6] = (i2x * fx + i2y * fy) / focal;
				NormalizeColumn(rot);
				rot[1] = i1x;
				rot[4] = i1y;
				rot[7] = (i1x * fx + i1y * fy) / focal;
				NormalizeColumn(rot + 1);
				rot[2] = -fx;
				rot[5] = -fy;
				rot[8] = focal;
				NormalizeColumn(rot + 2);
#endif	// ATIP_NEW_METHOD
			}
			break;

		// 2 finite vanishing points, 1 infinite one
		case 2:
			{
				// Compute the focal length and the field-of-view

				// Coordinates of the finite vanishing points F1 and F2
				// inside the camera base
				float f1x = (vanPoints[2] * 2.0f - (float)width) / (float)height;
				float f1y = - (vanPoints[3] * 2.0f - (float)height) / (float)height;
				float f2x = (vanPoints[6] * 2.0f - (float)width) / (float)height;
				float f2y = - (vanPoints[7] * 2.0f - (float)height) / (float)height;

				// Vector between the two vanishing points F1F2
				/*float f1f2x = f2x - f1x;
				float f1f2y = f2y - f1y;

				// Compute the vector F1P2 from the first vanishing point to the
				// orthogonal projection of the principal point to F1F2
				float dot = - (f1x * f1f2x + f1y * f1f2y)
							/ (f1f2x * f1f2x + f1f2y * f1f2y);
				float f1p2x = dot * f1f2x;
				float f1p2y = dot * f1f2y;

				// Compute the norms between the two vanishing points and P2
				float f1p2 = (float)sqrt(f1p2x * f1p2x + f1p2y * f1p2y);
				float p2f2 = (float)sqrt(f1f2x * f1f2x + f1f2y * f1f2y) - f1p2;

				// Coordinates of the orthogonal projection of the principal point to F1F2
				float p2x = f1x + f1p2x;
				float p2y = f1y + f1p2y;

				// Compute the focal length
				focal = (float)sqrt(f1p2 * p2f2 - p2x * p2x - p2y * p2y);*/
				focal = (float)sqrt(fabs(f1x * f2x + f1y * f2y));

				// Compute the field-of-view
				fov = float(atan(1.0f / focal) * 360.0 / M_2PI) * 2.0f;

				// Compute the rotation matrix

				// Norm of the vectors OF1 and OF2
				// (from the camera to the vanishing points)
				float of1 = (float)sqrt(f1x * f1x + f1y * f1y + focal * focal);
				float of2 = (float)sqrt(f2x * f2x + f2y * f2y + focal * focal);

				// Rotation matrix
				rot[0] = f1x / of1;
				rot[3] = f1y / of1;
				rot[6] = - focal / of1;
				rot[2] = - f2x / of2;
				rot[5] = - f2y / of2;
				rot[8] = focal / of2;
				CrossProductColumn(rot + 1, rot + 2, rot);

				// If the vertical direction is to the y < 0, rotate the base
				if (rot[4] < 0.0f)
				{
					float t;
					t = rot[0];		rot[0] = - rot[2];	rot[2] = - t;
					t = rot[3];		rot[3] = - rot[5];	rot[5] = - t;
					t = rot[6];		rot[6] = - rot[8];	rot[8] = - t;
					rot[1] = - rot[1];
					rot[4] = - rot[4];
					rot[7] = - rot[7];
				}
			}
			break;

		// 3 finite vanishing points
		case 3:
			{
				//! \todo If the 3 vanishing points represent a non-orthogonal basis
				//! (with a threshold), it is too hard to find a correct orthogonal basis

				// Coordinates of the finite vanishing points F0, F1 and F2
				// inside the camera base
				float vp[3][2];
                vp[0][0] = (vanPoints[2] * 2.0f - (float)width) / (float)height;
				vp[0][1] = - (vanPoints[3] * 2.0f - (float)height) / (float)height;
				vp[1][0] = (vanPoints[6] * 2.0f - (float)width) / (float)height;
				vp[1][1] = - (vanPoints[7] * 2.0f - (float)height) / (float)height;
				vp[2][0] = (vanPoints[10] * 2.0f - (float)width) / (float)height;
				vp[2][1] = - (vanPoints[11] * 2.0f - (float)height) / (float)height;

				// Compute the three possible focal lengths
				float focals[3] = {
					(float)sqrt(-(vp[0][0] * vp[1][0] + vp[0][1] * vp[1][1])),
					(float)sqrt(-(vp[1][0] * vp[2][0] + vp[1][1] * vp[2][1])),
					(float)sqrt(-(vp[2][0] * vp[0][0] + vp[2][1] * vp[0][1]))	};

				// Compute a quality factor for each focal length
				float u[3], v[3], w[3], f[3];
				float quality[3];
				unsigned int i, i1, i2;
				for (i = 0; i < 3; i++)
				{
					i1 = (i + 1) % 3;
					i2 = (i + 2) % 3;

					// Take two vanishing points directions
					u[i] = vp[i][0];	u[1] = vp[i][1];		u[2] = - focals[i];
					v[i] = vp[i1][0];	v[1] = vp[i1][1];		v[2] = - focals[i];

					// Compute a third direction orthogonal to the two firsts
					CrossProduct(w, u, v);
					Normalize(w);
					f[0] = vp[i2][0];	f[1] = vp[i2][1];		f[2] = - focals[i];
					Normalize(f);
					quality[i] = fabs(Dot3(w, f));
				}

				// Look for the maximum quality factor
				float maxQual = quality[0];
				unsigned int indexMax = 0;
				if (quality[1] > maxQual)
				{
					maxQual = quality[1];
					indexMax = 1;
				}
				if (quality[2] > maxQual)
				{
					maxQual = quality[2];
					indexMax = 2;
				}

				// The focal length and the fov are now known
				focal = focals[indexMax];
				fov = float(atan(1.0f / focal) * 360.0 / M_2PI) * 2.0f;

				// Compute the rotation matrix
				rot[0] = vp[indexMax][0];
				rot[3] = vp[indexMax][1];
				rot[6] = - focal;
				NormalizeColumn(rot);
				rot[1] = vp[(indexMax + 1) % 3][0];
				rot[4] = vp[(indexMax + 1) % 3][1];
				rot[7] = - focal;
				NormalizeColumn(rot + 1);
				CrossProductColumn(rot + 2, rot, rot + 1);
				NormalizeColumn(rot + 2);

				// Look for the direction with the maximum Y component and place
				// this direction in the second matrix column
				float t;
				if (fabs(rot[3]) > fabs(rot[4]))
				{
					t = rot[0];		rot[0] = rot[1];	rot[1] = t;
					t = rot[3];		rot[3] = rot[4];	rot[4] = t;
					t = rot[6];		rot[6] = rot[7];	rot[7] = t;
				}
				if (fabs(rot[5]) > fabs(rot[4]))
				{
					t = rot[2];		rot[2] = rot[1];	rot[1] = t;
					t = rot[5];		rot[5] = rot[4];	rot[4] = t;
					t = rot[8];		rot[8] = rot[7];	rot[7] = t;
				}

				// If the Y component of the second column is to the y < 0,
				// take the opposite of this vector
				if (rot[4] < 0.0f)
				{
					rot[1] = - rot[1];
					rot[4] = - rot[4];
					rot[7] = - rot[7];
				}

				// If the basis in indirect, swap the first and third vectors
				float vert[9];
				CrossProductColumn(vert, rot, rot + 6);
				if (vert[0] * rot[1] + vert[3] * rot[4] + vert[6] * rot[7] < 0.0f)
				{
					t = rot[0];		rot[0] = rot[2];	rot[2] = t;
					t = rot[3];		rot[3] = rot[5];	rot[5] = t;
					t = rot[6];		rot[6] = rot[8];	rot[8] = t;
				}

				// If the Z component of the third column is to the z < 0,
				// take the opposite of the first and the third vectors
				if (rot[8] < 0.0f)
				{
					rot[0] = - rot[0];
					rot[3] = - rot[3];
					rot[6] = - rot[6];
					rot[2] = - rot[2];
					rot[5] = - rot[5];
					rot[8] = - rot[8];
				}
			}
			break;
	}

	return true;
}
