/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_filter.cpp
//! \author	BOULANGER Kevin
//! \date	29th June 2004
//! \brief	Image filtering functions


#include <van_point/atip_filter.h>
#include <tools/atip_private_tools.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>


bool atipConvertToGrayscale(unsigned char * grayImage, const unsigned char * colorImage,
							unsigned int width, unsigned int height, unsigned char bpp)
{
	// Only if the image is coded in 24 or 32 bits
	ATIP_ASSERT((bpp == 24) || (bpp == 32));
	if ((bpp != 24) && (bpp != 32))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int nPixels = width * height;
	unsigned int nBytesPerPixel = bpp / 8;
	unsigned int i, p;
	for (i = 0, p = 0; i < nPixels; i++, p += nBytesPerPixel)
		grayImage[i] = (unsigned char)floor(  (float)colorImage[p] * 0.299f
											  + (float)colorImage[p + 1] * 0.587f
											  + (float)colorImage[p + 2] * 0.114f);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipConvertImageToFloats(float * dstImage, const unsigned char * srcImage,
							  unsigned int width, unsigned int height)
{
	ATIP_ASSERT(dstImage && srcImage && width && height);
	if (!dstImage || !srcImage || !width || !height)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int nPixels = width * height;
	for (unsigned int p = 0; p < nPixels; p++)
		dstImage[p] = (float)srcImage[p];

	return true;
}

//----------------------------------------------------------------------------------------

bool atipReduceImage2_8(unsigned char * & smallImage,
						unsigned int & smallWidth, unsigned int & smallHeight,
						const unsigned char * bigImage,
						unsigned int bigWidth, unsigned int bigHeight)
{
	ATIP_ASSERT(bigImage && (bigWidth > 0) && (bigHeight > 0));
	if (!bigImage || (bigWidth == 0) || (bigHeight == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// New size
	unsigned int bWidth = bigWidth;
	unsigned int bHeight = bigHeight;
	smallWidth = (bWidth + 1) >> 1;
	smallHeight = (bHeight + 1) >> 1;

	// Allocate the new image
	ATIP_NEW_ARRAY(unsigned char, smallImage, smallWidth * smallHeight);

	// Fill the small image with the means of the 2x2 blocks of the big image
	unsigned int i, j, p, p2;

	// If the width and the height are even numbers
	if (((bWidth & 1) == 0) && ((bHeight & 1) == 0))
		for (i = 0, p = 0, p2 = 0; i < smallHeight; i++, p2 += bWidth)
			for (j = 0; j < smallWidth; j++, p++, p2 += 2)
				smallImage[p] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + bWidth]
					+ (unsigned short)bigImage[p2 + bWidth + 1]) >> 2);

	// If the width is an odd number and the height an even number
	else if ((bWidth & 1) && ((bHeight & 1) == 0))
		for (i = 0, p = 0, p2 = 0; i < smallHeight; i++, p2 += bWidth)
		{
			for (j = 0; j < smallWidth - 1; j++, p++, p2 += 2)
				smallImage[p] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + bWidth]
					+ (unsigned short)bigImage[p2 + bWidth + 1]) >> 2);
			smallImage[p++] = (unsigned char)((
					  (unsigned short)bigImage[p2]
				    + (unsigned short)bigImage[p2 + bWidth]) >> 1);
			p2 += 2;
		}

	// If the width is an even number and the height an odd number
	else if (((bWidth & 1) == 0) && (bHeight & 1))
	{
		for (i = 0, p = 0, p2 = 0; i < smallHeight - 1; i++, p2 += bWidth)
			for (j = 0; j < smallWidth; j++, p++, p2 += 2)
				smallImage[p] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + bWidth]
					+ (unsigned short)bigImage[p2 + bWidth + 1]) >> 2);
		for (j = 0; j < smallWidth; j++, p++, p2 += 2)
			smallImage[p] = (unsigned char)((
				  (unsigned short)bigImage[p2]
				+ (unsigned short)bigImage[p2 + 1]) >> 1);
	}

	// If the width and the height are odd numbers
	else if ((bWidth & 1) && (bHeight & 1))
	{
		for (i = 0, p = 0, p2 = 0; i < smallHeight - 1; i++, p2 += bWidth)
		{
			for (j = 0; j < smallWidth - 1; j++, p++, p2 += 2)
				smallImage[p] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + bWidth]
					+ (unsigned short)bigImage[p2 + bWidth + 1]) >> 2);
			smallImage[p++] = (unsigned char)((
					  (unsigned short)bigImage[p2]
				    + (unsigned short)bigImage[p2 + bWidth]) >> 1);
			p2 += 2;
		}
		for (j = 0; j < smallWidth - 1; j++, p++, p2 += 2)
			smallImage[p] = (unsigned char)((
				  (unsigned short)bigImage[p2]
				+ (unsigned short)bigImage[p2 + 1]) >> 1);
		smallImage[p] = bigImage[p2];
	}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipReduceImage2_24(unsigned char * & smallImage,
						 unsigned int & smallWidth, unsigned int & smallHeight,
						 const unsigned char * bigImage,
						 unsigned int bigWidth, unsigned int bigHeight)
{
	ATIP_ASSERT(bigImage && (bigWidth > 0) && (bigHeight > 0));
	if (!bigImage || (bigWidth == 0) || (bigHeight == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// New size
	unsigned int bWidth = bigWidth;
	unsigned int bHeight = bigHeight;
	smallWidth = (bWidth + 1) >> 1;
	smallHeight = (bHeight + 1) >> 1;

	// Allocate the new image
	ATIP_NEW_ARRAY(unsigned char, smallImage, smallWidth * smallHeight * 3);

	// Fill the small image with the means of the 2x2 blocks of the big image
	unsigned int i, j, p, p2;
	unsigned int bWidth3 = bWidth * 3;

	// If the width and the height are even numbers
	if (((bWidth & 1) == 0) && ((bHeight & 1) == 0))
		for (i = 0, p = 0, p2 = 0; i < smallHeight; i++, p2 += bWidth3)
			for (j = 0; j < smallWidth; j++, p += 3, p2 += 6)
			{
				smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
					+ (unsigned short)bigImage[p2 + 3]
					+ (unsigned short)bigImage[p2 + bWidth3]
					+ (unsigned short)bigImage[p2 + bWidth3 + 3]) >> 2);
				smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]
					+ (unsigned short)bigImage[p2 + bWidth3 + 1]
					+ (unsigned short)bigImage[p2 + bWidth3 + 4]) >> 2);
				smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]
					+ (unsigned short)bigImage[p2 + bWidth3 + 2]
					+ (unsigned short)bigImage[p2 + bWidth3 + 5]) >> 2);
			}

	// If the width is an odd number and the height an even number
	else if ((bWidth & 1) && ((bHeight & 1) == 0))
		for (i = 0, p = 0, p2 = 0; i < smallHeight; i++, p2 += bWidth3)
		{
			for (j = 0; j < smallWidth - 1; j++, p += 3, p2 += 6)
			{
				smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
					+ (unsigned short)bigImage[p2 + 3]
					+ (unsigned short)bigImage[p2 + bWidth3]
					+ (unsigned short)bigImage[p2 + bWidth3 + 3]) >> 2);
				smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]
					+ (unsigned short)bigImage[p2 + bWidth3 + 1]
					+ (unsigned short)bigImage[p2 + bWidth3 + 4]) >> 2);
				smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]
					+ (unsigned short)bigImage[p2 + bWidth3 + 2]
					+ (unsigned short)bigImage[p2 + bWidth3 + 5]) >> 2);
			}					
			smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
				    + (unsigned short)bigImage[p2 + bWidth3]) >> 1);
			smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
				    + (unsigned short)bigImage[p2 + bWidth3 + 1]) >> 1);
			smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
				    + (unsigned short)bigImage[p2 + bWidth3 + 2]) >> 1);
			p += 3;
			p2 += 6;
		}

	// If the width is an even number and the height an odd number
	else if (((bWidth & 1) == 0) && (bHeight & 1))
	{
		for (i = 0, p = 0, p2 = 0; i < smallHeight - 1; i++, p2 += bWidth3)
			for (j = 0; j < smallWidth; j++, p += 3, p2 += 6)
			{
				smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
					+ (unsigned short)bigImage[p2 + 3]
					+ (unsigned short)bigImage[p2 + bWidth3]
					+ (unsigned short)bigImage[p2 + bWidth3 + 3]) >> 2);
				smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]
					+ (unsigned short)bigImage[p2 + bWidth3 + 1]
					+ (unsigned short)bigImage[p2 + bWidth3 + 4]) >> 2);
				smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]
					+ (unsigned short)bigImage[p2 + bWidth3 + 2]
					+ (unsigned short)bigImage[p2 + bWidth3 + 5]) >> 2);
			}
		for (j = 0; j < smallWidth; j++, p += 3, p2 += 6)
		{
			smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 3]) >> 1);
			smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]) >> 1);
			smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]) >> 1);
		}
	}

	// If the width and the height are odd numbers
	else if ((bWidth & 1) && (bHeight & 1))
	{
		for (i = 0, p = 0, p2 = 0; i < smallHeight - 1; i++, p2 += bWidth)
		{
			for (j = 0; j < smallWidth - 1; j++, p++, p2 += 2)
			{
				smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
					+ (unsigned short)bigImage[p2 + 3]
					+ (unsigned short)bigImage[p2 + bWidth3]
					+ (unsigned short)bigImage[p2 + bWidth3 + 3]) >> 2);
				smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]
					+ (unsigned short)bigImage[p2 + bWidth3 + 1]
					+ (unsigned short)bigImage[p2 + bWidth3 + 4]) >> 2);
				smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]
					+ (unsigned short)bigImage[p2 + bWidth3 + 2]
					+ (unsigned short)bigImage[p2 + bWidth3 + 5]) >> 2);
			}
			smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2    ]
				    + (unsigned short)bigImage[p2 + bWidth3]) >> 1);
			smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
				    + (unsigned short)bigImage[p2 + bWidth3 + 1]) >> 1);
			smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
				    + (unsigned short)bigImage[p2 + bWidth3 + 2]) >> 1);
			p += 3;
			p2 += 6;
		}
		for (j = 0; j < smallWidth - 1; j++, p += 3, p2 += 6)
		{
			smallImage[p    ] = (unsigned char)((
					  (unsigned short)bigImage[p2]
					+ (unsigned short)bigImage[p2 + 3]) >> 1);
			smallImage[p + 1] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 1]
					+ (unsigned short)bigImage[p2 + 4]) >> 1);
			smallImage[p + 2] = (unsigned char)((
					  (unsigned short)bigImage[p2 + 2]
					+ (unsigned short)bigImage[p2 + 5]) >> 1);
		}
		smallImage[p    ] = bigImage[p2    ];
		smallImage[p + 1] = bigImage[p2 + 1];
		smallImage[p + 2] = bigImage[p2 + 2];
	}

	return true;
}

//----------------------------------------------------------------------------------------

//! Apply the Deriche filter in a single line (used by atipDericheLaplacians())
//! \param dstLine Destination line, after filtering
//! \param srcLine Source line, before filtering
//! \param tempLine Allocated temporary line, used for the computations
//! \param width Width of the line to compute
//! \param c Coefficients of the filter (a1, a2, b1, b2)
void atipDericheLaplaciansLine(float * dstLine, const float * srcLine, float * tempLine,
							   int width, const float * c)
{
	ATIP_ASSERT(dstLine && srcLine && tempLine && width && c);

	// Apply the recursive filter from the beginning to the end of the line
	dstLine[0] = srcLine[0];
	dstLine[1] = srcLine[1] + c[0] * srcLine[0] + c[2] * dstLine[0];
	int i;
	for (i = 2; i < width; i++)
		dstLine[i] = srcLine[i] + c[0] * srcLine[i - 1]
								+ c[2] * dstLine[i - 1]
								+ c[3] * dstLine[i - 2];

	// Apply the recursive filter from the end to the beginning of the line
	tempLine[width - 1] = 0.0f;
	tempLine[width - 2] = c[1] * srcLine[width - 1];
	for (i = width - 3; i >= 0; i--)
		tempLine[i] =  c[1] * srcLine[i + 1]
					 + c[3] * srcLine[i + 2]
					 + c[2] * tempLine[i + 1]
					 + c[3] * tempLine[i + 2];

	// The result is the opposite of the sum of the two passes of the filter
	for (i = 0; i < width; i++)
		 dstLine[i] = - (dstLine[i] + tempLine[i]);
}

//----------------------------------------------------------------------------------------

//! Threshold of the zero detection of the laplacians coded with floats
#define EPS_F			0.0001f

//! Zero-crossing laplacians detection in a single line coded with floats
//! (used by atipDericheLaplacian())
//! \param dstLine Destination line, must contains 0's when the function is called,
//!                1's are added where there are edges
//! \param srcLine Source line, contains laplacians in a line
//! \param width Width of the lines
void atipZerosLaplaciansLine(unsigned char * dstLine, const float * srcLine,
							 unsigned int width)
{
	ATIP_ASSERT(dstLine && srcLine && width);

	float prod;
	unsigned int i;
	for (i = 0; i < width - 1; i++)
		if ((fabs(srcLine[i]) >= EPS_F) || (fabs(srcLine[i + 1]) >= EPS_F))
		{
			prod = srcLine[i] * srcLine[i + 1];
			if (prod <= 0.0f)
				if (fabs(srcLine[i]) < fabs(srcLine[i + 1]))
						dstLine[i] = 1;
				else	dstLine[i + 1] = 1;
		}
}

//----------------------------------------------------------------------------------------

//! Maximum macro
bool atipDericheLaplacians(unsigned char * dstImage, const float * srcImage,
						   unsigned int width, unsigned int height, float alpha)
{
	ATIP_ASSERT(dstImage && srcImage && width && height && (alpha >= 0.0f));
	if (!dstImage || !srcImage || !width || !height || (alpha < 0.0f))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Compute the coefficients of the laplacian filter
	float e = exp(-alpha);
	float k = (1.0f - e * e) / (2.0f * alpha * e);
	float c[4] = {	- (1.0f + k * alpha) * e,		// a1
					(1.0f - k * alpha) * e,			// a2
					2.0f * e,						// b1
					- e * e };						// b2

	// Apply the Deriche laplacian filter to the lines
	unsigned int nPixels = width * height;
	float * tempImage = 0;
	ATIP_NEW_ARRAY(float, tempImage, nPixels);
	float * tempLine = 0;
	ATIP_NEW_ARRAY(float, tempLine, MAX(width, height));
	unsigned int i;
	for (i = 0; i < height; i++)
		atipDericheLaplaciansLine(tempImage + i * width, srcImage + i * width,
								  tempLine, (int)width, c);

	// Apply the Deriche laplacian filter to the columns
	float * tempCol1 = 0;
	ATIP_NEW_ARRAY(float, tempCol1, height);
	float * tempCol2 = 0;
	ATIP_NEW_ARRAY(float, tempCol2, height);
	float * laplacians = 0;
	ATIP_NEW_ARRAY(float, laplacians, width * height);
	unsigned int j;
	for (j = 0; j < width; j++)
	{
		for (i = 0; i < height; i++) tempCol1[i] = tempImage[i * width + j];
		atipDericheLaplaciansLine(tempCol2, tempCol1, tempLine, (int)height, c);
		for (i = 0; i < height; i++) laplacians[i * width + j] = tempCol2[i];
	}

	// Look for the zeros of the laplacians in the lines
	memset(dstImage, 0, nPixels);
	for (i = 0; i < height; i++)
		atipZerosLaplaciansLine(dstImage + i * width, laplacians + i * width, width);

	// Look for the zeros of the laplacians in the columns
	unsigned char * tempUCharCol = 0;
	ATIP_NEW_ARRAY(unsigned char, tempUCharCol, height);
	for (j = 0; j < width; j++)
	{
		for (i = 0; i < height; i++) tempCol1[i] = laplacians[i * width + j];
		memset(tempUCharCol, 0, height);
		atipZerosLaplaciansLine(tempUCharCol, tempCol1, height);
		for (i = 0; i < height; i++)
			if (dstImage[i * width + j] == 0) dstImage[i * width + j] = tempUCharCol[i];
	}

	ATIP_DELETE_ARRAY(tempImage);
	ATIP_DELETE_ARRAY(tempLine);
	ATIP_DELETE_ARRAY(tempCol1);
	ATIP_DELETE_ARRAY(tempCol2);
	ATIP_DELETE_ARRAY(laplacians);
	ATIP_DELETE_ARRAY(tempUCharCol);
	return true;
}

//----------------------------------------------------------------------------------------

//! Threshold of the zero detection of the laplacians coded with short integers
#define EPS_SH			1

//! Zero-crossing laplacians detection in a single line coded with short integers
//! (used by atipSimpleLaplacians())
//! \param dstLine Destination line, must contains 0's when the function is called,
//!                1's are added where there are edges
//! \param srcLine Source line, contains laplacians in a line
//! \param width Width of the lines
void atipZerosLaplaciansLine(unsigned char * dstLine, const short * srcLine,
							 unsigned int width)
{
	ATIP_ASSERT(dstLine && srcLine && width);

	short prod;
	unsigned int i;
	for (i = 0; i < width - 1; i++)
		if ((abs(srcLine[i]) >= EPS_SH) || (abs(srcLine[i + 1]) >= EPS_SH))
		{
			prod = srcLine[i] * srcLine[i + 1];
			if (prod <= 0)
				if (abs(srcLine[i]) < abs(srcLine[i + 1]))
						dstLine[i] = 1;
				else	dstLine[i + 1] = 1;
		}
}

//----------------------------------------------------------------------------------------

bool atipSimpleLaplacians(unsigned char * dstImage, const unsigned char * srcImage,
						  unsigned int width, unsigned int height)
{
	ATIP_ASSERT(dstImage && srcImage && width && height);
	if (!dstImage || !srcImage || !width || !height)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// A laplacian filter like this is used :	1 | -2  1 -2 |
	//											- |  1  4  1 |
	//											8 | -2  1 -2 |

	// It is the sum of two filters :	1 | -1  2 -1 |   1 | -1 -1 -1 |
	//									- | -1  2 -1 | + - |  2  2  2 |
	//									8 | -1  2 -1 |   8 | -1 -1 -1 |

	// Each filter is separable, two 1D filters will be used :
	// [ -1 2 -1 ] and 1/8 [ 1 1 1 ]

	// Finally, the needed computations are approximately equals, so a direct
	// 2D convolution will be done

	// Conversion to an image of signed short integers
	unsigned int nPixels = width * height;
	short * shSrcImage = 0;
	ATIP_NEW_ARRAY(short, shSrcImage, nPixels);
	unsigned int p;
	for (p = 0; p < nPixels; p++)
		shSrcImage[p] = (short)srcImage[p];

	// The first and last lines and columns of the filtered image are filled with 0's
	short * laplacians = 0;
	ATIP_NEW_ARRAY(short, laplacians, nPixels);
	memset(laplacians, 0, width * sizeof(short));
	memset(laplacians + (height - 1) * width, 0, width * sizeof(short));
	unsigned int i;
	for (i = 2, p = width; i < height; i++, p += width)
	{
		laplacians[p] = 0;
		laplacians[p + width - 1] = 0;
	}

	// Apply the filter
	unsigned int j;
	for (i = 2, p = width + 1; i < height; i++, p += 2)
		for (j = 2; j < width; j++, p++)
			laplacians[p] = ( (shSrcImage[p] << 2)
							 + shSrcImage[p - width]
							 + shSrcImage[p - 1]
							 + shSrcImage[p + 1]
							 + shSrcImage[p + width]
							 - (shSrcImage[p - width - 1] << 1)
							 - (shSrcImage[p - width + 1] << 1)
							 - (shSrcImage[p + width - 1] << 1)
							 - (shSrcImage[p + width + 1] << 1) ) >> 3;

	// Look for the zeros of the laplacians in the lines
	memset(dstImage, 0, nPixels);
	for (i = 0; i < height; i++)
		atipZerosLaplaciansLine(dstImage + i * width, laplacians + i * width, width);

	// Look for the zeros of the laplacians in the columns
	short * tempCol = 0;
	ATIP_NEW_ARRAY(short, tempCol, height);
	unsigned char * tempUCharCol = 0;
	ATIP_NEW_ARRAY(unsigned char, tempUCharCol, height);
	for (j = 0; j < width; j++)
	{
		for (i = 0; i < height; i++) tempCol[i] = laplacians[i * width + j];
		memset(tempUCharCol, 0, height);
		atipZerosLaplaciansLine(tempUCharCol, tempCol, height);
		for (i = 0; i < height; i++)
			if (dstImage[i * width + j] == 0) dstImage[i * width + j] = tempUCharCol[i];
	}

	ATIP_DELETE_ARRAY(shSrcImage);
	ATIP_DELETE_ARRAY(laplacians);
	ATIP_DELETE_ARRAY(tempCol);
	ATIP_DELETE_ARRAY(tempUCharCol);
	return true;
}

//----------------------------------------------------------------------------------------

//@{
//! Frei-Chen coefficients
//! \note For the row gradients, the matrix is
//!               [[FC_A 0 -FC_A] [FC_B 0 -FC_B] [FC_A 0 -FC_A]]
//!       For the column gradients, the matrix is
//!               [[-FC_A -FC_B -FC_A] [0 0 0] [FC_A FC_B FC_A]]
#define FC_A		0.29289f				// = 1.0f / (2.0f + sqrt(2.0f));
#define FC_B		0.41421f				// = sqrt(2.0f) / (2.0f + sqrt(2.0f));
//@}

bool atipFreiChenFilter(float * gradX, float * gradY, const unsigned char * srcImage,
						unsigned int width, unsigned int height)
{
	ATIP_ASSERT(gradX && gradY && srcImage && width && height);
	if (!gradX || !gradY || !srcImage || !width || !height)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int i, j, p;
	unsigned int nPixels = width * height;
	
	// Fill the first and last rows and columns of gradX and gradY with 0's
	memset(gradX, 0, width * sizeof(float));
	memset(gradX + (height - 1) * width, 0, width * sizeof(float));
	memset(gradY, 0, width * sizeof(float));
	memset(gradY + (height - 1) * width, 0, width * sizeof(float));

	for (i = 1; i < height - 1; i++)
	{
		gradX[i * width] = 0.0f;
		gradX[i * width + width - 1] = 0.0f;
		gradY[i * width] = 0.0f;
		gradY[i * width + width - 1] = 0.0f;
	}

	// Conversion of the 8 bits original image to an image of floats
	float * image = 0;
	ATIP_NEW_ARRAY(float, image, nPixels);
	for (p = 0; p < nPixels; p++)
		image[p] = (float)srcImage[p];

	// Apply the filter for the remaining pixels
	p = width + 1;
	for (i = 1; i < height - 1; i++)
	{
		for (j = 1; j < width - 1; j++)
		{
			gradX[p] =   image[p - width - 1] * FC_A
					   - image[p - width + 1] * FC_A
					   + image[p - 1]         * FC_B
					   - image[p + 1]         * FC_B
					   + image[p + width - 1] * FC_A
					   - image[p + width + 1] * FC_A;

			gradY[p] = - image[p - width - 1] * FC_A
					   - image[p - width]     * FC_B
					   - image[p - width + 1] * FC_A
					   + image[p + width - 1] * FC_A
					   + image[p + width]     * FC_B
					   + image[p + width + 1] * FC_A;
			p++;
		}
		p += 2;
	}

	ATIP_DELETE_ARRAY(image);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipSimpleThreshold(unsigned char * & edges,
						 int * edgePoints, unsigned int & nEdgePoints,
						 const unsigned char * mask, const float * srcImage,
						 unsigned int width, unsigned int height,
						 float threshold)
{
	ATIP_ASSERT(edgePoints && mask && srcImage && width && height
				&& (threshold >= 0.0f) && (threshold <= ATIP_MAX_THRESHOLD));
	if (   !edgePoints || !mask || !srcImage || !width || !height
		|| (threshold < 0.0f) || (threshold > ATIP_MAX_THRESHOLD))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Allocation of the final edge points map
	ATIP_NEW_ARRAY(unsigned char, edges, width * height);
	memset(edges, 0, width * height);

	nEdgePoints = 0;
	unsigned int ep = 0;
	float gradX, gradY;
	unsigned int i, j;
	unsigned int p = width + 1;
	double dThreshold = (double)threshold;

	for (i = 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
			if (mask[p])
			{
				// Compute the gradients using a Frei-Chen filter
				gradX =   srcImage[p - width - 1] * FC_A
						- srcImage[p - width + 1] * FC_A
						+ srcImage[p - 1]         * FC_B
						- srcImage[p + 1]         * FC_B
						+ srcImage[p + width - 1] * FC_A
						- srcImage[p + width + 1] * FC_A;
				gradY = - srcImage[p - width - 1] * FC_A
						- srcImage[p - width]     * FC_B
						- srcImage[p - width + 1] * FC_A
						+ srcImage[p + width - 1] * FC_A
						+ srcImage[p + width]     * FC_B
						+ srcImage[p + width + 1] * FC_A;

				// If the norm of the gradient is greater than a threshold,
				// add it to the list of edge points
				if (sqrt(gradX * gradX + gradY * gradY) > dThreshold)
				{
					edges[p] = 1;
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}
			}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipSimpleThreshold(unsigned char * & edges,
						 int * edgePoints, unsigned int & nEdgePoints,
						 const float * gradX, const float * gradY,
						 unsigned int width, unsigned int height, float threshold)
{
	ATIP_ASSERT(edgePoints && gradX && gradY && width && height
				&& (threshold >= 0.0f) && (threshold <= ATIP_MAX_THRESHOLD));
	if (   !edgePoints || !gradX || !gradY || !width || !height
		|| (threshold < 0.0f) || (threshold > ATIP_MAX_THRESHOLD))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Allocation of the final edge points map
	ATIP_NEW_ARRAY(unsigned char, edges, width * height);
	memset(edges, 0, width * height);

	nEdgePoints = 0;
	unsigned int ep = 0;
	unsigned int i, j;
	unsigned int p = width + 1;
	double dThreshold = (double)threshold;

	for (i = 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
			// If the norm of the gradient is greater than a threshold,
			// add it to the list of edge points
			if (sqrt(gradX[p] * gradX[p] + gradY[p] * gradY[p]) > dThreshold)
			{
				edges[p] = 1;
				edgePoints[ep++] = (int)j;
				edgePoints[ep++] = (int)i;
				nEdgePoints++;
			}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipHysteresisThresholdNeighbor(unsigned char * & edges,
									 int * edgePoints, unsigned int & nEdgePoints,
									 const unsigned char * mask, const float * srcImage,
									 unsigned int width, unsigned int height,
			                         float lowThreshold, float highThreshold)
{
	ATIP_ASSERT(edgePoints && mask && srcImage && width && height
				&& (lowThreshold >= 0.0f) && (lowThreshold < highThreshold)
				&& (highThreshold <= ATIP_MAX_THRESHOLD));
	if (   !edgePoints || !mask || !srcImage || !width || !height
		|| (lowThreshold < 0.0f) || (lowThreshold >= highThreshold)
		|| (highThreshold > ATIP_MAX_THRESHOLD))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Allocation of the final edge points map
	ATIP_NEW_ARRAY(unsigned char, edges, width * height);
	memset(edges, 0, width * height);

	// Compute the gradients of the whole image
	unsigned int nPixels = width * height;
	float * gradX = 0;
	ATIP_NEW_ARRAY(float, gradX, nPixels);
	float * gradY = 0;
	ATIP_NEW_ARRAY(float, gradY, nPixels);
	double * norm = 0;
	ATIP_NEW_ARRAY(double, norm, nPixels);
	bool * higher = 0;
	ATIP_NEW_ARRAY(bool, higher, nPixels);
	memset(higher, 0, nPixels);
	double dLowThreshold = (double)lowThreshold;
	double dHighThreshold = (double)highThreshold;
	nEdgePoints = 0;
	unsigned int ep = 0;
	unsigned int i, j;
	unsigned int p = width + 1;
	for (i = 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
		{
			// Compute the gradients using a Frei-Chen filter
			gradX[p] =  srcImage[p - width - 1] * FC_A
					  - srcImage[p - width + 1] * FC_A
					  + srcImage[p - 1]         * FC_B
					  - srcImage[p + 1]         * FC_B
					  + srcImage[p + width - 1] * FC_A
					  - srcImage[p + width + 1] * FC_A;

			gradY[p] = - srcImage[p - width - 1] * FC_A
					   - srcImage[p - width]     * FC_B
					   - srcImage[p - width + 1] * FC_A
					   + srcImage[p + width - 1] * FC_A
					   + srcImage[p + width]     * FC_B
					   + srcImage[p + width + 1] * FC_A;

			// Compute the L2 norm
			norm[p] = sqrt(gradX[p] * gradX[p] + gradY[p] * gradY[p]);

			// If the norm is greater than the high threshold
			if (norm[p] > dHighThreshold)
			{
				higher[p] = true;
				
				// If the pixel is inside the mask, it is an edge pixel
				if (mask[p])
				{
					edges[p] = 1;
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}
			}
		}

	// Look for the norms that are between the low and the high thresholds, and
	// for which at least one neighbor pixel has a norm greater than the high level
	for (i = 2, p = width + width + 2; i < height - 2; i++, p += 4)
		for (j = 2; j < width - 2; j++, p++)
			if (mask[p] && !higher[p] && (norm[p] > dLowThreshold))
				if (   higher[p - width] || higher[p - 1]
					|| higher[p + 1] || higher[p + width]
					|| higher[p - width - 1] || higher[p - width + 1]
					|| higher[p + width - 1] || higher[p + width + 1])
				{
					edges[p] = 1;
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}

	ATIP_DELETE_ARRAY(gradX);
	ATIP_DELETE_ARRAY(gradY);
	ATIP_DELETE_ARRAY(norm);
	ATIP_DELETE_ARRAY(higher);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipHysteresisThreshold(unsigned char * & edges,
							 int * edgePoints, unsigned int & nEdgePoints,
							 const unsigned char * mask, const float * srcImage,
							 unsigned int width, unsigned int height,
							 float lowThreshold, float highThreshold)
{
	ATIP_ASSERT(edgePoints && mask && srcImage && width && height
				&& (lowThreshold >= 0.0f) && (lowThreshold < highThreshold)
				&& (highThreshold <= ATIP_MAX_THRESHOLD));
	if (   !edgePoints || !mask || !srcImage || !width || !height
		|| (lowThreshold < 0.0f) || (lowThreshold >= highThreshold)
		|| (highThreshold > ATIP_MAX_THRESHOLD))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Allocation of the final edge points map
	ATIP_NEW_ARRAY(unsigned char, edges, width * height);
	memset(edges, 0, width * height);

	// Compute the gradients of the pixels inside the mask
	unsigned int nPixels = width * height;
	float * gradX = 0;
	ATIP_NEW_ARRAY(float, gradX, nPixels);
	float * gradY = 0;
	ATIP_NEW_ARRAY(float, gradY, nPixels);
	double * norm = 0;
	ATIP_NEW_ARRAY(double, norm, nPixels);
	bool * higher = 0;
	ATIP_NEW_ARRAY(bool, higher, nPixels);
	memset(higher, 0, nPixels);
	double dLowThreshold = (double)lowThreshold;
	double dHighThreshold = (double)highThreshold;
	nEdgePoints = 0;
	unsigned int ep = 0;
	unsigned int i, j;
	unsigned int p = width + 1;
	for (i = 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
			if (mask[p])
			{
				// Compute the gradients using a Frei-Chen filter
				gradX[p] =  srcImage[p - width - 1] * FC_A
						  - srcImage[p - width + 1] * FC_A
						  + srcImage[p - 1]         * FC_B
						  - srcImage[p + 1]         * FC_B
						  + srcImage[p + width - 1] * FC_A
						  - srcImage[p + width + 1] * FC_A;

				gradY[p] = - srcImage[p - width - 1] * FC_A
						   - srcImage[p - width]     * FC_B
						   - srcImage[p - width + 1] * FC_A
						   + srcImage[p + width - 1] * FC_A
						   + srcImage[p + width]     * FC_B
						   + srcImage[p + width + 1] * FC_A;

				// Compute the L2 norm
				norm[p] = sqrt(gradX[p] * gradX[p] + gradY[p] * gradY[p]);

				// If the norm is greater than the high threshold, it is an edge pixel
				if (norm[p] > dHighThreshold)
				{
					higher[p] = true;
					edges[p] = 1;
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}
			}

	// Look for the norms that are between the low and the high thresholds, and
	// for which at least one neighbor pixel in the mask has a norm greater
	// than the high level
	for (i = 2, p = width + width + 2; i < height - 2; i++, p += 4)
		for (j = 2; j < width - 2; j++, p++)
			if (mask[p] && !higher[p] && (norm[p] > dLowThreshold))
				if (   higher[p - width] || higher[p - 1]
					|| higher[p + 1] || higher[p + width]
					|| higher[p - width - 1] || higher[p - width + 1]
					|| higher[p + width - 1] || higher[p + width + 1])
				{
					edges[p] = 1;
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}

	ATIP_DELETE_ARRAY(gradX);
	ATIP_DELETE_ARRAY(gradY);
	ATIP_DELETE_ARRAY(norm);
	ATIP_DELETE_ARRAY(higher);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipHysteresisThreshold(unsigned char * & edges,
							 int * edgePoints, unsigned int & nEdgePoints,
							 const float * gradX, const float * gradY,
							 unsigned int width, unsigned int height,
							 float lowThreshold, float highThreshold)
{
	ATIP_ASSERT(edgePoints && gradX && gradY && width && height
				&& (lowThreshold >= 0.0f) && (lowThreshold < highThreshold)
				&& (highThreshold <= ATIP_MAX_THRESHOLD));
	if (   !edgePoints || !gradX || !gradY || !width || !height
		|| (lowThreshold < 0.0f) || (lowThreshold >= highThreshold)
		|| (highThreshold > ATIP_MAX_THRESHOLD))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Allocation of the final edge points map
	ATIP_NEW_ARRAY(unsigned char, edges, width * height);
	memset(edges, 0, width * height);

	unsigned int nPixels = width * height;
	double * norm = 0;
	ATIP_NEW_ARRAY(double, norm, nPixels);
	bool * higher = 0;
	ATIP_NEW_ARRAY(bool, higher, nPixels);
	memset(higher, 0, nPixels);
	double dLowThreshold = (double)lowThreshold;
	double dHighThreshold = (double)highThreshold;
	nEdgePoints = 0;
	unsigned int ep = 0;
	unsigned int i, j;
	unsigned int p = width + 1;
	for (i = 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
		{
			// Compute the L2 norm
			norm[p] = sqrt(gradX[p] * gradX[p] + gradY[p] * gradY[p]);

			// If the norm is greater than the high threshold
			if (norm[p] > dHighThreshold)
			{
				higher[p] = true;
				edges[p] = 1;
				edgePoints[ep++] = (int)j;
				edgePoints[ep++] = (int)i;
				nEdgePoints++;
			}
		}

	// Look for the norms that are between the low and the high thresholds, and
	// for which at least one neighbor pixel has a norm greater than the high level
	for (i = 2, p = width + width + 2; i < height - 2; i++, p += 4)
		for (j = 2; j < width - 2; j++, p++)
			if (!higher[p] && (norm[p] > dLowThreshold))
				if (   higher[p - width] || higher[p - 1]
					|| higher[p + 1] || higher[p + width]
					|| higher[p - width - 1] || higher[p - width + 1]
					|| higher[p + width - 1] || higher[p + width + 1])
				{
					edges[p];
					edgePoints[ep++] = (int)j;
					edgePoints[ep++] = (int)i;
					nEdgePoints++;
				}

	ATIP_DELETE_ARRAY(norm);
	ATIP_DELETE_ARRAY(higher);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipMeanFilter8(unsigned char * dstImage, const unsigned char * srcImage,
					 unsigned int width, unsigned int height,
					 unsigned int windowSize)
{
	ATIP_ASSERT(   dstImage && srcImage && width && height
				&& (windowSize >= 3) && (windowSize & 1)
				&& (width >= windowSize) && (height >= windowSize));
	if (   !dstImage || !srcImage || !width || !height
		|| (windowSize < 3) || ((windowSize & 1) == 0)
		|| (width < windowSize) || (height < windowSize))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int winWidth = windowSize >> 1;
	unsigned int nWinPixels = windowSize * windowSize;
	unsigned int i, j, k, p;

	// Fill the side rows and columns with 0's
	for (i = 0; i < winWidth; i++)
	{
		memset(dstImage + i * width, 0, width);
		memset(dstImage + width * (height - 1 - i), 0, width);
	}
	for (i = winWidth; i < height - winWidth; i++)
	{
		memset(dstImage + i * width, 0, winWidth);
		memset(dstImage + (i + 1) * width - winWidth, 0, winWidth);
	}

	// Compute the first pixel of the final image
	unsigned int sum = 0;
	for (i = 0, p = 0; i < windowSize; i++, p += width)
		for (j = 0; j < windowSize; j++)
			sum += (unsigned int)srcImage[p + j];
	dstImage[winWidth * width + winWidth] = (unsigned char)(sum / nWinPixels);

	// Compute the remaining pixel of the final image
	for (i = winWidth; i < height - winWidth; )
	{
		// Move the window to the right
		for (j = winWidth + 1; j < width - winWidth; j++)
		{
			for (k = i - winWidth; k <= i + winWidth; k++)
			{
				// Remove the first column of the previous window from the sum
				sum -= (unsigned int)srcImage[k * width + j - winWidth - 1];

				// Add the last column of the new window to the sum
				sum += (unsigned int)srcImage[k * width + j + winWidth];
			}
			dstImage[i * width + j] = (unsigned char)(sum / nWinPixels);
		}

		// Next line if possible
		i++;
		j--;
		if (i < height - winWidth)
		{
			// Move the window one step to the bottom
			for (k = j - winWidth; k <= j+ winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= (unsigned int)srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += (unsigned int)srcImage[(i + winWidth) * width + k];
			}
			dstImage[i * width + j] = (unsigned char)(sum / nWinPixels);

			// Move the window to the left
			for (j = width - winWidth - 2; j >= winWidth; j--)
			{
				for (k = i - winWidth; k <= i + winWidth; k++)
				{
					// Remove the last column of the previous window from the sum
					sum -= (unsigned int)srcImage[k * width + j + winWidth + 1];

					// Add the first column of the new window to the sum
					sum += (unsigned int)srcImage[k * width + j - winWidth];
				}
				dstImage[i * width + j] = (unsigned char)(sum / nWinPixels);
			}
		}

		// Move the window one step to the bottom
		i++;
		j++;
		if (i < height - winWidth)
		{
			for (k = j - winWidth; k <= j+ winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= (unsigned int)srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += (unsigned int)srcImage[(i + winWidth) * width + k];
			}
			dstImage[i * width + j] = (unsigned char)(sum / nWinPixels);
		}
	}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipMeanFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					  unsigned int width, unsigned int height,
					  unsigned int windowSize)
{
	ATIP_ASSERT(   dstImage && srcImage && width && height
				&& (windowSize >= 3) && (windowSize & 1)
				&& (width >= windowSize) && (height >= windowSize));
	if (   !dstImage || !srcImage || !width || !height
		|| (windowSize < 3) || ((windowSize & 1) == 0)
		|| (width < windowSize) || (height < windowSize))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int winWidth = windowSize >> 1;
	unsigned int nWinPixels = windowSize * windowSize;
	unsigned int i, j, k, p;

	// Fill the side rows and columns with 0's
	for (i = 0; i < winWidth; i++)
	{
		memset(dstImage + i * width, 0, width * sizeof(unsigned int));
		memset(dstImage + width * (height - 1 - i), 0, width * sizeof(unsigned int));
	}
	for (i = winWidth; i < height - winWidth; i++)
	{
		memset(dstImage + i * width, 0, winWidth * sizeof(unsigned int));
		memset(dstImage + (i + 1) * width - winWidth, 0, winWidth * sizeof(unsigned int));
	}

	// Compute the first pixel of the final image
	unsigned int sum = 0;
	for (i = 0, p = 0; i < windowSize; i++, p += width)
		for (j = 0; j < windowSize; j++)
			sum += srcImage[p + j];
	dstImage[winWidth * width + winWidth] = sum / nWinPixels;

	// Compute the remaining pixel of the final image
	for (i = winWidth; i < height - winWidth; )
	{
		// Move the window to the right
		for (j = winWidth + 1; j < width - winWidth; j++)
		{
			for (k = i - winWidth; k <= i + winWidth; k++)
			{
				// Remove the first column of the previous window from the sum
				sum -= srcImage[k * width + j - winWidth - 1];

				// Add the last column of the new window to the sum
				sum += srcImage[k * width + j + winWidth];
			}
			dstImage[i * width + j] = sum / nWinPixels;
		}

		// Next line if possible
		i++;
		j--;
		if (i < height - winWidth)
		{
			// Move the window one step to the bottom
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += srcImage[(i + winWidth) * width + k];
			}
			dstImage[i * width + j] = sum / nWinPixels;

			// Move the window to the left
			for (j = width - winWidth - 2; j >= winWidth; j--)
			{
				for (k = i - winWidth; k <= i + winWidth; k++)
				{
					// Remove the last column of the previous window from the sum
					sum -= srcImage[k * width + j + winWidth + 1];

					// Add the first column of the new window to the sum
					sum += srcImage[k * width + j - winWidth];
				}
				dstImage[i * width + j] = sum / nWinPixels;
			}
		}

		// Move the window one step to the bottom
		i++;
		j++;
		if (i < height - winWidth)
		{
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum	+= srcImage[(i + winWidth) * width + k];
			}
		}
		dstImage[i * width + j] = sum / nWinPixels;
	}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipMeanFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					  unsigned int width, unsigned int height,
					  unsigned int windowSize, unsigned int & max)
{
	ATIP_ASSERT(   dstImage && srcImage && width && height
				&& (windowSize >= 3) && (windowSize & 1)
				&& (width >= windowSize) && (height >= windowSize));
	if (   !dstImage || !srcImage || !width || !height
		|| (windowSize < 3) || ((windowSize & 1) == 0)
		|| (width < windowSize) || (height < windowSize))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int winWidth = windowSize >> 1;
	unsigned int nWinPixels = windowSize * windowSize;
	unsigned int i, j, k, p, v;

	// Fill the side rows and columns with 0's
	for (i = 0; i < winWidth; i++)
	{
		memset(dstImage + i * width, 0, width * sizeof(unsigned int));
		memset(dstImage + width * (height - 1 - i), 0, width * sizeof(unsigned int));
	}
	for (i = winWidth; i < height - winWidth; i++)
	{
		memset(dstImage + i * width, 0, winWidth * sizeof(unsigned int));
		memset(dstImage + (i + 1) * width - winWidth, 0, winWidth * sizeof(unsigned int));
	}

	// Compute the first pixel of the final image
	unsigned int sum = 0;
	for (i = 0, p = 0; i < windowSize; i++, p += width)
		for (j = 0; j < windowSize; j++)
			sum += srcImage[p + j];
	max = dstImage[winWidth * width + winWidth] = sum / nWinPixels;

	// Compute the remaining pixel of the final image
	for (i = winWidth; i < height - winWidth; )
	{
		// Move the window to the right
		for (j = winWidth + 1; j < width - winWidth; j++)
		{
			for (k = i - winWidth; k <= i + winWidth; k++)
			{
				// Remove the first column of the previous window from the sum
				sum -= srcImage[k * width + j - winWidth - 1];

				// Add the last column of the new window to the sum
				sum += srcImage[k * width + j + winWidth];
			}
			v = sum / nWinPixels;
			if (v > max) max = v;
			dstImage[i * width + j] = v;
		}

		// Next line if possible
		i++;
		j--;
		if (i < height - winWidth)
		{
			// Move the window one step to the bottom
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += srcImage[(i + winWidth) * width + k];
			}
			v = sum / nWinPixels;
			if (v > max) max = v;
			dstImage[i * width + j] = v;

			// Move the window to the left
			for (j = width - winWidth - 2; j >= winWidth; j--)
			{
				for (k = i - winWidth; k <= i + winWidth; k++)
				{
					// Remove the last column of the previous window from the sum
					sum -= srcImage[k * width + j + winWidth + 1];

					// Add the first column of the new window to the sum
					sum += srcImage[k * width + j - winWidth];
				}
				v = sum / nWinPixels;
				if (v > max) max = v;
				dstImage[i * width + j] = v;
			}
		}

		// Move the window one step to the bottom
		i++;
		j++;
		if (i < height - winWidth)
		{
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += srcImage[(i + winWidth) * width + k];
			}
			v = sum / nWinPixels;
			if (v > max) max = v;
			dstImage[i * width + j] = v;
		}
	}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipSumFilter32(unsigned int * dstImage, const unsigned int * srcImage,
					 unsigned int width, unsigned int height,
					 unsigned int windowSize, unsigned int & max,
					 unsigned int & xMax, unsigned int & yMax)
{
	ATIP_ASSERT(   dstImage && srcImage && width && height
				&& (windowSize >= 3) && (windowSize & 1)
				&& (width >= windowSize) && (height >= windowSize));
	if (   !dstImage || !srcImage || !width || !height
		|| (windowSize < 3) || ((windowSize & 1) == 0)
		|| (width < windowSize) || (height < windowSize))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int winWidth = windowSize >> 1;
	unsigned int i, j, k, p;

	// Fill the side rows and columns with 0's
	for (i = 0; i < winWidth; i++)
	{
		memset(dstImage + i * width, 0, width * sizeof(unsigned int));
		memset(dstImage + width * (height - 1 - i), 0, width * sizeof(unsigned int));
	}
	for (i = winWidth; i < height - winWidth; i++)
	{
		memset(dstImage + i * width, 0, winWidth * sizeof(unsigned int));
		memset(dstImage + (i + 1) * width - winWidth, 0, winWidth * sizeof(unsigned int));
	}

	// Compute the first pixel of the final image
	unsigned int sum = 0;
	for (i = 0, p = 0; i < windowSize; i++, p += width)
		for (j = 0; j < windowSize; j++)
			sum += srcImage[p + j];
	max = dstImage[winWidth * width + winWidth] = sum;
	xMax = yMax = 0;

	// Compute the remaining pixel of the final image
	for (i = winWidth; i < height - winWidth; )
	{
		// Move the window to the right
		for (j = winWidth + 1; j < width - winWidth; j++)
		{
			for (k = i - winWidth; k <= i + winWidth; k++)
			{
				// Remove the first column of the previous window from the sum
				sum -= srcImage[k * width + j - winWidth - 1];

				// Add the last column of the new window to the sum
				sum += srcImage[k * width + j + winWidth];
			}
			dstImage[i * width + j] = sum;
			if (sum > max)
			{
				max = sum;
				xMax = j;
				yMax = i;
			}
		}

		// Next line if possible
		i++;
		j--;
		if (i < height - winWidth)
		{
			// Move the window one step to the bottom
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += srcImage[(i + winWidth) * width + k];
			}
			dstImage[i * width + j] = sum;
			if (sum > max)
			{
				max = sum;
				xMax = j;
				yMax = i;
			}

			// Move the window to the left
			for (j = width - winWidth - 2; j >= winWidth; j--)
			{
				for (k = i - winWidth; k <= i + winWidth; k++)
				{
					// Remove the last column of the previous window from the sum
					sum -= srcImage[k * width + j + winWidth + 1];

					// Add the first column of the new window to the sum
					sum += srcImage[k * width + j - winWidth];
				}
				dstImage[i * width + j] = sum;
				if (sum > max)
				{
					max = sum;
					xMax = j;
					yMax = i;
				}
			}
		}

		// Move the window one step to the bottom
		i++;
		j++;
		if (i < height - winWidth)
		{
			for (k = j - winWidth; k <= j + winWidth; k++)
			{
				// Remove the first row of the previous window from the sum
				sum -= srcImage[(i - winWidth - 1) * width + k];

				// Add the last row of the new window to the sum
				sum += srcImage[(i + winWidth) * width + k];
			}
			dstImage[i * width + j] = sum;
			if (sum > max)
			{
				max = sum;
				xMax = j;
				yMax = i;
			}
		}
	}

	return true;
}

//----------------------------------------------------------------------------------------

bool atipHighPassFilter32(unsigned int * dstImage, const unsigned int * srcImage,
						  unsigned int width, unsigned int height, unsigned int & max)
{
	ATIP_ASSERT(dstImage && srcImage && width && height);
	if (!dstImage || !srcImage || !width || !height)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Fill the first and last lines and columns with 0's
	memset(dstImage, 0, width * sizeof(unsigned int));
	memset(dstImage + (height - 1) * width, 0, width * sizeof(unsigned int));
	unsigned int i, j, p;
	for (i = 1, p = width; i < height - 1; i++, p += width)
	{
		dstImage[p] = 0;
		dstImage[p + width - 1] = 0;
	}

	// Apply a high-pass filter to the image
	max = 0;
	int v;
	for (i = 1, p = width + 1; i < height - 1; i++, p += 2)
		for (j = 1; j < width - 1; j++, p++)
		{
			v = 4 * (int)srcImage[p] - (int)srcImage[p - width] - (int)srcImage[p - 1]
								- (int)srcImage[p + 1] - (int)srcImage[p + width];
			if (v < 0) v = 0;
			if ((unsigned int)v > max) max = (unsigned int)v;
			dstImage[p] = (unsigned int)v;
		}

	return true;
}