/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_gaussian_hemisphere.cpp
//! \author	BOULANGER Kevin
//! \date	20th July 2004
//! \brief	Hough transform on a gaussian hemisphere


#include <van_point/atip_gaussian_hemisphere.h>
#include <van_point/atip_filter.h>
#include <tools/atip_private_tools.h>
#include <tools/atip_primitives.h>
#include <math.h>


//! Draw a segment which lies in 2 lines, adding the given coefficient to the pixels
//! of an accumulator coded with 32 bits
//! \param x1 X coordinate of the point in the first line (drawn)
//! \param x2 X coordinate of the point in the second line (not drawn)
//! \param accum Pointer to the row containing x1
//! \param width Width of the accumulator in pixels
//! \param maximum If an accumulator is greater than this parameter, it is updated
//! \param coeff Coefficient to add to each accumulator cell
void atipDrawSegmentY2Accum32(int x1, int x2, unsigned int * accum, unsigned int width,
							  unsigned int & maximum, unsigned int coeff)
{
	// Remove the segments which are completely outside the accumulation space
	if (x1 < 0 && x2 < 0) return;
	if (x1 >= (int)width && x2 >= (int)width) return;

#define CLAMP_WIDTH(a)	((a) < 0 ? 0 : ((a) >= (int)width ? (int)width - 1 : (a)))

	int center, x, p;
	int i1, ic1, ic2, i2;
	unsigned int newAccum;

	// If the second point is on the right of the first point
	if (x1 <= x2)
	{
		center = (x1 + x2) >> 1;
		i1 = CLAMP_WIDTH(x1);
		ic1 = CLAMP_WIDTH(center);
		ic2 = CLAMP_WIDTH(center + 1);
		i2 = CLAMP_WIDTH(x2);

		for (x = i1, p = i1; x <= ic1; x++, p++)
		{
			newAccum = (accum[p] += coeff);
			if (newAccum > maximum) maximum = newAccum;
		}
		for (x = ic2, p += width; x < i2; x++, p++)
		{
			newAccum = (accum[p] += coeff);
			if (newAccum > maximum) maximum = newAccum;
		}
	}

	// If the second point is on the left of the first point
	else
	{
		center = (x1 + x2 + 1) >> 1;
		i1 = CLAMP_WIDTH(x1);
		ic1 = CLAMP_WIDTH(center);
		ic2 = CLAMP_WIDTH(center - 1);
		i2 = CLAMP_WIDTH(x2);

		for (x = i1, p = i1; x >= ic1; x--, p--)
		{
			newAccum = (accum[p] += coeff);
			if (newAccum > maximum) maximum = newAccum;
		}
		for (x = ic2, p += width; x > i2; x--, p--)
		{
			newAccum = (accum[p] += coeff);
			if (newAccum > maximum) maximum = newAccum;
		}
	}
}

//----------------------------------------------------------------------------------------

//! Draw a segment which lies in 2 lines, substracting the given coefficient to the pixels
//! of an accumulator coded with 32 bits
//! \param x1 X coordinate of the point in the first line (drawn)
//! \param x2 X coordinate of the point in the second line (not drawn)
//! \param accum Pointer to the row containing x1
//! \param width Width of the accumulator in pixels
//! \param coeff Coefficient to add to each accumulator cell
void atipDrawSegmentY2MinusAccum32(int x1, int x2, unsigned int * accum,
								   unsigned int width, unsigned int coeff)
{
	// Remove the segments which are completely outside the accumulation space
	if (x1 < 0 && x2 < 0) return;
	if (x1 >= (int)width && x2 >= (int)width) return;

#define CLAMP_WIDTH(a)	((a) < 0 ? 0 : ((a) >= (int)width ? (int)width - 1 : (a)))

	int center, x, p;
	int i1, ic1, ic2, i2;

	// If the second point is on the right of the first point
	if (x1 <= x2)
	{
		center = (x1 + x2) >> 1;
		i1 = CLAMP_WIDTH(x1);
		ic1 = CLAMP_WIDTH(center);
		ic2 = CLAMP_WIDTH(center + 1);
		i2 = CLAMP_WIDTH(x2);

		for (x = i1, p = i1; x <= ic1; x++, p++)
			accum[p] -= coeff;
		for (x = ic2, p += width; x < i2; x++, p++)
			accum[p] -= coeff;
	}

	// If the second point is on the left of the first point
	else
	{
		center = (x1 + x2 + 1) >> 1;
		i1 = CLAMP_WIDTH(x1);
		ic1 = CLAMP_WIDTH(center);
		ic2 = CLAMP_WIDTH(center - 1);
		i2 = CLAMP_WIDTH(x2);

		for (x = i1, p = i1; x >= ic1; x--, p--)
			accum[p] -= coeff;
		for (x = ic2, p += width; x > i2; x--, p--)
			accum[p] -= coeff;
	}
}

//----------------------------------------------------------------------------------------

bool atipGaussHough(unsigned int * & hough, unsigned int & maxHough,
					const float * lines, const unsigned int * lineCoeffs,
					unsigned int nLines, unsigned int nTheta, unsigned int nPhi,
					unsigned int windowSize, unsigned int width, unsigned int height)
{
	ATIP_ASSERT(   lines && (nLines > 0) && (nTheta > 0) && (nPhi > 0)
				&& ((windowSize & 1) == 1) && (windowSize >= 3)
				&& (width > 0) && (height > 0));
	if (   !lines || (nLines == 0) || (nTheta == 0) || (nPhi == 0)
		|| ((windowSize & 1) == 0) || (windowSize < 3) || (width == 0) || (height == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Window half size without the center value
	int windowHalf = (int)windowSize >> 1;

	// Corrected dimensions of the transform
	unsigned int nTheta2 = nTheta + windowSize - 1;
	unsigned int nPhi2 = nPhi + windowSize - 1;

	// Precompute the cosinus and sinus tables
	float * cosTable = 0;
	ATIP_NEW_ARRAY(float, cosTable, nPhi2);
	float * sinTable = 0;
	ATIP_NEW_ARRAY(float, sinTable, nPhi2);
	int phi;
	double dPhi;
	for (phi = -windowHalf; phi < (int)nPhi + windowHalf; phi++)
	{
		dPhi = phi / (double)nPhi * M_2PI;
		cosTable[phi + windowHalf] = (float)cos(dPhi);
		sinTable[phi + windowHalf] = (float)sin(dPhi);
	}

	// Create the destination array
	ATIP_NEW_ARRAY(unsigned int, hough, nTheta2 * nPhi2);
	memset(hough, 0, nTheta2 * nPhi2 * sizeof(unsigned int));
	maxHough = 0;

	// Line coefficients
	unsigned int * lCoeffs = 0;
	ATIP_NEW_ARRAY(unsigned int, lCoeffs, nLines);
	if (lineCoeffs) memcpy(lCoeffs, lineCoeffs, nLines * sizeof(unsigned int));
	else			for (unsigned int c = 0; c < nLines; c++) lCoeffs[c] = 1;

	// Radius of the gaussian sphere
	// (0.25f is a bias to avoid the future divisions by zero)
	float centerX = (float)width / 2.0f - 0.25f;
	float centerY = (float)height / 2.0f - 0.25f;
	float sphereRadius = MIN(centerX, centerY);

	// For each line, do the Hough transform on the gaussian hemisphere
	float nx, ny, nz;
	float t;
	int i, previousI;
	unsigned int offset;
	unsigned int nCoords = nLines << 2;
	unsigned int l;
	maxHough = 0;
	for (l = 0; l < nCoords; l += 4)
	{
		// Compute the normal of the interpretation plane (contains the line
		// and the camera point)
		// n = v1 x v2 (cross product)
		// v1 : vector from the camera to the point (x0, y0) of the line
		// v2 : vector from the camera to the point (x0 + dx, y0 + dy) of the line
		nx = sphereRadius * lines[l + 3];
		ny = -sphereRadius * lines[l + 2];
        nz =   lines[l + 3] * (lines[l    ] - centerX)
			 - lines[l + 2] * (lines[l + 1] - centerY);
		if (nz < 0.0f)
		{
			nx = -nx;
			ny = -ny;
			nz = -nz;
		}

		// For each point along phi
		for (phi = -windowHalf, offset = 0;
			 phi < (int)nPhi + windowHalf;
			 phi++, offset += nTheta2)
		{
			t = nx * cosTable[phi + windowHalf] + ny * sinTable[phi + windowHalf];
			i = (int)floor(t / (float)sqrt(nz * nz + t * t) * nTheta);
			if (phi > -windowHalf)
				atipDrawSegmentY2Accum32(previousI + windowHalf, i + windowHalf,
										 hough - nTheta2 + offset, nTheta2,
										 maxHough, lCoeffs[l >> 2]);
			previousI = i;
		}
	}

	ATIP_DELETE_ARRAY(lCoeffs);
	ATIP_DELETE_ARRAY(cosTable);
	ATIP_DELETE_ARRAY(sinTable);

	return true;
}

//----------------------------------------------------------------------------------------

//! Dot product between two unit vectors defined by their spherical coordinates
//! \param phi1 Phi coordinate of the first vector
//! \param theta1 Theta coordinate of the first vector
//! \param phi2 Phi coordinate of the second vector
//! \param theta2 Theta coordinate of the second vector
//! \return The dot product between the two vectors
double SphereDotProduct(double phi1, double theta1, double phi2, double theta2)
{
	return	sin(theta1) * sin(theta2) *
				(cos(phi1) * cos(phi2) + sin(phi1) * sin(phi2))
			+ cos(theta1) * cos(theta2);
}

//----------------------------------------------------------------------------------------

//! When two finite vanishing points coordinates are too similar, or when two infinite
//! vanishing points directions are too similar, the second vanishing point is removed.
//! To detect this, a scalar product is done, and when it is greater than this threshold,
//! the removal is done
#define ATIP_SIMILAR_VANPOINTS_THRESHOLD		0.999f

//! Maximum number of retries when 3 different vanishing points cannot be found
#define ATIP_RETRIES_MAX_NUMBER					10


bool atipSearchVanPoints(float vanPoints[12], bool infiniteVanPoints[3],
						 float * lineClusters[4], unsigned int nLinesPerCluster[4],
						 unsigned int * hough, const float * lines,
						 const unsigned int * lineCoeffs, unsigned int nLines,
						 unsigned int nTheta, unsigned int nPhi, unsigned int windowSize,
						 unsigned int width, unsigned int height,
						 float maxLineDistance, float thetaInfinite,
						 unsigned int maxLineAngle)
{
	ATIP_ASSERT(   vanPoints && infiniteVanPoints && lineClusters && nLinesPerCluster
				&& hough && lines && (nLines > 0) && (nTheta >= 4) && (nPhi >= 4)
				&& ((windowSize & 1) == 1) && (windowSize >= 3)
				&& (width > 0) && (height > 0)
				&& (maxLineDistance > 0.0f) && (thetaInfinite > 0.0f)
				&& (maxLineAngle > 0) && (maxLineAngle < 90));
	if (   !vanPoints || !infiniteVanPoints || !lineClusters || !nLinesPerCluster
		|| !hough || !lines || (nLines == 0) || (nTheta < 4) || (nPhi < 4)
		|| ((windowSize & 1) == 0) || (windowSize < 3)
		|| (width == 0) || (height == 0)
		/*|| (maxLineDistance <= 0.0f) || (thetaInfinite <= 0.0f)
		|| (maxLineAngle == 0) || (maxLineAngle >= 90)*/)
//! \bug If the test is completely done, in Release mode under Visual Studio .NET 2003,
//!      the function gives bad results
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Window half size without the center value
	int windowHalf = (int)windowSize >> 1;

	// Corrected dimensions of the transform
	unsigned int nTheta2 = nTheta + windowSize - 1;
	unsigned int nPhi2 = nPhi + windowSize - 1;

	// Precompute the cosinus and sinus tables
	float * cosTable = 0;
	ATIP_NEW_ARRAY(float, cosTable, nPhi2);
	float * sinTable = 0;
	ATIP_NEW_ARRAY(float, sinTable, nPhi2);
	int iPhi;
	double dPhi;
	for (iPhi = -windowHalf; iPhi < (int)nPhi + windowHalf; iPhi++)
	{
		dPhi = iPhi / (double)nPhi * M_2PI;
		cosTable[iPhi + windowHalf] = (float)cos(dPhi);
		sinTable[iPhi + windowHalf] = (float)sin(dPhi);
	}

	// Temporary allocation for the line clusters
	float * tempLineClusters[3] = { 0 };
	ATIP_NEW_ARRAY(float, tempLineClusters[0], nLines * 4);
	ATIP_NEW_ARRAY(float, tempLineClusters[1], nLines * 4);
	ATIP_NEW_ARRAY(float, tempLineClusters[2], nLines * 4);
	nLinesPerCluster[0] = 0;
	nLinesPerCluster[1] = 0;
	nLinesPerCluster[2] = 0;

	// Copy of the source lines
	float * srcLines = 0;
	ATIP_NEW_ARRAY(float, srcLines, nLines * 4);
	memcpy(srcLines, lines, nLines * 4 * sizeof(float));

	// Line coefficients
	unsigned int * lCoeffs = 0;
	ATIP_NEW_ARRAY(unsigned int, lCoeffs, nLines);
	if (lineCoeffs) memcpy(lCoeffs, lineCoeffs, nLines * sizeof(unsigned int));
	else			for (unsigned int c = 0; c < nLines; c++) lCoeffs[c] = 1;

	// Infinite vanishing points booleans initialization
	infiniteVanPoints[0] = false;
	infiniteVanPoints[1] = false;
	infiniteVanPoints[2] = false;

	// Info about the gaussian hemisphere
	float centerX = (float)width / 2.0f - 0.25f;
	float centerY = (float)height / 2.0f - 0.25f;
	float sphereRadius = MIN(centerX, centerY);

	// Modified Hough transform
	unsigned int * hough2 = 0;
	ATIP_NEW_ARRAY(unsigned int, hough2, nTheta2 * nPhi2);

	// Max line distance parameter in pixel
	float maxLineDistancePixels = MIN((float)width, (float)height) * maxLineDistance;

	// Minimum dot product between two vectors to consider them as colinear
	float minDotProduct = (float)cos((double)maxLineAngle * M_2PI / 360.0);

	// For each pass
	unsigned int max, xMax, yMax;
	double phi, theta;
	float x, y;
	float nx, ny, nz, t;
	int i, previousI;
	unsigned int l;
	float x0, y0, dx, dy, d;
	unsigned int offset;
	unsigned int retries = 0;
	unsigned int pass;
	for (pass = 0; pass < 3; pass++)
	{
		// Apply a low-pass filter to the transform to improve the quality of the
		// maximum detection
		if (!atipSumFilter32(hough2, hough, nTheta2, nPhi2, windowSize, max, xMax, yMax))
		{
			ATIP_DELETE_ARRAY(tempLineClusters[0]);
			ATIP_DELETE_ARRAY(tempLineClusters[1]);
			ATIP_DELETE_ARRAY(tempLineClusters[2]);
			ATIP_DELETE_ARRAY(srcLines);
			ATIP_DELETE_ARRAY(lCoeffs);
			ATIP_DELETE_ARRAY(hough2);
			ATIP_DELETE_ARRAY(cosTable);
			ATIP_DELETE_ARRAY(sinTable);
		}

		// Once the maximum detected, compute the coordinates of the point
		phi	= (double)(yMax - windowHalf) / (double)nPhi * M_2PI;
		theta = acos((double)(xMax - windowHalf) / (double)nTheta);
		vanPoints[ pass << 2     ] = (float)phi;
		vanPoints[(pass << 2) + 1] = (float)theta;

		// Is the detected point an infinite one ?
		if (theta >= thetaInfinite)
		{
			infiniteVanPoints[pass] = true;

			// Compute the direction inside the original image
			vanPoints[(pass << 2) + 2] = x = (float)cos(phi);
			vanPoints[(pass << 2) + 3] = y = -(float)sin(phi);

			// For each remaining detected line, test if the line direction is
			// similar to the current direction and then add them to the current cluster
			for (l = 0; l < nLines; l++)
			{
				dx = srcLines[(l << 2) + 2];
				dy = srcLines[(l << 2) + 3];
			
				if ((dx != 0.0f) || (dy != 0.0f))
				{
					x0 = srcLines[ l << 2     ];
					y0 = srcLines[(l << 2) + 1];

					// If the dot product between the current direction and the line
					// direction is greater than a threshold, these directions are
					// considered equal
					if (fabs(dx * x + dy * y) >= minDotProduct)
					{
						// Copy the current line to the current cluster
						memcpy(tempLineClusters[pass] + (nLinesPerCluster[pass] << 2),
							   srcLines + (l << 2), 4 * sizeof(float));
						memset(srcLines + (l << 2), 0, 4 * sizeof(float));
						nLinesPerCluster[pass]++;

						// Compute the normal of the interpretation plane
						// (contains the line and the camera point)
						// n = v1 x v2 (cross product)
						// v1 : vector from the camera to the point
						//      (x0, y0) of the line
						// v2 : vector from the camera to the point
						//      (x0 + dx, y0 + dy) of the line
						nx = sphereRadius * dy;
						ny = -sphereRadius * dx;
						nz = dy * (x0 - centerX) - dx * (y0 - centerY);
						if (nz < 0.0f)
						{
							nx = -nx;
							ny = -ny;
							nz = -nz;
						}

						// Remove the line contribution from the
						// gaussian hemisphere Hough transform
						for (iPhi = -windowHalf, offset = 0;
							 iPhi < (int)nPhi + windowHalf;
							 iPhi++, offset += nTheta2)
						{
							t =   nx * cosTable[iPhi + windowHalf]
								+ ny * sinTable[iPhi + windowHalf];
							i = (int)floor(t / (float)sqrt(nz * nz + t * t) * nTheta);
							if (iPhi > -windowHalf)
								atipDrawSegmentY2MinusAccum32(previousI + windowHalf,
											i + windowHalf, hough - nTheta2 + offset,
											nTheta2, lCoeffs[l]);
							previousI = i;
						}
					}
				}
			}
		}

		// If the detected point is finite
		else
		{
			infiniteVanPoints[pass] = false;

			// Compute the coordinates inside the original image
			// x = t.r.sin(theta).cos(phi)
			// y = t.r.sin(theta).sin(phi)
			// z = -t.r.cos(theta) = -r		so t = 1 / cos(theta)
			x = sphereRadius * (float)tan(theta);
			vanPoints[(pass << 2) + 3] = y = x * (float)sin(phi) + centerY;
			vanPoints[(pass << 2) + 2] = x = x * (float)cos(phi) + centerX;

			// For each remaining detected line, test the proximity of the detected point
			// and then add them to the current cluster
			for (l = 0; l < nLines; l++)
			{
				dx = srcLines[(l << 2) + 2];
				dy = srcLines[(l << 2) + 3];

				if ((dx != 0.0f) || (dy != 0.0f))
				{
					// Distance of the line to the origin
					x0 = srcLines[ l << 2     ];
					y0 = srcLines[(l << 2) + 1];
					d = dy * x0 - dx * y0;

					// If the distance from to current point to the current line is
					// less than a threshold, the line is considered inside the current
					// cluster
					if (fabs(dx * y - dy * x + d) <= maxLineDistancePixels)
					{
						// Copy the current line to the current cluster
						memcpy(tempLineClusters[pass] + (nLinesPerCluster[pass] << 2),
							   srcLines + (l << 2), 4 * sizeof(float));
						memset(srcLines + (l << 2), 0, 4 * sizeof(float));
						nLinesPerCluster[pass]++;

						// Compute the normal of the interpretation plane (contains the line
						// and the camera point)
						// n = v1 x v2 (cross product)
						// v1 : vector from the camera to the point (x0, y0) of the line
						// v2 : vector from the camera to the point (x0 + dx, y0 + dy) of the line
						nx = sphereRadius * dy;
						ny = -sphereRadius * dx;
						nz = dy * (x0 - centerX) - dx * (y0 - centerY);
						if (nz < 0.0f)
						{
							nx = -nx;
							ny = -ny;
							nz = -nz;
						}

						// Remove the line contribution from the
						// gaussian hemisphere Hough transform
						for (iPhi = -windowHalf, offset = 0;
							 iPhi < (int)nPhi + windowHalf;
							 iPhi++, offset += nTheta2)
						{
							t =   nx * cosTable[iPhi + windowHalf]
								+ ny * sinTable[iPhi + windowHalf];
							i = (int)floor(t / (float)sqrt(nz * nz + t * t) * nTheta);
							if (iPhi > -windowHalf)
								atipDrawSegmentY2MinusAccum32(previousI + windowHalf,
											i + windowHalf, hough - nTheta2 + offset,
											nTheta2, lCoeffs[l]);
							previousI = i;
						}
					}
				}
			}
		}

		// If two passes are done, and if the two vanishing points directions are
		// too similar, remove the second and do an additional pass
		if (pass == 1)
			if (infiniteVanPoints[0] == infiniteVanPoints[1])
				if (fabs(SphereDotProduct(vanPoints[0], vanPoints[1],
									 vanPoints[4], vanPoints[5]))
						> ATIP_SIMILAR_VANPOINTS_THRESHOLD)
				{
					retries++;
					if (retries >= ATIP_RETRIES_MAX_NUMBER)
					{
						atipSetError(ATIP_NO_THREE_DIFFERENT_VAN_POINTS);
						return false;
					}
					pass--;
					continue;
				}

		// If three passes are done, and if two vanishing points directions are
		// too similar, remove one and do an additional pass
		if (pass == 2)
		{
			if (infiniteVanPoints[0] == infiniteVanPoints[2])
				if (fabs(SphereDotProduct(vanPoints[0], vanPoints[1],
									 vanPoints[8], vanPoints[9]))
						> ATIP_SIMILAR_VANPOINTS_THRESHOLD)
				{
					retries++;
					if (retries >= ATIP_RETRIES_MAX_NUMBER)
					{
						atipSetError(ATIP_NO_THREE_DIFFERENT_VAN_POINTS);
						return false;
					}
					pass--;
					continue;
				}

			if (infiniteVanPoints[1] == infiniteVanPoints[2])
				if (fabs(SphereDotProduct(vanPoints[4], vanPoints[5],
									 vanPoints[8], vanPoints[9]))
						> ATIP_SIMILAR_VANPOINTS_THRESHOLD)
				{
					retries++;
					if (retries >= ATIP_RETRIES_MAX_NUMBER)
					{
						atipSetError(ATIP_NO_THREE_DIFFERENT_VAN_POINTS);
						return false;
					}
					pass--;
					continue;
				}
		}
	}

	// Sort the vanishing points (finite ones at the beginning)
	unsigned int order[] = { 0, 1, 2 };
	float tv[4];
	unsigned int ti;
	if (infiniteVanPoints[0] && !infiniteVanPoints[1])
	{
		memcpy(tv, vanPoints, 4 * sizeof(float));
		memcpy(vanPoints, vanPoints + 4, 4 * sizeof(float));
		memcpy(vanPoints + 4, tv, 4 * sizeof(float));
		ti = order[0];		order[0] = order[1];			order[1] = ti;
		ti = nLinesPerCluster[0];	nLinesPerCluster[0] = nLinesPerCluster[1];
		nLinesPerCluster[1] = ti;
		infiniteVanPoints[0] = false;
		infiniteVanPoints[1] = true;
	}
	if (infiniteVanPoints[1] && !infiniteVanPoints[2])
	{
		memcpy(tv, vanPoints + 4, 4 * sizeof(float));
		memcpy(vanPoints + 4, vanPoints + 8, 4 * sizeof(float));
		memcpy(vanPoints + 8, tv, 4 * sizeof(float));
		ti = order[1];		order[1] = order[2];			order[2] = ti;
		ti = nLinesPerCluster[1];	nLinesPerCluster[1] = nLinesPerCluster[2];
		nLinesPerCluster[2] = ti;
		infiniteVanPoints[1] = false;
		infiniteVanPoints[2] = true;
	}
	if (infiniteVanPoints[0] && !infiniteVanPoints[1])
	{
		memcpy(tv, vanPoints, 4 * sizeof(float));
		memcpy(vanPoints, vanPoints + 4, 4 * sizeof(float));
		memcpy(vanPoints + 4, tv, 4 * sizeof(float));
		ti = order[0];		order[0] = order[1];			order[1] = ti;
		ti = nLinesPerCluster[0];	nLinesPerCluster[0] = nLinesPerCluster[1];
		nLinesPerCluster[1] = ti;
		infiniteVanPoints[0] = false;
		infiniteVanPoints[1] = true;
	}

	// Create correctly sized arrays for the line clusters and copy the data
	ATIP_NEW_ARRAY(float, lineClusters[0], nLinesPerCluster[0] * 4);
	ATIP_NEW_ARRAY(float, lineClusters[1], nLinesPerCluster[1] * 4);
	ATIP_NEW_ARRAY(float, lineClusters[2], nLinesPerCluster[2] * 4);
	ATIP_NEW_ARRAY(float, lineClusters[3], (nLines - nLinesPerCluster[0]
					- nLinesPerCluster[1] - nLinesPerCluster[2]) * 4);
	memcpy(lineClusters[0], tempLineClusters[order[0]],
		   nLinesPerCluster[0] * 4 * sizeof(float));
	memcpy(lineClusters[1], tempLineClusters[order[1]],
		   nLinesPerCluster[1] * 4 * sizeof(float));
	memcpy(lineClusters[2], tempLineClusters[order[2]],
		   nLinesPerCluster[2] * 4 * sizeof(float));
	nLinesPerCluster[3] = 0;
	for (l = 0; l < nLines; l++)
		if ((srcLines[(l << 2) + 2] != 0.0f) || (srcLines[(l << 2) + 3] != 0.0f))
		{
			memcpy(lineClusters[3] + (nLinesPerCluster[3] << 2),
				   srcLines + (l << 2), 4 * sizeof(float));
			nLinesPerCluster[3]++;
		}
	ATIP_ASSERT(nLinesPerCluster[3] == (nLines - nLinesPerCluster[0]
					- nLinesPerCluster[1] - nLinesPerCluster[2]));

	ATIP_DELETE_ARRAY(tempLineClusters[0]);
	ATIP_DELETE_ARRAY(tempLineClusters[1]);
	ATIP_DELETE_ARRAY(tempLineClusters[2]);
	ATIP_DELETE_ARRAY(srcLines);
	ATIP_DELETE_ARRAY(lCoeffs);
	ATIP_DELETE_ARRAY(hough2);
	ATIP_DELETE_ARRAY(cosTable);
	ATIP_DELETE_ARRAY(sinTable);

	return true;
}

//----------------------------------------------------------------------------------------

bool atipDrawVanPoints24(unsigned char * dstImage, const unsigned char * srcImage,
						 unsigned int width, unsigned int height,
						 const float vanPoints[12], const bool infiniteVanPoints[3],
						 float * const lineClusters[4],
						 const unsigned int nLinesPerCluster[4])
{
	//! Width of the cross representing a vanishing point
#define ATIP_CROSS_SIZE		100

	ATIP_ASSERT(   dstImage && srcImage && (width > 0) && (height > 0)
				&& vanPoints && infiniteVanPoints && lineClusters && nLinesPerCluster);
	if (   !dstImage || !srcImage || (width == 0) || (height == 0) || !vanPoints
		|| !infiniteVanPoints || !lineClusters || !nLinesPerCluster)
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Copy the source image to the destination image
	memcpy(dstImage, srcImage, width * height * 3);

	// Center of the image
	float centerX = (float)width / 2.0f - 0.25f;
	float centerY = (float)height / 2.0f - 0.25f;

	// Color of the line clusters
	unsigned char lineColor[] = {	96,		0,		0,			// First vanishing point
									0,		96,		0,			// Second vanishing point
									0,		0,		96,			// Third vanishing point
									192,	192,	192 };		// Remaining lines

	// Color of the crosses
	unsigned char crossColor[] = {	255,	0,		0,			// First vanishing point
									0,		255,	0,			// Second vanishing point
									0,		128,	255	};		// Third vanishing point

	// For each clustered line
	float x0, y0, dx, dy;
	float t[4];
	int ix0, iy0, ix1, iy1;
	unsigned int lx0, ly0, lx1, ly1;
	unsigned int cluster, l;
	for (cluster = 0; cluster < 4; cluster++)
		for (l = 0; l < nLinesPerCluster[cluster]; l++)
		{
			// Point on the line
			x0 = lineClusters[cluster][ l << 2     ];
			y0 = lineClusters[cluster][(l << 2) + 1];

			// Direction vector of the line
			dx = lineClusters[cluster][(l << 2) + 2];
			dy = lineClusters[cluster][(l << 2) + 3];

			// t parameter for the 4 image edge points
			t[0] = -x0 / dx;
			t[1] = (float(width - 1) - x0) / dx;
			t[2] = -y0 / dy;
			t[3] = (float(height - 1) - y0) / dy;

			// Bubble sort of the the t parameters
			if (t[0] > t[1]) swap(t[0], t[1]);
			if (t[1] > t[2]) swap(t[1], t[2]);
			if (t[2] > t[3]) swap(t[2], t[3]);
			if (t[0] > t[1]) swap(t[0], t[1]);
			if (t[1] > t[2]) swap(t[1], t[2]);
			if (t[0] > t[1]) swap(t[0], t[1]);
			
			ix0 = (int)floor(x0 + t[1] * dx);
			if (ix0 < 0) lx0 = 0;
			else if (ix0 >= (int)width) lx0 = width - 1;
			else lx0 = (unsigned int)ix0;
			iy0 = (int)floor(y0 + t[1] * dy);
			if (iy0 < 0) ly0 = 0;
			else if (iy0 >= (int)height) ly0 = height - 1;
			else ly0 = (unsigned int)iy0;
			ix1 = (int)floor(x0 + t[2] * dx);
			if (ix1 < 0) lx1 = 0;
			else if (ix1 >= (int)width) lx1 = width - 1;
			else lx1 = (unsigned int)ix1;
			iy1 = (int)floor(y0 + t[2] * dy);
			if (iy1 < 0) ly1 = 0;
			else if (iy1 >= (int)height) ly1 = height - 1;
			else ly1 = (unsigned int)iy1;
			atipDrawSegment24(dstImage, width, height, lx0, ly0, lx1, ly1,
							  lineColor[cluster * 3], lineColor[cluster * 3 + 1],
							  lineColor[cluster * 3 +2]);
		}

	// For each vanishing point
	int xp, yp, xi, yi;
	unsigned char * center;
	unsigned int p, v;
	for (v = 0; v < 3; v++)
	{
		// If the vanishing point is infinite, draw a line from the center of the image
		if (infiniteVanPoints[v])
		{
			// Point on the line
			x0 = centerX;
			y0 = centerY;

			// Direction vector of the line
			dx = vanPoints[(v << 2) + 2];
			dy = vanPoints[(v << 2) + 3];

			// t parameter for the 4 image edge points
			t[0] = -x0 / dx;
			t[1] = (float(width - 1) - x0) / dx;
			t[2] = -y0 / dy;
			t[3] = (float(height - 1) - y0) / dy;

			// Bubble sort of the the t parameters
			if (t[0] > t[1]) swap(t[0], t[1]);
			if (t[1] > t[2]) swap(t[1], t[2]);
			if (t[2] > t[3]) swap(t[2], t[3]);
			if (t[0] > t[1]) swap(t[0], t[1]);
			if (t[1] > t[2]) swap(t[1], t[2]);
			if (t[0] > t[1]) swap(t[0], t[1]);

			ix0 = (int)floor(x0 + t[1] * dx);
			if (ix0 < 0) lx0 = 0;
			else if (ix0 >= (int)width) lx0 = width - 1;
			else lx0 = (unsigned int)ix0;
			iy0 = (int)floor(y0 + t[1] * dy);
			if (iy0 < 0) ly0 = 0;
			else if (iy0 >= (int)height) ly0 = height - 1;
			else ly0 = (unsigned int)iy0;
			ix1 = (int)floor(x0 + t[2] * dx);
			if (ix1 < 0) lx1 = 0;
			else if (ix1 >= (int)width) lx1 = width - 1;
			else lx1 = (unsigned int)ix1;
			iy1 = (int)floor(y0 + t[2] * dy);
			if (iy1 < 0) ly1 = 0;
			else if (iy1 >= (int)height) ly1 = height - 1;
			else ly1 = (unsigned int)iy1;
			if (t[1] > 0.0f)
				atipDrawSegment24(dstImage, width, height, lx0, ly0,
								  (unsigned int)floor(x0), (unsigned int)floor(y0),
								  crossColor[v * 3], crossColor[v * 3 + 1],
								  crossColor[v * 3 + 2]);
			else
				atipDrawSegment24(dstImage, width, height, (unsigned int)floor(x0),
								  (unsigned int)floor(y0), lx1, ly1,
								  crossColor[v * 3], crossColor[v * 3 + 1],
								  crossColor[v * 3 + 2]);
		}

		// If the vanishing point is finite, draw a cross
		else
		{
			xp = (int)floor(vanPoints[(v << 2) + 2]);
			yp = (int)floor(vanPoints[(v << 2) + 3]);
			center = dstImage + (yp * (int)width + xp) * 3;
			for (xi = xp - ATIP_CROSS_SIZE, p = -ATIP_CROSS_SIZE * 3;
				xi <= xp + ATIP_CROSS_SIZE; xi++, p += 3)
					if (   (xi >= 0) && (xi < (int)width)
						&& (yp >= 0) && (yp < (int)height))
						memcpy(center + p, crossColor + v * 3, 3);

			for (yi = yp - ATIP_CROSS_SIZE, p = -ATIP_CROSS_SIZE * width * 3;
				yi <= yp + ATIP_CROSS_SIZE; yi++, p += width * 3)
					if (   (xp >= 0) && (xp < (int)width)
						&& (yi >= 0) && (yi < (int)height))
						memcpy(center + p, crossColor + v * 3, 3);
		}
	}
	return true;
}