/****************************************************************************************/
/*																						*/
/*							Automatic Tour Into the Picture (ATIP)						*/
/*																						*/
/****************************************************************************************/

//! \file	atip_hough.cpp
//! \author	BOULANGER Kevin
//! \date	13th July 2004
//! \brief	Hough transforms


#include <van_point/atip_hough.h>
#include <tools/atip_private_tools.h>
#include <tools/atip_primitives.h>
#include <math.h>


bool atipPolarHough(unsigned int * & hough, int & rhoMax, unsigned int & maxHough,
					const int * points, unsigned int nPoints,
					unsigned int width, unsigned int height, unsigned int nTheta)
{
	ATIP_ASSERT(points && (nPoints > 0) && (nTheta > 0));
	if (!points || (nPoints == 0) || (nTheta == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Fixed point numbers are used to improve the computation speed

	// Compute the cosinus and sinus tables
	int * cosTable = 0;
	ATIP_NEW_ARRAY(int, cosTable, nTheta);
	int * sinTable = 0;
	ATIP_NEW_ARRAY(int, sinTable, nTheta);
	unsigned int theta;
	double dTheta;
	for (theta = 0; theta < nTheta; theta++)
	{
		dTheta = theta / (double)nTheta * M_2PI;
		cosTable[theta] = (int)floor(cos(dTheta) * 65535.0);
		sinTable[theta] = (int)floor(sin(dTheta) * 65535.0);
	}

	// Create the destination array
	rhoMax = (int)ceil(sqrt(double(width * width + height * height)));
	ATIP_NEW_ARRAY(unsigned int, hough, nTheta * rhoMax);
	memset(hough, 0, rhoMax * nTheta * sizeof(unsigned int));

	// For each point, do the Hough 1 to m transform
	unsigned int nCoords = nPoints << 1;
	int rho;
	maxHough = 0;
	unsigned int newAccum;
	unsigned int offset;
	unsigned int p;
	for (p = 0; p < nCoords; p += 2)
	{
		for (theta = 0, offset = 0; theta < nTheta; theta++, offset += rhoMax)
		{
			rho = ( ( (points[p] * cosTable[theta] + points[p + 1] * sinTable[theta])
					>> 16) + rhoMax ) >> 1;
			ATIP_ASSERT((rho >= 0) && (rho < (int)rhoMax));
			newAccum = ++hough[offset + rho];
			if (newAccum > maxHough) maxHough = newAccum;
		}
	}

	ATIP_DELETE_ARRAY(cosTable);
	ATIP_DELETE_ARRAY(sinTable);

	return true;
}

//----------------------------------------------------------------------------------------

bool atipPolarHoughThreshold(float * & lines, unsigned int * & lineCoeffs,
							 unsigned int & nLines,
							 const unsigned int * houghHigh, const unsigned int * hough,
							 unsigned int nTheta, int rhoMax,
							 unsigned int maxHoughHigh, float threshold)
{
	ATIP_ASSERT(   houghHigh && hough && (nTheta > 0) && (rhoMax > 0)
				&& (maxHoughHigh > 0) && (threshold > 0.0f) && (threshold < 1.0f));
	if (   !houghHigh || !hough || (nTheta == 0) || (rhoMax <= 0)
		|| (maxHoughHigh == 0) || (threshold <= 0.0f) || (threshold >= 1.0f))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	unsigned int iThresh = (unsigned int)floor(threshold * (float)maxHoughHigh);
	float * tempLines = 0;
	ATIP_NEW_ARRAY(float, tempLines, rhoMax * nTheta * 4);
	unsigned int * tempCoeffs = 0;
    ATIP_NEW_ARRAY(unsigned int, tempCoeffs, rhoMax * nTheta);
	nLines = 0;
	unsigned int theta, p;
	int rho;
	double r, t;
	for (theta = 0, p = 0; theta < nTheta; theta++)
		for (rho = 0; rho < rhoMax; rho++, p++)
			// If a pixel of the Hough transform is greater than the threshold,
			// store the (rho, theta) pair
			if (houghHigh[p] >= iThresh)
			{
				// Rho
				r = double(((rho << 1) - rhoMax) << 16) / 65535.0;

				// Theta
				t = (double)theta / (double)nTheta * M_2PI;

				// Point on the line
				tempLines[ nLines << 2     ] = (float)(r * cos(t));
				tempLines[(nLines << 2) + 1] = (float)(r * sin(t));

				// Direction vector of the line
				tempLines[(nLines << 2) + 2] = (float)cos(t + M_0_5PI);
				tempLines[(nLines << 2) + 3] = (float)sin(t + M_0_5PI);

				// Coefficient of the line
				tempCoeffs[nLines] = hough[p];
				nLines++;
			}

	// Create correctly sized arrays for the lines
	ATIP_NEW_ARRAY(float, lines, nLines * 4);
	memcpy(lines, tempLines, nLines * 4 * sizeof(float));
	ATIP_DELETE_ARRAY(tempLines);
	ATIP_NEW_ARRAY(unsigned int, lineCoeffs, nLines);
	memcpy(lineCoeffs, tempCoeffs, nLines * sizeof(unsigned int));
	ATIP_DELETE_ARRAY(tempCoeffs);
	return true;
}

//----------------------------------------------------------------------------------------

bool atipDrawHoughLines24(unsigned char * dstImage, const unsigned char * srcImage,
						  unsigned int width, unsigned int height,
						  const float * lines, const unsigned int * lineCoeffs,
						  unsigned int maxLineCoeff, unsigned int nLines)
{
	ATIP_ASSERT(   dstImage && srcImage && (width > 0) && (height > 0)
				&& lines && (nLines > 0));
	if (   !dstImage || !srcImage || (width == 0) || (height == 0)
		|| !lines || (nLines == 0))
	{
		atipSetError(ATIP_INVALID_ARGUMENT);
		return false;
	}

	// Copy the source image to the destination image
	memcpy(dstImage, srcImage, width * height * 3);
	
	// For each line
	float x0, y0, dx, dy;
	float t[4];
	int ix0, iy0, ix1, iy1;
	unsigned int lx0, ly0, lx1, ly1;
	unsigned int l;
	for (l = 0; l < nLines; l++)
	{
		// Point on the line
		x0 = lines[l << 2];
		y0 = lines[(l << 2) + 1];

		// Direction vector of the line
		dx = lines[(l << 2) + 2];
		dy = lines[(l << 2) + 3];

		// t parameter for the 4 image edge points
		t[0] = -x0 / dx;
		t[1] = (float(width - 1) - x0) / dx;
		t[2] = -y0 / dy;
		t[3] = (float(height - 1) - y0) / dy;

		// Bubble sort of the the t parameters
		if (t[0] > t[1]) swap(t[0], t[1]);
		if (t[1] > t[2]) swap(t[1], t[2]);
		if (t[2] > t[3]) swap(t[2], t[3]);
		if (t[0] > t[1]) swap(t[0], t[1]);
		if (t[1] > t[2]) swap(t[1], t[2]);
		if (t[0] > t[1]) swap(t[0], t[1]);
		
		ix0 = (int)floor(x0 + t[1] * dx);
		if (ix0 < 0) lx0 = 0;
		else if (ix0 >= (int)width) lx0 = width - 1;
		else lx0 = (unsigned int)ix0;
		iy0 = (int)floor(y0 + t[1] * dy);
		if (iy0 < 0) ly0 = 0;
		else if (iy0 >= (int)height) ly0 = height - 1;
		else ly0 = (unsigned int)iy0;
		ix1 = (int)floor(x0 + t[2] * dx);
		if (ix1 < 0) lx1 = 0;
		else if (ix1 >= (int)width) lx1 = width - 1;
		else lx1 = (unsigned int)ix1;
		iy1 = (int)floor(y0 + t[2] * dy);
		if (iy1 < 0) ly1 = 0;
		else if (iy1 >= (int)height) ly1 = height - 1;
		else ly1 = (unsigned int)iy1;
		unsigned char color = lineCoeffs[l] * 255 / maxLineCoeff;
		atipDrawSegment24(dstImage, width, height, lx0, ly0, lx1, ly1,
						  color, color, color);
	}

	return true;
}
